{*****************************************************************************}
{  								              }
{	Copyright (c) 2002  Vampyre SoftWorks			              }
{								              }
{								              }                                                                             
{*****************************************************************************}
{								              }
{	TABLEMAIN.PAS						              }
{	Simple class for managing string grids		                      }
{								              }
{	Created by:						              }
{	Marek "Mar" "Galfar"  Mauder  [pentar@seznam.cz]	              }
{                                                                             }
{ Created:                                                                    }
{ 15.03.2002                                                                  }
{							                      }
{	Last Modified:						              }
{	29.09.2002						              }
{								              }
{*****************************************************************************}



unit TableMain;

interface

uses Windows, Classes, SysUtils, WinUtils, FileVFormat, Graphics;

type

  ENoGrid = class(Exception);
  ENoProps = class(Exception);
  EResize = class(Exception);
  ESaveError = class(Exception);
  ELoadError = class(Exception);
  EUnknownFmt = class(Exception);
  EExport = class(Exception);

  PDynStrGrid = ^TDynStrGrid;
  TDynStrGrid = array of array of string;

  PDynLongArray = ^TDynLongArray;
  TDynLongArray = array of LongInt;

  TTable = class(TPersistent)
  private
    FName: string;
    FColCount: LongInt;
    FRowCount: LongInt;
    FCompress: boolean;

    FCells: TDynStrGrid;
    FColWidths: TDynLongArray;
    FRowHeights: TDynLongArray;

    FHeaderFont:TFont;
    FValueFont:TFont;
    FTitleFont:TFont;

    FComment: TStringList;

    procedure Resize(ACols, ARows: LongInt);

  protected

    procedure SetRowCount(Value: LongInt); virtual;
    procedure SetColCount(Value: LongInt); virtual;
    procedure SetFont(Idx:LongInt; Value:TFont);virtual;

    procedure SetCellIdx(ACol, ARow: Integer; Value: string); virtual;
    procedure SetCellStr(ACol, ARow: string; Value: string); virtual;
    function GetCellIdx(ACol, ARow: Integer): string; virtual;
    function GetCellStr(ACol, ARow: string): string; virtual;

    procedure SetColWidth(Idx: LongInt; AWidth: LongInt);
    procedure SetRowHeight(Idx: LongInt; AHeight: LongInt);

    function GetColWidth(Idx: LongInt): LongInt;
    function GetRowHeight(Idx: LongInt): LongInt;

    procedure SetComment(Value: TStringList);

  public

    constructor Create(ACols: LongInt = 10; ARows: LongInt = 20); virtual;
    destructor Destroy; override;

    procedure SaveToStream(Stream: TStream); virtual;
    procedure SaveToFile(AName: string); virtual;
    procedure LoadFromStream(Stream: TStream); virtual;
    procedure LoadFromFile(AName: string); virtual;

    procedure FillCells(str: string; stCol: LongInt = 0; stRow: LongInt = 0);
    procedure SetDefaultMeasures;

    procedure ExportTextFile(FileName: string);
    procedure ExportCSVXP(FileName: string);
    procedure ExportCSV2K(FileName: string);


    procedure Assign(Source: TPersistent); override;

    property CellsIdx[ACol, ARow: Integer]: string read GetCellIdx write
    SetCellIdx;
    property CellsStr[ACol, ARow: string]: string read GetCellStr write
    SetCellStr;

    property ColWidths[Index: LongInt]: LongInt read GetColWidth write
      SetColWidth;
    property RowHeights[Index: LongInt]: LongInt read GetRowHeight write
      SetRowHeight;


  published

    property Name: string read FName write FName;
    property RowCount: LongInt read FRowCount write SetRowCount;
    property ColCount: LongInt read FColCount write SetColCount;
    property Compress: boolean read FCompress write FCompress;
    property Comment: TStringList read FComment write SetComment;

    property HeaderFont:TFont index 0 read FHeaderFont write SetFont;
    property TitleFont:TFont index 1 read FTitleFont write SetFont;
    property ValueFont:TFont index 2 read FValueFont write SetFont;


  end;


type

  THdr = record
    ID: array[0..3] of char;
    S1: char;
    Info: array[0..19] of char;
    VerHi: word;
    VerLo: word;
    Reserved: array[0..11] of char;

    S2: char;
  end;

const

  VerHI = 0;
  VerLO = 3;

  MaxRows = 256;
  MinRows = 2;

  MaxCols = 256;
  MinCols = 2;

  DefColWidth = 70;
  DefRowHeight = 20;

  MinColWidth = 8;
  MaxColWidth = 256;

  MinRowHeight = 2;
  MaxRowHeight = 64;

  TabEXT = 'VDT';
  TabID = 'VDT@';
  TabINFO = '<VAMPYRE DATA TABLE>';

  TabHDR: THdr = (ID: TabID; S1: #0; Info: TabINFO;
    VERHI: VerHI; VERLO: VerLO; Reserved: ''; S2: #0);

  sDefFontSize=10;

const

  sTableProps = 'Table Props';
  sStringGrid = 'String Grid';
  sFileOptions = 'File Options';
  sMeasures = 'Measures';
  sComment = 'Comment';


  sNoGrid = 'Error - No String Grid in file';
  sNoProps = 'Error - No Table Properties in file';
  sResize = 'Error when resizing grid';
  sExport = 'Error when exporting table';
  sSaveError = 'Error when saving file';
  sLoadError = 'Error when loading file';
  sUnknownFmt = 'Unknown Table format';

  sDefaultFont='Courrier New';

implementation

uses TableExport;

constructor TTable.Create(ACols, ARows: LongInt);
begin
  inherited Create;
  FName := 'untitled';
  FColCount := ACols;
  FRowCount := ARows;
  FComment := TStringList.Create;
  FCompress:=false;
  Resize(FColCount, FRowCount);

  FHeaderFont:=TFont.Create;
  FTitleFont:=TFont.Create;
  FValueFont:=TFont.Create;

  FHeaderFont.Name:=sDefaultFont;
  FTitleFont.Name:=sDefaultFont;
  FValueFont.Name:=sDefaultFont;

  FHeaderFont.Size:=sDefFontSize;
  FTitleFont.Size:=sDefFontSize;
  FValueFont.Size:=sDefFontSize;

end;



destructor TTable.Destroy;
begin

  FComment.Free;
  FRowHeights := nil;
  FColWidths := nil;
  FCells := nil;

  FHeaderFont.Free;
  FTitleFont.Free;
  FValueFont.Free;

  inherited Destroy;
end;


procedure TTable.Resize(ACols, ARows: LongInt);
var
  IsNew: boolean;
  Temp: TDynStrGrid;
  x, y: LongInt;
begin
  if FCells = nil then
    Isnew := true
  else
    IsNew := false;

  ARows := Min(MaxRows, Max(MinRows, ARows));
  ACols := Min(MaxCols, Max(MinCols, ACols));

  try

    SetLength(Temp, FColCount, FRowCount);
    if not IsNew then
      for x := 0 to Min(ACols - 1, FColCount - 1) do
        for y := 0 to Min(ARows - 1, FRowCount - 1) do
          Temp[x, y] := FCells[x, y];


    FCells := nil;
    SetLength(FCells, ACols, ARows);

    if not IsNew then
      for x := 0 to Min(ACols - 1, FColCount - 1) do
        for y := 0 to Min(ARows - 1, FRowCount - 1) do
          FCells[x, y] := Temp[x, y];


    Temp := nil;


    SetLength(FColWidths, ACols);
    SetLength(FRowHeights, ARows);

    if ACols > FColCount then
      for x := FColCount to ACols - 1 do
        ColWidths[x] := DefColWidth;

    if ARows > FRowCount then
      for x := FRowCount to ARows - 1 do
        RowHeights[x] := DefRowHeight;

  except
    raise EResize.Create(sResize);
  end;

  FColCount := ACols;
  FRowCount := ARows;

end;

procedure TTable.SetRowCount(Value: LongInt);
begin
  Resize(FColCount, Value);
end;

procedure TTable.SetColCount(Value: LongInt);
begin
  Resize(Value, FRowCount);
end;


procedure TTable.SetCellIdx(ACol, ARow: Integer; Value: string);
begin
  FCells[ACol, ARow] := Value;
end;

procedure TTable.SetCellStr(ACol, ARow: string; Value: string);
var C, R: LongInt;
  FC, FR: LongInt;
begin
  if (ACol = '') and (ARow = '') then Exit;
  FC := -1;
  FR := -1;

  for C := 0 to ColCount - 1 do
    if Compstr(ACol, FCells[C, 0]) then
    begin
      FC := C;
      Break;
    end;

  for R := 0 to RowCount - 1 do
    if Compstr(ARow, FCells[0, R]) then
    begin
      FR := R;
      Break;
    end;

  if (FC >= 0) and (FR >= 0) then
    FCells[FC, FR] := Value;

end;

procedure TTable.SetFont(Idx:LongInt; Value:TFont);
begin
 case Idx of
  0:FHeaderFont.Assign(Value);
  1:FTitleFont.Assign(Value);
  2:FValueFont.Assign(Value);
 end;
end;


function TTable.GetCellIdx(ACol, ARow: Integer): string;
begin
  Result := FCells[ACol, ARow];
end;


function TTable.GetCellStr(ACol, ARow: string): string;
var C, R: LongInt;
  FC, FR: LongInt;
begin
  Result := '';
  if (ACol = '') and (ARow = '') then Exit;
  FC := -1;
  FR := -1;

  for C := 0 to ColCount - 1 do
    if Compstr(ACol, FCells[C, 0]) then
    begin
      FC := C;
      Break;
    end;

  for R := 0 to RowCount - 1 do
    if Compstr(ARow, FCells[0, R]) then
    begin
      FR := R;
      Break;
    end;

  if (FC >= 0) and (FR >= 0) then
    Result := FCells[FC, FR];

end;


procedure TTable.SaveToStream(Stream: TStream);
var x, y: longint;
  TF: TVFile;
  MS: TMemoryStream;
begin

  TF := TVFile.Create;
  TF.Header.ID := TabID;
  TF.Header.Desc := TabINFO;
  TF.Header.VerHI := VerHI;
  TF.Header.VerLO := VerLO;
  TF.Header.Compress := Compress;


  MS := TMemoryStream.Create;

  try
    MS.Clear;
    WriteStr(MS, Name);
    MS.WriteBuffer(FRowCount, SizeOf(FRowCount));
    MS.WriteBuffer(FColCount, SizeOf(FColCount));
    TF.AddData(sTableProps, MS);

    MS.Clear;

    for x := 0 to FColCount - 1 do
      for y := 0 to FRowCount - 1 do
        WriteStr(MS, CellsIdx[x, y]);
    TF.AddData(sStringGrid, MS);

    MS.Clear;
    MS.WriteBuffer(FCompress, SizeOf(FCompress));
    MS.WriteBuffer(x, SizeOf(x)); //reserved bytes
    MS.WriteBuffer(x, SizeOf(x));
    MS.WriteBuffer(x, SizeOf(x));
    TF.AddData(sFileOptions, MS);

    MS.Clear;
    for x := 0 to FColCount - 1 do
      MS.WriteBuffer(FColWidths[x], SizeOf(FColWidths[x]));
    for x := 0 to FRowCount - 1 do
      MS.WriteBuffer(FRowHeights[x], SizeOf(FRowHeights[x]));
    TF.AddData(sMeasures, MS);

    MS.Clear;
    FComment.SaveToStream(MS);
    TF.AddData(sComment, MS);



    TF.SaveToStream(Stream);

  except
    raise ESaveError.Create(sSaveError);
  end;

  TF.Free;
  MS.Free;

end;

procedure TTable.SaveToFile(AName: string);
var
  Stream: TStream;
begin
  Stream := TFileStream.Create(AName, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TTable.LoadFromStream(Stream: TStream);
var Hdr: THdr;
  VDO: TVDataObj;
  ID: TID;
  Pos: Int64;
  TF: TVFile;
  x, y, i: longint;
begin

  Pos := Stream.Position;
  Stream.ReadBuffer(ID, SizeOf(ID));
  Stream.Position := Pos;
  if ID = TabID then
  begin
    try
      Stream.ReadBuffer(HDR, SizeOf(HDR));

      Name := ReadStr(Stream);
      Stream.ReadBuffer(y, SizeOf(FRowCount));
      Stream.ReadBuffer(x, SizeOf(FColCount));
      Stream.ReadBuffer(i, SizeOf(i));
      Stream.ReadBuffer(i, SizeOf(i));

      Resize(x, y);

      for x := 0 to FColCount - 1 do
        for y := 0 to FRowCount - 1 do
          CellsIdx[x, y] := ReadStr(Stream);

      SetDefaultMeasures;

    except
      raise ESaveError.Create(sSaveError);
    end;

  end
  else
    if ID = sVFID then
    begin

      TF := TVFile.Create('', Stream);

      try
        Pos := TF.GetData(sTableProps, VDO);
        if Pos <> -1 then
        begin
          Name := ReadStr(VDO.Stream);
          VDO.Stream.ReadBuffer(y, SizeOf(FRowCount));
          VDO.Stream.ReadBuffer(x, SizeOf(FColCount));
          VDO.Free;
        end
        else
          raise ENoProps.Create(sNoProps);


        Pos := TF.GetData(sStringGrid, VDO);
        if Pos <> -1 then
        begin
          Resize(x, y);
          for x := 0 to FColCount - 1 do
            for y := 0 to FRowCount - 1 do
              CellsIdx[x, y] := ReadStr(VDO.Stream);
          VDO.Free;
        end
        else
        begin
          raise ENoGrid.Create(sNoGrid);
        end;


        Pos := TF.GetData(sFileOptions, VDO);
        if Pos <> -1 then
        begin
          VDO.Stream.ReadBuffer(FCompress, SizeOf(FCompress));
          VDO.Free;
        end
        else
        begin
          FCompress := false;
        end;


        Pos := TF.GetData(sMeasures, VDO);
        if Pos <> -1 then
        begin
          for x := 0 to FColCount - 1 do
            VDO.Stream.ReadBuffer(FColWidths[x], SizeOf(FColWidths[x]));
          for x := 0 to FRowCount - 1 do
            VDO.Stream.ReadBuffer(FRowHeights[x], SizeOf(FRowHeights[x]));
          VDO.Free;
        end
        else
        begin
          SetDefaultMeasures;
        end;


        Pos := TF.GetData(sComment, VDO);
        if Pos <> -1 then
        begin
          FComment.LoadFromStream(VDO.Stream);
          VDO.Free;
        end
        else
        begin
          FComment.Clear;
        end;



      except
        raise ELoadError.Create(sLoadError);
      end;

      TF.Free;

    end

    else
      raise EUnknownFmt.Create(sUnknownFmt);

end;

procedure TTable.LoadFromFile(AName: string);
var
  Stream: TStream;
begin
  if not FileExists(AName) then Exit;
  Stream := TFileStream.Create(AName, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;


procedure TTable.FillCells(str: string; stCol: LongInt; stRow: LongInt);
var ix, iy: LongInt;
begin

  for ix := stCol to ColCount - 1 do
    for iy := stRow to RowCount - 1 do
      CellsIdx[ix, iy] := str;

end;


procedure TTable.SetColWidth(Idx: LongInt; AWidth: LongInt);
begin
  if Idx in [0..FColCount - 1] then
    FColWidths[Idx] := Min(MaxColWidth, Max(MinColWidth, AWidth));

end;

procedure TTable.SetRowHeight(Idx: LongInt; AHeight: LongInt);
begin
  if Idx in [0..FRowCount - 1] then
    FRowHeights[Idx] := Min(MaxRowHeight, Max(MinRowHeight, AHeight));

end;

function TTable.GetColWidth(Idx: LongInt): LongInt;
begin
  if Idx in [0..FColCount - 1] then
    Result := FColWidths[Idx]
  else
    Result := DefColWidth;
end;

function TTable.GetRowHeight(Idx: LongInt): LongInt;
begin
  if Idx in [0..FRowCount - 1] then
    Result := FRowHeights[Idx]
  else
    Result := DefRowHeight;


end;

procedure TTable.SetDefaultMeasures;
var i: LongInt;
begin

  for i := 0 to FColCount - 1 do
    FColWidths[i] := DefColWidth;

  for i := 0 to FRowCount - 1 do
    FRowHeights[i] := DefRowHeight;

end;

procedure TTable.SetComment(Value: TStringList);
begin
  FComment.Assign(Value);
end;

procedure TTable.Assign(Source: TPersistent);
var Tab: TTable;
  I, J: LongInt;
begin
  if Source is TTable then
  begin
    Tab := TTable(Source);
    Name := Tab.Name;
    Comment := Tab.Comment;
    Resize(Tab.ColCount, Tab.RowCount);
    Compress := Tab.Compress;
    for I := 0 to ColCount - 1 do
      for J := 0 to RowCount - 1 do
        FCells[I, J] := Tab.FCells[I, J];

    for I := 0 to ColCount - 1 do
      FColWidths[I] := Tab.ColWidths[I];

    for I := 0 to RowCount - 1 do
      FRowHeights[I] := Tab.RowHeights[I];

  end
  else
    inherited Assign(Source);

end;

procedure TTable.ExportTextFile(FileName: string);
begin
  try
    TableToText(Self, FileName);
  except
    raise EExport.Create(sExport);
  end;
end;


procedure TTable.ExportCSVXP(FileName: string);
begin
  try
    TableToCSVXP(Self, FileName);
  except
    raise EExport.Create(sExport);
  end;
end;


procedure TTable.ExportCSV2K(FileName: string);
begin
  try
    TableToCSV2K(Self, FileName);
  except
    raise EExport.Create(sExport);
  end;
end;




end.

