unit FileIOEngine;

interface

uses Windows,SysUtils,Classes,WinUtils,Engines,BaseObjs,
  FileRec,Archivers,BaseMgrs;

type

  EFileIO=class(Exception)
  end;

  TFileIOEngine=class(TEngine)
  private
    FLastFile:TFileRec;
    FCacheS:DWord;
    FFiles:TBaseMgr;
    FArchivers:TSortedMgr;

    procedure CreateArchivers;

    procedure SetLastFile(FR:TFileRec);
    procedure CheckCache;

  protected

     function GetCFile(Index:longInt):TFileRec;
     function GetCUsed:Dword;
     function GetFCount:LongInt;
     function GetArc(Index:LongInt):TArchiver;
     function GetArcCount:LongInt;

     procedure SetCache(Value:DWord);

  public
    constructor Create(AOwner:TEngine=nil); override;
    destructor Destroy; override;

    class function GetInstance:TFileIOEngine;

    procedure Init;override;
    procedure Done;override;

    function GetFile(AName:string):TFileRec;virtual; // for loaders that can load from memory - images,sounds,models,...
    function GetFullPath(AName:string):string;virtual; // for loaders that can't load from memory - videos,music,...
    function IsFile(AName:string):boolean;virtual;


    function AddArc(Arc:TArchiver):LongInt;
    function GetArcByName(AName:string):TArchiver;  // obj reference
    procedure SetArcOrder(Arc:TArchiver;Ord:LongInt);
    procedure ClearCache;
    property CacheFiles[Index:LongInt]:TFileRec read GetCFile;
    property Archivers[Index:LongInt]:TArchiver read GetArc;
    property ArcCount:LongInt read GetArcCount;
    property FilesCount:LongInt read GetFCount;
    property LastFile:TFileRec read FLastFile;
    property CacheSize:DWord read FCacheS write SetCache;  //in kilobytes
    property CacheUsed:DWord read GetCUsed;                // in kB

  published

  end;



implementation

uses
    RawArc,
    ZipArc;     //default archiver

var
  FE:TFileIoEngine;

constructor TFileIOEngine.Create(AOwner:TEngine);
begin
 inherited Create(AOwner);
 FCacheS:=4096;
 Timer.Enabled:=false;
 FFiles:=TBaseMgr.Create();
 FArchivers:=TSortedMgr.Create();
 FArchivers.Sorting:=stDescending;
 CreateArchivers;

 FE.Free;
 FE:=Self;

end;

destructor TFileIOEngine.Destroy;
begin

 ClearCache;
 FArchivers.Free;
 FFiles.Free;
 FE:=nil;
 inherited Destroy;
end;

class function TFileIOEngine.GetInstance:TFileIOEngine;
begin
 if FE=nil
 then
 begin
 FE:=TFileIOEngine.Create();
 FE.Init;
 end;
 Result:=FE;
end;

procedure TFileIOEngine.CreateArchivers;
var I:LongInt;
begin
for I:=0 to RegArchivers.Count-1 do
begin
 AddArc(TArchiverInfo(RegArchivers[I]).Cls.Create);
end;
end;

procedure TFileIOEngine.Init;
var I:LongInt;
begin
 for I:=0 to FArchivers.Count-1 do
  Archivers[I].Init;


 inherited Init;
end;


procedure TFileIOEngine.Done;
var I:LongInt;
begin
 inherited Done;

 for I:=0 to FArchivers.Count-1 do
  Archivers[I].Done;

 ClearCache;

end;

function TFileIOEngine.GetFile(AName:string):TFileRec;
var I:LongInt;
    FR:TFileRec;
begin
FR:=nil;

 for I:=0 to FilesCount-1 do
  if CompStr(ExtractFilename(CacheFiles[I].Name),ExtractFileName(AName))then
  begin
   FR:=TFileRec.Create();
   CacheFiles[I].CopyTo(FR);
  end;


 if FR=nil then
 for I:=0 to ArcCount-1 do
  begin
   FR:=Archivers[I].GetFile(AName);
   if FR<>nil then
      Break;
   end;


if FR=nil then
 raise EFileIO.CreateFmt('FileIOEngine Error : File not found [ %s ]',[AName])
 else
  SetLastFile(FR);

Result:=FR;


end;

function TFileIOEngine.GetFullPath(AName:string):string;
var I:LongInt;
begin
Result:='';

 for I:=0 to ArcCount-1 do
  begin
   Result:=Archivers[I].GetFullPath(AName);
   if Result<>'' then
      Break;
   end;


if Result='' then
 raise EFileIO.CreateFmt('FileIOEngine Error : File not found [ %s ]',[AName]);

end;

function TFileIOEngine.IsFile(AName:string):boolean;
var I:Longint;
begin
 Result:=false;
 for I:=0 to ArcCount-1 do
  if Archivers[I].IsFileInArc(AName)then
   begin
    Result:=true;
    Exit;
   end;

end;

procedure TFileIOEngine.SetCache(Value:DWord);
begin
FCacheS:=Value;
CheckCache;
end;

function TFileIOEngine.GetCFile(Index:longInt):TFileRec;
begin
 Result:=nil;
 if not ((Index>-1)and(Index<FilesCount)) then Exit;
 Result:=TFileRec(FFiles[Index]);

end;

function TFileIOEngine.GetArc(Index:LongInt):TArchiver;
begin
 Result:=nil;
 if not ((Index>-1)and(Index<FArchivers.Count)) then Exit;
 Result:=TArchiver(FArchivers[Index]);

end;

function TFileIOEngine.GetCUsed:Dword;
var I:LongInt;
begin
 Result:=0;
 for I := 0 to FilesCount-1 do
  Result:=Result+CacheFiles[I].Size;

 Result:=Result div 1024;

end;

function TFileIOEngine.GetFCount:LongInt;
begin
 Result:=FFiles.Count;
end;

function TFileIOEngine.GetArcCount:LongInt;
begin
 Result:=FArchivers.Count;
end;

function TFileIOEngine.AddArc(Arc:TArchiver):LongInt;
begin
Result:=-1;
if Arc<>nil then
begin
 Result:=FArchivers.Add(Arc);
end;
end;

procedure TFileIOEngine.ClearCache;
begin
 FFiles.Clear;

end;

procedure TFileIOEngine.SetLastFile(FR:TFileRec);
var I:LongInt;
begin
 for I:=0 to FilesCount-1 do
  if CompStr(ExtractFilename(CacheFiles[I].Name),ExtractFileName(FR.Name))then
   begin
    FLastFile:=CacheFiles[I];
    FLastFile.RefCount:=FLastFile.RefCount+1;
    Exit;
   end;


 FLastFile:=TFileRec.Create();
 FR.CopyTo(FLastFile);
 FLastFile.RefCount:=1;
 FFiles.Add(TBaseObj(FLastFile));
 CheckCache;

end;

procedure TFileIOEngine.CheckCache;
var Idx,I,MinRef:LongInt;

begin

while CacheUsed>CacheSize do
begin
 Idx:=-1;
 MinRef:=$FFFF;
 for I:=0 to FilesCount-1 do
  if CacheFiles[I].RefCount<MinRef then
   begin
   MinRef:=CacheFiles[I].RefCount;
   Idx:=I;
   end;
 FFiles.Delete(Idx);

end;


end;


function TFileIOEngine.GetArcByName(AName:string):TArchiver;
var I:LOngInt;
begin
Result:=nil;
for I:=0 to ArcCount-1 do
 if CompStr(Archivers[I].Name,AName)then
  begin
   Result:=Archivers[I];
   Exit;
  end;


end;

procedure TFileIOEngine.SetArcOrder(Arc:TArchiver;Ord:LongInt);
begin
 if Arc=nil then Exit;
 if FArchivers.IndexOf(Arc)<0 then Exit;
 FArchivers.Extract(Arc);
 Arc.Order:=Ord;
 Farchivers.Add(Arc);


end;

end.
