unit Fonts;

interface

uses Windows, SysUtils, Classes, WinUtils, BaseObjs, Objs3D, Types3D, FileDirs,
  Utils3D, Colors, Textures, BaseMgrs, Graphics, Images,
  CharMgrs, Chars, FontTypes, GfxTypes, ImgUtils, FileRec;

type

  // if loading Windows fonts make sure that font files are outside of
  // archives - AddFontResource can register only files not memory.
  //  ( There is AddFontMemResourceEx that can load from archives
  //  but is supported only on Windows 2000 and Windows XP and there
  //  are still many people who have Win98/WinMe (blee))

  // IMPORTANT WARNING
  // When usign Windows fonts make sure that font file name has same
  // name as font face name.After registering font with AddFontResource we know
  // only file name not font face name and font can not be found by GDI.
  // Example:  Call__.ttf contains font internaly named as Calligraphia
  //    >> we must rename file to Calligraphia.ttf
  // You can find out font face name using standart Window app called fontview.exe


  TFont = class(TBaseObj)
  private
    FChars: TCharMgr;
    FGDIFont: Graphics.TFont;
    FFileName: string;
    FAddResult: LongInt;

    FHeight: LongInt;
    FColor: TARGB;
    FOutlineColor: TARGB;
    FStyle: TFontStyles;
    FCanvas: TCanvas;
    FCharsHeight: LongInt;
    FRenInited: boolean;

    FUsedCharSet: TFontChars;
    FChanged: boolean;
    FTexSize: LongInt;

    procedure LoadWindowsFont(AFile: string);
    procedure LoadBitMapFont(AFile: string);
    procedure LoadVampFont(AFile: string);
    procedure InitChar(C: Char; Obj: TChar);

  protected

    function GetIsWindowsFont: boolean;

    procedure SetColor(Value: TARGB);
    procedure SetOutlineColor(Value: TARGB);
    procedure SetHeight(Value: LongInt);
    procedure SetStyle(Value: TFontStyles);

  public
    constructor Create; override;
    destructor Destroy; override;

    procedure Init; override;
    procedure Done; override;

    procedure Load(AFile: string);
    procedure LoadRegistered(AFaceName: string);
      // loads registered Windows Font ;use 'Arial',...

    procedure BuildChars;
    procedure FreeChars;

    procedure Save(AFile: string; Compress: boolean = false); overload;
      // saves chars to Vampyre Font File
    procedure Export(AFile: string); overload;
      // saves chars to Image  (ALine - number of chars in line   0=all chars in one line
    procedure Save(AStream: TStream; Compress: boolean = false); overload;
    procedure Export(AStream: TStream); overload;


    function TextWidth(Text: string; AHeight: LongInt = 0): LongInt;


    //  functios for rendering text using Chars

    procedure RenderInit;
    procedure RenderDone;

    procedure RenderText(X, Y: single; Text: string); overload;
    procedure RenderText(X, Y: single; ACol: TARGB; AHeight: single; Text:
      string); overload;

    procedure RenderShadowText(X, Y: single; Text: string); overload;
    procedure RenderShadowText(X, Y: single; ACol, ShadowCol: TARGB; AHeight,
      SOffset: single; Text:
      string); overload;

    procedure RenderAlphaText(X, Y: single; Alpha: byte; Text: string);
      overload;
    procedure RenderAlphaText(X, Y: single; ACol: TARGB; Alpha: byte; AHeight:
      single; Text:
      string); overload;

    procedure RenderTextFX(X, Y: single; Text: string; FX: TFontFX); overload;
    procedure RenderTextFX(X, Y: single; ACol: TARGB; AHeight: single; Text:
      string; FX: TFontFX); overload;




    property FileName: string read FFileName;
    property IsWindowsFont: boolean read GetIsWindowsFont;
    property GDIFont: Graphics.TFont read FGDIFont;
    property Chars: TCharMgr read FChars;
    property UsedCharSet: TFontChars read FUsedCharSet write FUsedCharset;
    property CharTextureSize: LongInt read FTexSize write FTexSize;

          // dyn propertys
    property Height: LongInt read FHeight write SetHeight;

          // static propertys - must call TFont.BuildChars for Win Fonts
    property Color: TARGB read FColor write SetColor;
    property Style: TFontStyles read FStyle write SetStyle;
    property OutlineColor: TARGB read FOutlineColor write SetOutlineColor;


  published



  end;



implementation

uses GGraphic, FontFiles, GFileIO;


constructor TFont.Create;
begin
  inherited Create;
  FChars := TCharMgr.Create;
  FColor := $FFFFFFFF;
  FOutlineColor := $FF000000;
  FHeight := 20;
  FStyle := [fsBold];
  FRenInited := false;
  FChanged := false;
  FTexSize := CharTexSize;
end;

destructor TFont.Destroy;
begin
  Done;
  FreeChars;
  FChars.Free;
  FGDIFont.Free;
  inherited Destroy;
end;

procedure TFont.Init;
begin
  inherited Init;
  if FChanged then
    BuildChars;
  Chars.InitDevTextures;
end;

procedure TFont.Done;
begin
  inherited Done;
  Chars.FreeDevTextures;
end;

function TFont.GetIsWindowsFont: boolean;
begin
  Result := (FGDIFont <> nil);
end;

procedure TFont.SetColor(Value: TARGB);
begin
  FColor := Value;
  if IsWindowsFont then
  begin
    FGDIFont.Color := ARGBToBGR(FColor);
  end;
  FChanged := true;
end;

procedure TFont.SetOutlineColor(Value: TARGB);
begin
  FOutlineColor := Value;
  FChanged := true;
end;

procedure TFont.SetHeight(Value: LongInt);
begin
  FHeight := Value;
  if IsWindowsFont then
  begin
    FGDIFont.Height := ARGBToBGR(FColor);
  end;
end;

procedure TFont.SetStyle(Value: TFontStyles);
var FS: Graphics.TFontStyles;
begin
  FStyle := Value;
  if IsWindowsFont then
  begin
    FS := [];
    if fsBold in Style then
      Include(FS, Graphics.fsBold);
    if fsItalic in Style then
      Include(FS, Graphics.fsItalic);
    if fsUnderLine in Style then
      Include(FS, Graphics.fsUnderLine);
    if fsStrikeOut in Style then
      Include(FS, Graphics.fsStrikeOut);
    FGDIFont.Style := FS;
  end;
  if fsFreeImages in FStyle then
  begin
    if not Inited then Exit;
    if Chars.Char[0] = nil then Exit;
    if Chars.Char[0].Texture.Image <> nil then
      Chars.FreeImages;
  end;
  FChanged := true;
end;


procedure TFont.Load(AFile: string);
begin
  case GetFileType(AFile) of
    ftGraphic: LoadBitmapFont(AFile);
    ftFont:
      if CompStr(ExtractFileExt(AFile), '.' + FontFiles.FntEXT) then
        LoadVampFont(AFile)
      else
        LoadWindowsFont(AFile);
  else
    begin
      raise EFontError.Create('Error in TFont.Load - "' + ExtractFileName(AFile)
        + '" is not supported font file.');
      Exit;
    end;
  end;
end;

procedure TFont.LoadRegistered(AFaceName: string);
begin
  if FGDIFont = nil then
    FGDIFont := Graphics.TFont.Create;
  FGDIFont.Name := AFaceName;
  Name := FGDIFont.Name;

  Color := FColor;
  Height := FHeight;
  Style := FStyle;

  FAddResult := 0;
  FChanged := true;
end;

procedure TFont.LoadWindowsFont(AFile: string);
var Path: string;
begin
  Path := FileIO.GetFullPath(AFile);
  FAddResult := AddFontResource(PChar(Path));
  if FAddResult = 0 then
  begin
    raise EFontError.Create('Error in TFont.LoadWindowsFont - Font file "' +
      ExtractFileName(AFile) + '" can not be registered by Windows.');
    Exit;
  end;

  Name := ChangeFileExt(ExtractFileName(AFile), '');
  FFileName := Path;
  if FGDIFont = nil then
    FGDIFont := Graphics.TFont.Create;

  FGDIFont.Name := Name;
  Color := FColor;
  Height := FHeight;
  Style := FStyle;
  FChanged := true;
end;

procedure TFont.LoadBitMapFont(AFile: string);
var FFile: FileRec.TFileRec;
begin
  FFile := FileIO.GetFile(AFile);
  FontFiles.StreamFmt := FindFormat(ExtractFileExt(AFile));
  FontFiles.ImportFont(Self, FFile.Stream);
  FontFiles.StreamFmt := nil;
  FFile.Free;

  FGDIFont.Free;
  FGDIFont:=nil;

 //if Inited then
  //Chars.InitDevTextures;
end;

procedure TFont.LoadVampFont(AFile: string);
var FFile: FileRec.TFileRec;
begin

  FFile := FileIO.GetFile(AFile);
  FontFiles.LoadFont(Self, FFile.Stream);
  FFile.Free;

  FGDIFont.Free;
  FGDIFont:=nil;

  // if Inited then
//  Chars.InitDevTextures;
end;

procedure TFont.InitChar(C: Char; Obj: TChar);
var
  Len, X, Y, S, I: LongInt;
  Alpha: TImage;
begin
  Len := FCanvas.TextWidth(C);
  Obj.Texture.Image.Resize(FTexSize, FTexSize);
  Obj.Texture.Image.BitCount := 24;
  if (fsOutlined in Style) or (fsOutlineOnly in Style) then
  begin
    Obj.Texture.Image.ClearARGB(FOutlineColor);
    Obj.Texture.Image.Canvas.Font := FCanvas.Font;
    SetBkMode(Obj.Texture.Image.Canvas.Handle, Transparent);

    X := (FTexSize - Len) div 2;
    Y := (FTexSize - FCanvas.TextHeight(C)) div 2;

    if not (fsOutLineOnly in Style) then
    begin
      Obj.Texture.Image.Canvas.TextOut(X, Y, C);
      if not (fsNoAntialias in Style) then ;
      ImgUtils.FastBlur(Obj.Texture.Image);
    end;
  end
  else
    Obj.Texture.Image.ClearARGB(FColor);

  Alpha := TImage.Create();
  Alpha.Resize(FTexSize, FTexSize);
  Alpha.BitCount := 24;
  Alpha.ClearARGB($000000);
  Alpha.Canvas.Font := FCanvas.Font;
  Alpha.Canvas.Font.Color := clWhite;
  if not ((fsOutlined in Style) or (fsOutlineOnly in Style)) then
    Alpha.Canvas.Font.Height := FTexSize - 2;

  Len := Alpha.Canvas.TextWidth(C);

  X := (FTexSize - Len) div 2;
  Y := (FTexSize - Alpha.Canvas.TextHeight(C)) div 2;

  SetBkMode(Alpha.Canvas.Handle, Transparent);
  S := 0;

  if (fsOutlined in Style) or (fsOutlineOnly in Style) then
  begin
    S := FTexSize div 32 * 2;

    for I := 1 to S do
    begin
      Alpha.Canvas.TextOut(X, Y - I, C);
      Alpha.Canvas.TextOut(X, Y + I, C);
      Alpha.Canvas.TextOut(X + I, Y, C);
      Alpha.Canvas.TextOut(X - I, Y, C);

      Alpha.Canvas.TextOut(X - I, Y - I, C);
      Alpha.Canvas.TextOut(X + I, Y - I, C);
      Alpha.Canvas.TextOut(X + I, Y + I, C);
      Alpha.Canvas.TextOut(X - I, Y + I, C);
    end;

    if fsOutlineOnly in Style then
    begin
      Alpha.Canvas.Font.Color := clBlack;
      Alpha.Canvas.TextOut(X, Y, C);
    end;

  end
  else
    Alpha.Canvas.TextOut(X, Y, C);

  if not (fsNoAntiAlias in Style) then
    ImgUtils.FastBlur(Alpha);

  GreyScale(Alpha, 8);
  Obj.Texture.Image.SetAlphaChannel(Alpha);

  Alpha.Free;

  Obj.Width := Len + S;
  if fsFreeImages in Style then
  begin
       //  Obj.Texture.Image.Free;
    Obj.Texture.Image := nil;
  end;


end;

procedure TFont.BuildChars;
var I, Len: LongInt;
  Ch: TChar;

  procedure PreInit;
  begin
    FCanvas := TCanvas.Create;
    FCanvas.Handle := CreateCompatibleDC(0);
    FCanvas.Font := FGDIFont;
    FCanvas.Font.Height := FTexSize - 2;

    if (fsOutlined in Style) or (fsOutlineOnly in Style) then
      FCanvas.Font.Height := FTexSize - 2 - 2 * FTexSize div 32;

    FCharsHeight := FCanvas.TextHeight(FUsedCharSet);

  end;

  procedure AfterInit;
  begin
    DeleteDC(FCanvas.Handle);
    FCanvas.Free;
  end;



begin
  if not IsWindowsFont then Exit;
  //if not Inited then Exit;
  FChanged := false;

  FreeChars;

  FUsedCharSet := ACharSet;
  Len := Length(FUsedCharSet); // not using TFontEngine.ActiveCharSet so that
                          // TFont can be used independently from TFontEngine

  PreInit;

  for I := 1 to Len do
  begin
    Ch := TChar.Create(FUsedCharSet[I]);
    InitChar(FUsedCharSet[I], Ch);
    FChars.Add(Ch);
  end;

  AfterInit;

end;

procedure TFont.FreeChars;
begin
  FChars.FreeDevTextures;
  FChars.FreeImages;
  FChars.Clear;
end;

procedure TFont.Save(AFile: string; Compress: boolean);
begin
  FontFiles.SaveFont(Self, AFile, Compress);
end;

procedure TFont.Export(AFile: string);
begin
  FontFiles.ExportFont(Self, AFile);
end;

procedure TFont.Save(AStream: TStream; Compress: boolean);
begin
  FontFiles.SaveFont(Self, AStream, Compress);
end;

procedure TFont.Export(AStream: TStream);
begin
  FontFiles.ExportFont(Self, AStream);
end;


procedure TFont.RenderInit;
begin
  if FRenInited then Exit;
  Graphic.PushViewPort;
  Graphic.PushFeatures;
  Graphic.PushDWords;
  Renderer.SetDWord(sdTexAdrX, TEXADR_CLAMP);
  Renderer.SetDWord(sdTexAdrY, TEXADR_CLAMP);
  Renderer.SetDWord(sdCullMode, CULL_CW);
  ViewPort.Orthogonal(Renderer.UsedParams.Width, Renderer.UsedParams.Height);
  Graphic.Disable(gfLighting);
  Graphic.Disable(gfZBuffer);
  FRenInited := true;
end;

procedure TFont.RenderDone;
begin
  if not FRenInited then Exit;
  Graphic.PopFeatures([gfLighting, gfZBuffer]);
  Graphic.PopDWords([sdTexAdrX, sdTexAdrY, sdCullMode]);
  Graphic.PopViewPort;
  FRenInited := false;
end;



procedure TFont.RenderText(X, Y: single; Text: string);
begin
  RenderText(X, Y, $FFFFFFFF, FHeight, Text);
end;

procedure TFont.RenderText(X, Y: single; ACol: TARGB; AHeight: single; Text:
  string);
var I: LongInt;
  Ch: TChar;
  Org, P: TPoint3S;
  Pos: single;
  Scale: single;
  FI: boolean;
begin
  if not Inited then Exit;

  if not FRenInited then
    FI := true
  else
    FI := false;
  RenderInit;

  Org := Point3S(X, Y, 0);
  Pos := 0;
  Scale := AHeight / Chars.Char[0].Texture.Height;
  for I := 0 to Length(Text) - 1 do
  begin
    Ch := Chars[Text[I + 1]];
    if Ch = nil then
    begin
      Pos := Pos + Chars.Char[0].Width * Scale;
      Continue;
    end;

    P := Point3S(X + Pos - ((Ch.Texture.Width - Ch.Width) div 2) * Scale, Y, 0);

    Renderer.DrawTriStrip([
      P,
        Point3S(P.X + Ch.Texture.Width * Scale, P.Y, P.Z),
        Point3S(P.X, P.Y + Ch.Texture.Height * Scale, P.Z),
        Point3S(P.X + Ch.Texture.Width * Scale, P.Y + Ch.Texture.Height * Scale,
        P.Z)],
        [Coord(0, 0), Coord(1, 0), Coord(0, 1), Coord(1, 1)],
      ACol, Ch.Texture, nil);

    Pos := Pos + Ch.Width * Scale;
  end;

  if FI then
    RenderDone;

end;

procedure TFont.RenderShadowText(X, Y: single; Text: string);
begin
  RenderShadowText(X, Y, $FFFFFFFF, $FF000000, FHeight, Max(1.0, FHeight / 20),
    Text);
end;

procedure TFont.RenderShadowText(X, Y: single; ACol, ShadowCol: TARGB; AHeight,
  SOffset: single; Text:
  string);
begin
  RenderText(X + SOffset, Y + SOffset, ShadowCol, AHeight, Text);
  RenderText(X, Y, ACol, AHeight, Text);
end;

procedure TFont.RenderAlphaText(X, Y: single; Alpha: byte; Text: string);
begin
  RenderAlphaText(X, Y, $FFFFFFFF, Alpha, FHeight, Text);
end;

procedure TFont.RenderAlphaText(X, Y: single; ACol: TARGB; Alpha: byte; AHeight:
  single; Text:
  string);
begin
  RenderText(X, Y, SetA(ACol, Alpha), AHeight, Text);
end;

procedure TFont.RenderTextFX(X, Y: single; Text: string; FX: TFontFX);
begin
  RenderTextFX(X, Y, $FFFFFFFF, FHeight, Text, FX);
end;

procedure TFont.RenderTextFX(X, Y: single; ACol: TARGB; AHeight: single; Text:
  string; FX: TFontFX);
var I: LongInt;
  Ch: TChar;
  P: TPoint3S;
  SX, SY, Pos: single;
  Scale: single;
  FI: boolean;
  Coords: array[0..3] of TTexCoord;
  Points: array[0..3] of TPoint3S;
  Mat: TMatrix;
  Org: TPoint3S;

  procedure DrawChars(X, Y: single; Col: TARGB);
  var I, J: LongInt;
  begin
    Pos := 0;
    Scale := AHeight / Chars.Char[0].Texture.Height;

    for I := 0 to Length(Text) - 1 do
    begin
      Ch := Chars[Text[I + 1]];
      if Ch = nil then
      begin
        Pos := Pos + Chars.Char[0].Width * Scale * SX;
        Continue;
      end;

      P := Point3S(X + Pos - ((Ch.Texture.Width - Ch.Width) div 2) * Scale * SX,
        Y, 0);
      Points[0] := P;
      Points[1] := Point3S(P.X + Ch.Texture.Width * Scale * SX, P.Y, P.Z);
      Points[2] := Point3S(P.X, P.Y + Ch.Texture.Height * Scale * SY, P.Z);
      Points[3] := Point3S(P.X + Ch.Texture.Width * Scale * SX, P.Y +
        Ch.Texture.Height * Scale * SY, P.Z);

      if fxRotate in FX.FX then
      begin
        Mat := CreateRotationMatrixZ(Sin(DegToRad(FX.Angle)),
          Cos(DegToRad(FX.Angle)));
        Org := Point3S(X, Y, 0);
        Points[0] := AddPoint(Org, MatrixTransform(SubPoint(Points[0], Org),
          Mat));
        Points[1] := AddPoint(Org, MatrixTransform(SubPoint(Points[1], Org),
          Mat));
        Points[2] := AddPoint(Org, MatrixTransform(SubPoint(Points[2], Org),
          Mat));
        Points[3] := AddPoint(Org, MatrixTransform(SubPoint(Points[3], Org),
          Mat));
      end;

      Renderer.DrawTriStrip(Points, Coords, Col, Ch.Texture, nil);

      Pos := Pos + Ch.Width * Scale * SX;
    end;

  end;

begin
  if not Inited then Exit;

  if not FRenInited then
    FI := true
  else
    FI := false;
  RenderInit;

  if fxBlend in FX.FX then
  begin
    Renderer.SetDWord(sdSrcBlend, FX.SrcBlend);
    Renderer.SetDWord(sdDstBlend, FX.DestBlend);
  end;

  if fxAlpha in FX.FX then
    ACol := SetA(ACol, FX.Alpha);

  SX := 1;
  SY := 1;

  if fxStretch in FX.FX then
  begin
    SX := FX.StretchX;
    SY := FX.StretchY;
  end;

  Coords[0] := Coord(0, 0);
  Coords[1] := Coord(1, 0);
  Coords[2] := Coord(0, 1);
  Coords[3] := Coord(1, 1);

  if fxFlip in FX.FX then
  begin
    if FX.FlipX then
    begin
      Coords[0].X := 1;
      Coords[1].X := 0;
      Coords[2].X := 1;
      Coords[3].X := 0;
    end;
    if FX.FlipY then
    begin
      Coords[0].Y := 1;
      Coords[1].Y := 1;
      Coords[2].Y := 0;
      Coords[3].Y := 0;
    end;
  end;


  if fxShadow in FX.FX then
    DrawChars(X + FX.SOffsetX, Y + FX.SOffsetY, FX.SColor);

  DrawChars(X, Y, ACol);

  Graphic.PopDWords([sdSrcBlend, sdDstBlend]);

  if FI then
    RenderDone;

end;

function TFont.TextWidth(Text: string; AHeight: LongInt): LongInt;
var Pos: single;
  Scale: single;
  I: LongInt;
  Ch: TChar;
begin
  if AHeight = 0 then
    AHeight := FHeight;

  Pos := 0;
  Scale := AHeight / Chars.Char[0].Texture.Height;
  for I := 0 to Length(Text) - 1 do
  begin
    Ch := Chars[Text[I + 1]];
    if Ch = nil then
    begin
      Pos := Pos + Chars.Char[0].Width * Scale;
      Continue;
    end;


    Pos := Pos + Ch.Width * Scale;
  end;

  Result := Round(Pos);
end;



end.

