unit Cameras;

interface

uses Windows, SysUtils, Classes, WinUtils, BaseObjs, GfxTypes,
  Renderers, Objs3D, Types3D, Utils3D;

type

  TCamera = class(TBase3DObj)
  private
    FUp: TPoint3S;
    FFrustum: TFrustum;

  protected

    procedure SetPosition(Value: TPoint3S); override;
    procedure SetAngle(Value: TPoint3S); override;
    procedure SetUp(Value: TPoint3S); virtual;

  public
    constructor Create; override;
    destructor Destroy; override;

    procedure Init; override;
    procedure Done; override;
    procedure Draw; override;

    procedure Assign(Source: TBaseObj); override;

    procedure SetIdentityView; virtual;
    procedure UpdateView; virtual;
    procedure Configure(APos, ADir, AUp: TPoint3S); virtual;

    procedure CalcClipInfo; virtual;

    function IsLineVisible(P1, P2: TPoint3S): Boolean;
    function IsPointVisible(P: TPoint3S): Boolean;
    function IsBoxVisible(P1, P2: TPoint3S): boolean;overload;
    function IsBoxVisible(BB: TBoundingInfo): boolean;overload;
    function IsSphereVisible(Center : TPoint3S;Radius : single):boolean;


  published

    property Up: TPoint3S read FUp write SetUp;

  end;




implementation

uses GGraphic,ViewPorts;

constructor TCamera.Create;
begin
  inherited Create;
  Up := Point3S(0, 0, 1);
  Position := Point3S(0, 0, 0);
  LookAt(Point3S(100, 100, 0));

end;

destructor TCamera.Destroy;
begin
  Done;
  inherited Destroy;
end;


procedure TCamera.Init;
begin
  inherited Init;
  UpdateView;

end;


procedure TCamera.Done;
begin
  inherited Done;

end;

procedure TCamera.Assign(Source: TBaseObj);
var C: TCamera;
begin
  inherited Assign(Source);
  if not (Source is TCamera) then Exit;
  C := TCamera(Source);
  Up := C.Up;
  FFrustum:=C.FFrustum;

end;

procedure TCamera.SetPosition(Value: TPoint3S);
begin
  inherited SetPosition(Value);
  UpdateView;
  CalcClipInfo;
end;

procedure TCamera.SetAngle(Value: TPoint3S);
begin
  inherited SetAngle(Value);
  UpdateView;
  CalcClipInfo;
end;

procedure TCamera.SetUp(Value: TPoint3S);
begin
  FUp := Value;
  UpdateView;
end;


procedure TCamera.SetIdentityView;
begin
  if not Inited then Exit;
  Renderer.ViewIdentity;
end;

procedure TCamera.UpdateView;
begin
  if not Inited then Exit;
  if ViewPort.Projection = pmOrthogonal then
    SetIdentityView
  else
    Renderer.ViewLookAt(Position, Direction, Up);
end;

procedure TCamera.Draw;
begin


end;

procedure TCamera.Configure(APos, ADir, AUp: TPoint3S);
begin
  Done;
  Position := APos;
  LookAt(ADir);
  Up := AUp;
  Init;
end;

procedure TCamera.CalcClipInfo;
var
  View, Proj: TMatrix;
  PF: TPoint3S;
  Clip: TMatrix;
  V:TViewPort;
begin
  if not Inited then Exit;

  View := Renderer.GetViewMatrix;
  V:=ViewPort;
  CreateProjectionMatrix(Proj,V.Fov,1/V.AspectRatio,V.NearPlane,V.BackPlane);
  Clip := MatrixMultiply(View, Proj);

  FFrustum[fpRight].a := clip._14 - clip._11;
  FFrustum[fpRight].b := clip._24 - clip._21;
  FFrustum[fpRight].c := clip._34 - clip._31;
  FFrustum[fpRight].d := clip._44 - clip._41;
  NormalizePlane(FFrustum[fpRight]);

  FFrustum[fpLeft].a := clip._14 + clip._11;
  FFrustum[fpLeft].b := clip._24 + clip._21;
  FFrustum[fpLeft].c := clip._34 + clip._31;
  FFrustum[fpLeft].d := clip._44 + clip._41;
  NormalizePlane(FFrustum[fpLeft]);

  FFrustum[fpBottom].a := clip._14 + clip._12;
  FFrustum[fpBottom].b := clip._24 + clip._22;
  FFrustum[fpBottom].c := clip._34 + clip._32;
  FFrustum[fpBottom].d := clip._44 + clip._42;
  NormalizePlane(FFrustum[fpBottom]);

  FFrustum[fpTop].a := clip._14 - clip._12;
  FFrustum[fpTop].b := clip._24 - clip._22;
  FFrustum[fpTop].c := clip._34 - clip._32;
  FFrustum[fpTop].d := clip._44 - clip._42;
  NormalizePlane(FFrustum[fpTop]);

  FFrustum[fpFar].a := clip._14 - clip._13;
  FFrustum[fpFar].b := clip._24 - clip._23;
  FFrustum[fpFar].c := clip._34 - clip._33;
  FFrustum[fpFar].d := clip._44 - clip._43;
  NormalizePlane(FFrustum[fpFar]);

  FFrustum[fpNear].a := clip._14 + clip._13;
  FFrustum[fpNear].b := clip._24 + clip._23;
  FFrustum[fpNear].c := clip._34 + clip._33;
  FFrustum[fpNear].d := clip._44 + clip._43;
  NormalizePlane(FFrustum[fpNear]);



end;

function TCamera.IsLineVisible(P1, P2: TPoint3S): Boolean;
var
  F: TFrustumPlanes;
begin
  if IsPointVisible(P1) or IsPointVisible(P2) then
    Result := true
  else
  begin
    Result := false;
    if IsBoxVisible(P1, P2) then
      Result := true;
  end;

end;

function TCamera.IsPointVisible(P: TPoint3S): Boolean;
var
  I: TFrustumPLanes;
begin
  for I := TFrustumPlanes(0) to TFrustumPlanes(5) do
    if (FFrustum[I].a * p.x + FFrustum[i].b * p.y + FFrustum[i].c * p.z
      + FFrustum[I].d <= 0) then
    begin
      Result := false;
      Exit;
    end;
  Result := true;
end;

function TCamera.IsBoxVisible(P1, P2: TPoint3S): boolean;
var I: TFrustumPlanes;
  amin, amax, bmin, bmax, cmin, cmax: single;

begin

  for I := TFrustumPlanes(0) to TFrustumPlanes(5) do
  begin
    amin := FFrustum[I].a * P1.x;
    bmin := FFrustum[I].b * P1.y;
    cmin := FFrustum[I].c * P1.z;

    if (amin + bmin + cmin + FFrustum[I].d > 0) then continue;

    amax := FFrustum[I].a * P2.x;

    if (amax + bmin + cmin + FFrustum[I].d > 0) then continue;

    bmax := FFrustum[I].b * P2.y;

    if (amin + bmax + cmin + FFrustum[I].d > 0) then continue;
    if (amax + bmax + cmin + FFrustum[I].d > 0) then continue;

    cmax := FFrustum[I].c * P2.z;

    if (amin + bmin + cmax + FFrustum[I].d > 0) then continue;
    if (amax + bmin + cmax + FFrustum[I].d > 0) then continue;
    if (amin + bmax + cmax + FFrustum[I].d > 0) then continue;
    if (amax + bmax + cmax + FFrustum[I].d > 0) then continue;

    Result := false;
    Exit;

  end;
  Result := true
end;

function TCamera.IsBoxVisible(BB:TBoundingInfo): boolean;
begin
 Result:=IsBoxVisible(BB.Min,BB.Max);
end;

function TCamera.IsSphereVisible(Center : TPoint3S;Radius : single):boolean;
var I : TFrustumPlanes;
    d : single;
begin
  for I := TFrustumPlanes(0) to TFrustumPlanes(5) do
  begin
    d := FFrustum[I].a * Center.x + FFrustum[I].b * Center.y + FFrustum[I].c * Center.z + FFrustum[I].d;
    // if there is d<=Radius  func will return true only if whole spherer is in frustum
    if (d <= -Radius) then
    begin
      Result:=false;
      Exit;
    end;
  end;
  Result:=true;
end;

end.

