unit RenD3D;

interface

uses Windows, Classes, SysUtils, Renderers, DirectXGraphics, ImgUtils,
  D3DMisc, Engines, WinUtils, ContNRS, GfxWin, GfxTypes, Colors, Images,
  DXGUtils, Textures, Types3D, Timing, DrawBuffers, Utils3D, BaseMgrs,
  TexUtils, GMessages, GfxMessages, DynLights, WindowCls, Lights;

type

  // WARNING - It sem that on newer graphic cards (geForce2 an newer)
  //  IDirect3dDevice8.GetTransform() does not work.So do not
  // use GetXXXXMatrix,WorldPop,WordlPush 

 
  ED3DInit = class(Exception)
  end;
  ED3DError = class(Exception)
  end;


  TD3DRenderer = class(TRenderer)
  private
    FDevice: IDevice;
    FD3D: ID3D;
    FD3DPP: TD3DPP;
    FDevType: TD3DDevType;
    FDevCaps: TD3DCaps8;
    FBehaviour: DWord;
    FDesktopMode: TD3DDisplayMode;
    FModes: TObjectList;
    FAdapter: TAdapterID;
    FWin: TGfxWIn;

    FBegin: boolean;
    FLastTex: ITexture;

    FDefMat: TD3DMaterial8;
    FWorldMat: TD3DMatrix;
    FWorldStack: TStack; //slow
    FStackPos:LongInt;
    FStack:array[0..31]of TD3DMatrix;

    procedure SetD3DParams;
    function DepthExists(Fmt, Depth: TD3DFORMAT): boolean;
    function FullScrModeExists(W, H: Dword; Fmt: TD3DFORMAT): boolean;
    procedure FindFullMode(var W, H: Dword; var Fmt: TD3DFORMAT);
    function GetBitCount(Fmt: TD3DFormat): byte;
    function GetHighestRefresh(W, H: Dword; Fmt: TD3DFORMAT): Dword;
    procedure FindDepth(var Fmt, Depth: TD3DFORMAT);
    function GetState(F: TFeature): TD3DRenderStateType;
    function ModeExists(W, H, B, R: DWord): boolean;
    procedure GetModes;

    procedure BeginScene;
    procedure EndScene;

  protected
    function GetCanDraw: boolean;
    function GetDeviceDesc: string; override;
    function GetDeviceDriver: string; override;
    function GetTexMem: Dword; override;

  public
    constructor Create; override;
    destructor Destroy; override;

    procedure Init; override;
    procedure Done; override;
    function Reset: boolean; override;

    procedure EnumFullScrModes(List: TBaseMgr); override;

    // sets functions

    function SetDWord(Mode: TSetDWord; Value: DWord): boolean; override;
    function SetFloat(Mode: TSetFloat; Value: Single): boolean; override;

    // projection/viewports
    function SetOrtho(Width, Height: LongInt): boolean; override;
    function SetPerspective(Fov, NearP, BackP: single; Aspect: single = 0):
    boolean; override;
    function ConfigView(X, Y, Width, Height: LongInt): boolean; override;
    function GetProjMatrix: TMatrix; override;

    /////////// features
    function Enable(F: TFeature): boolean; override;
    function Disable(F: TFeature): boolean; override;



    /////////// drawing
    function Present: boolean; override;
    function ClearTarget(Color: TARGB): boolean; override;
    function ClearZBuffer(Value: single): boolean; override;
    function DrawBuffer(Buff: TDrawBuffer): boolean; override;
    function FreeDrawBuffer(Buff: TDrawBuffer): boolean; override;
    procedure DrawTriStrip(Points:array of TPoint3S;Coords:array of TTexCoord;
      Col:TARGB=$FFFFFFFF;Tex:TTexture=nil;Norm:PPoint3S=nil);override;
    procedure DrawFace(Points:array of TPoint3S;Coords:array of TTexCoord;
      Cols:array of TARGB;Tex:TTexture=nil;Norm:PPoint3S=nil);override;


    // tex
    procedure CreateDevTex(Tex: TTexture); override;
    procedure DeleteDevTex(Tex: TTexture); override;
    function SetTexture(Tex: TTexture): boolean; override;

    // view
    function ViewIdentity: boolean; override;
    function ViewLookAt(Pos, At, Up: TPoint3S): boolean; override;
    function GetViewMatrix: TMatrix; override;

    // world
    function WorldIdentity: boolean; override;
    function WorldTranslate(X, Y, Z: single): boolean; override;
    function WorldScale(X, Y, Z: single): boolean; override;
    function WorldRotate(X, Y, Z: single): boolean; override;
    function WorldMatrix(Mat: TMatrix): boolean; override;
    function GetWorldMatrix: TMatrix; override;
    function WorldPush: boolean; override;
    function WorldPop: boolean; override;
    function WorldMultiply(Mat: TMatrix): boolean; override;

    // dynamic lights
    function CreateLight(Light: TDynamicLight): boolean; override;
    function EnableLight(Idx: LongInt): boolean; override;
    function DisableLight(Idx: LongInt): boolean; override;


    property D3D: ID3D read FD3D;
    property Device: IDevice read FDevice;
    property Modes: TObjectList read FModes;
    property CanDraw: boolean read GetCanDraw;

  published

  end;





const
  RenName = 'D3DRenderer';
  RenDesc = 'Direct3D8 Renderer';

  sD3DInit = 'Error when initializing Direct3D';
  sD3DErr = 'D3D Error';


var
  D3DResPool: DWord = D3DPOOL_MANAGED;
  TextureMipLevels: LongInt = 5;


implementation

uses GGraphic;

var Mat: TD3DMatrix;
  DispMode: TD3DDisplayMode;

constructor TD3DRenderer.Create;
begin
  inherited Create;
  Name := RenName;
  Description := RenDesc;
  FModes := TObjectList.Create;
  FBegin := false;
  FWorldStack := TStack.Create;

end;

destructor TD3DRenderer.Destroy;
begin
  FWorldStack.Free;
  FModes.Free;
  inherited Destroy;
end;

procedure TD3DRenderer.GetModes;
var NumModes, I: LongInt;
  D3DMode: TD3DDisplayMode;
  NewMode: TDisplayMode;

begin
  if FD3D = nil then Exit;

  FModes.Clear;
  NumModes := FD3D.GetAdapterModeCount(0);
  for I := 0 to NumModes - 1 do
  begin
    FD3D.EnumAdapterModes(D3DADAPTER_DEFAULT, I, D3DMode);
    NewMode := TDisplayMode.Create;
    NewMode.Width := D3DMode.Width;
    NewMode.Height := D3DMode.Height;
    NewMode.Format := D3DMode.Format;
    NewMode.Refresh := D3DMode.RefreshRate;
    FModes.Add(NewMode);
  end;

end;

procedure TD3DRenderer.Init;
var HRes: HResult;

begin

  try
    FD3D := Direct3DCreate8(D3D_SDK_VERSION);
    FD3D.GetAdapterDisplayMode(D3DADAPTER_DEFAULT, FDesktopMode);
    FD3D.GetAdapterIdentifier(D3DADAPTER_DEFAULT, 0, FAdapter);

    SetD3DParams;


    HRes := FD3D.CreateDevice(D3DADAPTER_DEFAULT, FDevType,
      FD3DPP.hDeviceWindow, FBehaviour, FD3DPP, FDevice);
    if HRes <> D3D_OK then
      raise ED3DInit.Create(sD3DInit + ' :' + DXGErrorString(HRes));


  except
    raise ED3DInit.Create(sD3DInit);
  end;

  FDevice.Clear(0, nil, D3DCLEAR_TARGET, $FF303030, 1, 0);

  DXGutils.SetIdentityMatrix(Mat);
  DXGutils.SetIdentityMatrix(FWorldMat);
  FDevice.SetTransform(D3DTS_WORLD, Mat);
  FDevice.SetTransform(D3DTS_VIEW, Mat);
  FWorldMat:=TD3DMatrix(IdentityMatrix);

  FDevice.GetDisplayMode(DispMode);
  UsedParams.RefreshRate := DispMode.RefreshRate;

  // remove
  FDevice.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
  FDefMat := DXGUtils.InitMaterial(1, 1, 1, 1);
  FDefMat.Power := 100;
  FDevice.SetMaterial(FDefMat);


  Present;

  inherited Init;
end;

function TD3DRenderer.Reset: boolean;
begin
  Result := false;
  if not Inited then Exit;
  if FBegin then EndScene;
  SetD3DParams;
  Result := not Failed(FDevice.Reset(FD3DPP));
  FLastTex := nil;

  FDevice.GetDisplayMode(DispMode);
  UsedParams.RefreshRate := DispMode.RefreshRate;

  // remove after implementation

  FDevice.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
  FDevice.SetMaterial(FDefMat);
 //

  inherited Reset;
end;

procedure TD3DRenderer.Done;
begin
  inherited Done;
  FModes.Clear;
  FDevice := nil;
  FD3D := nil;

end;


function TD3DRenderer.GetBitCount(Fmt: TD3DFormat): byte;
begin

  case Fmt of
    D3DFMT_A8R8G8B8,
      D3DFMT_X8R8G8B8,
      D3DFMT_D32: Result := 32;
    D3DFMT_R5G6B5,
      D3DFMT_X1R5G5B5,
      D3DFMT_D16,
      D3DFMT_D16_LOCKABLE: Result := 16;
    D3DFMT_R8G8B8,
      D3DFMT_D24X8, //it is 32bit but we count only Z-Buffer
      D3DFMT_D24S8: Result := 24;

  else

    raise ED3DError.Create('Bad D3D Format');
  end;

end;

function TD3DRenderer.DepthExists(Fmt, Depth: TD3DFORMAT): boolean;
begin
  Result := false;

  if (SUCCEEDED(FD3D.CheckDeviceFormat(D3DADAPTER_DEFAULT, FDevType,
    Fmt, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, Depth))) then
    if (SUCCEEDED(FD3D.CheckDepthStencilMatch(D3DADAPTER_DEFAULT, FDevType,
      Fmt, Fmt, Depth))) then Result := true;


end;

procedure TD3DRenderer.FindDepth(var Fmt, Depth: TD3DFORMAT);
var Found: boolean;
begin
  Found := false;
  while not Found do
    case GetBitCount(Depth) of
      16:
        begin
          if DepthExists(Fmt, Depth) then
            Found := true
          else
          begin
            raise ED3DError.Create('16 bit Z Buffer format not found.' + #10 +
              'Please disable your video card with hammer.');
          end;

        end;
      24:
        begin
          if DepthExists(Fmt, Depth) then
            Found := true
          else
            case Depth of
              D3DFMT_D24S8: Depth := D3DFMT_D24X8;
              D3DFMT_D24X8: Depth := D3DFMT_D16;
            end;
        end;
      32:
        begin
          if DepthExists(Fmt, Depth) then
            Found := true
          else
            Depth := D3DFMT_D24S8;

        end;
    end;

end;

function TD3DRenderer.FullScrModeExists(W, H: Dword; Fmt: TD3DFORMAT): boolean;
var I: LongInt;
  Chk: boolean;
begin
  Result := false;
  if FModes.Count = 0 then Exit;

  for I := 0 to FModes.Count - 1 do
  begin
    with TDisplayMode(FModes[I]) do
      Chk := (Width = W) and (Height = H) and (Format = Fmt);
    if Chk then
    begin
      Result := true;
      Break;
    end;
  end;
end;

function TD3DRenderer.ModeExists(W, H, B, R: DWord): boolean;
var I: LongInt;
  Chk: boolean;
begin
  Result := false;
  if FModes.Count = 0 then Exit;

  for I := 0 to FModes.Count - 1 do
  begin
    with TDisplayMode(FModes[I]) do
      Chk := (Width = W) and (Height = H) and (Format = B) and (Refresh = R);
    if Chk then
    begin
      Result := true;
      Break;
    end;
  end;
end;

procedure TD3DRenderer.FindFullMode(var W, H: Dword; var Fmt: TD3DFORMAT);
var I: LongInt;
  Found: boolean;
begin
  Found := false;
  while not Found do
  begin
    if FullScrModeExists(W, H, Fmt) then
      Found := true
    else
    begin

      for I := 0 to FModes.Count - 1 do
        with TDisplayMode(FModes[I]) do
        begin

          if (Width = W) and (Height = H) and (Format <> Fmt) then

          begin
            if GetBitCount(Fmt) = 32 then
              Fmt := D3DFMT_R5G6B5;
            Break;
          end;

          raise ED3DError.Create('Display Mode not supported');
        end;
    end;
  end;
end;

function TD3DRenderer.GetHighestRefresh(W, H: Dword; Fmt: TD3DFORMAT): Dword;
var Rate, I: Dword;

begin
  Rate := D3DPRESENT_RATE_DEFAULT;
  for I := 0 to FModes.Count - 1 do
  begin
    with TDisplayMode(FModes[I]) do
      if (Width = W) and (Height = H) and (Format = Fmt) then
        if Refresh > Rate then Rate := Refresh;


  end;

// on my monitor 800x600 @ 75 Hz does HuhlyHuhly
// remove it

  if (W = 800) and (Rate = 75) then Rate := 72;


  Result := Rate;

end;

procedure TD3DRenderer.SetD3DParams;
var Format, Depth: TD3DFormat;
  Sync, Rate, W, H: Dword;
  R: Windows.TRect;
begin
  ZeroMemory(@FD3DPP, SizeOf(FD3DPP));
  CheckParams;

  GetModes;
  FDevType := Iff(Params.UseHardWare, D3DDEVTYPE_HAL, D3DDEVTYPE_REF);
  UsedParams.UseHardWare := Params.UseHardWare;

  FD3D.GetDeviceCaps(D3DADAPTER_DEFAULT, FDevType, FDevCaps);

  FD3DPP.hDeviceWindow := Params.Handle;
  FD3DPP.SwapEffect := D3DSWAPEFFECT_DISCARD;
  FD3DPP.BackBufferCount := Params.BackBuffers;
  FD3DPP.EnableAutoDepthStencil := Params.UseDepth;

  UsedParams.Handle := Params.Handle;
  UsedParams.UseDepth := Params.UseDepth;
  UsedParams.BackBuffers := Params.BackBuffers;

 // check hardware transforming and lighting support
  FBehaviour := D3DCREATE_SOFTWARE_VERTEXPROCESSING;
  if Params.UseHTL then
  begin
    if (FDevCaps.DevCaps and D3DDEVCAPS_HWTRANSFORMANDLIGHT) <> 0 then
    begin
      FBehaviour := D3DCREATE_HARDWARE_VERTEXPROCESSING;
      UsedParams.UseHTL := true;
    end;

    if (FDevCaps.DevCaps and D3DDEVCAPS_PUREDEVICE) <> 0 then
      FBehaviour := FBehaviour or D3DCREATE_PUREDEVICE;
  end;


 // check windowed rendering support
  FD3DPP.Windowed := false;
  if not Params.FullScreen then
  begin
    if (FDevCaps.Caps2 and D3DCAPS2_CANRENDERWINDOWED) <> 0 then
      FD3DPP.Windowed := true;
  end;
  UsedParams.FullScreen := not FD3DPP.Windowed;

  FWin := TGfxWin(GetWin(Params.Handle));
  if FWin = nil then
  begin
    FWin := TGfxWin.Create();
    FWin.SetHandle(Params.Handle);
  end;

  if (FWin <> nil) then
  begin
    if UsedParams.FullScreen then
    begin
      FWin.CanSetMode := false;
      FWin.SetFullScreen(Params.Width, Params.Height, Params.ColorBits);
    end
    else
      FWin.SetWindowed;
    FWin.CanSetMode := true;
  end;

 //seting parameters that differs in fullscr/window
  if UsedParams.FullScreen then
  begin
    if not Params.UseVSync then
      Sync := D3DPRESENT_INTERVAL_IMMEDIATE
    else
      Sync := D3DPRESENT_INTERVAL_ONE;


    W := Params.Width;
    H := Params.Height;
    case Params.ColorBits of
      16: Format := D3DFMT_R5G6B5;
      32: Format := D3DFMT_X8R8G8B8;
    else
      Format := D3DFMT_X8R8G8B8;
    end;

    FindFullMode(W, H, Format);
    UsedParams.ColorBits := GetBitCount(Format);

    Rate := GetHighestRefresh(W, H, Format);
    if Params.RefreshRate > 0 then
      if ModeExists(W, H, Format, Params.RefreshRate) then
        Rate := Params.RefreshRate
      else
        Rate := D3DPRESENT_RATE_DEFAULT;

  end

  else
  begin
    Format := FDesktopMode.Format;
    Rate := D3DPRESENT_RATE_DEFAULT;
    Sync := D3DPRESENT_INTERVAL_DEFAULT;

    GetClientRect(UsedParams.Handle, R);
    W := R.Right;
    H := R.Bottom;

    UsedParams.ColorBits := GetBitCount(FDesktopMode.Format);
  //  UsedParams.RefreshRate:=FDeskTopMode.RefreshRate;

  end;

  FD3DPP.FullScreen_PresentationInterval := Sync;
  FD3DPP.BackBufferFormat := Format;
  FD3DPP.BackBufferWidth := W;
  FD3DPP.BackBufferHeight := H;
  FD3DPP.FullScreen_RefreshRateInHz := Rate;

  UsedParams.UseVSync := Params.UseVSync;

  UsedParams.Width := W;
  UsedParams.Height := H;

  // check multisample support
  FD3DPP.MultiSampleType := Params.MultiSamples;
  while not Succeeded(FD3D.CheckDeviceMultiSampleType(D3DADAPTER_DEFAULT,
    FDevType,
    Format, not UsedParams.FullScreen, FD3DPP.MultiSampleType)) do
    Dec(FD3DPP.MultiSampleType);
  UsedParams.MultiSamples := FD3DPP.MultiSampleType;


  // Depth Buffer Settings
  UsedParams.DepthBits := 0;
  UsedParams.StencilBits := 0;
  if UsedParams.UseDepth then
  begin
    if Params.DepthBits < 16 then
      Params.DepthBits := 16;
    case Params.DepthBits of
      16: Depth := D3DFMT_D16;
      24: Depth := D3DFMT_D24S8;
      32: Depth := D3DFMT_D32;
    end;

    if (Params.StencilBits > 0) then
      Depth := D3DFMT_D24S8;

    FindDepth(Format, Depth);
    FD3DPP.AutoDepthStencilFormat := Depth;
    UsedParams.DepthBits := GetBitCount(Depth);
    UsedParams.StencilBits := Iff(Depth = D3DFMT_D24S8, 8, 0);

  end;


  if UsedParams.FullScreen then
    Messages.SendMessage(MSG_GFX_TOFULLSCREEN, Self)
  else
    Messages.SendMessage(MSG_GFX_TOWINDOW, Self)

end;

procedure TD3DRenderer.EnumFullScrModes(List: TBaseMgr);
var I: LongInt;
  DM: TDisplayMode;
begin
  if not Inited then
  begin
    FD3D := Direct3DCreate8(D3D_SDK_VERSION);
    GetModes;
    FD3D := nil;
  end;

  for I := 0 to FModes.Count - 1 do
  begin
    DM := TDisplayMode(FModes[I]);
    List.Add(TFullScreenMode.Create(DM.Width, DM.Height, GetBitCount(DM.Format),
      DM.Refresh));
  end;

end;


function TD3DRenderer.GetDeviceDesc: string;
begin
  Result := FAdapter.Description;
end;

function TD3DRenderer.GetDeviceDriver: string;
begin
  Result := FAdapter.Driver;
end;

function TD3DRenderer.GetTexMem: Dword;
begin
  Result := 0;
  if not Inited then Exit;
  Result := FDevice.GetAvailableTextureMem;
end;




function TD3DRenderer.GetState(F: TFeature): TD3DRenderStateType;
begin
  Result := 0;
  case F of
    gfLighting: Result := D3DRS_LIGHTING;
    gfAlphaBlend: Result := D3DRS_ALPHABLENDENABLE;
    gfStencil: Result := D3DRS_STENCILENABLE;
    gfFog: Result := D3DRS_FOGENABLE;
    gfDither: Result := D3DRS_DITHERENABLE;
    gfAlphaTest: Result := D3DRS_ALPHATESTENABLE;
    gfClipPlanes: Result := D3DRS_CLIPPLANEENABLE;
    gfVertexColors:
      begin
        Result := D3DRS_COLORVERTEX;
        FDevice.SetRenderState(D3DRS_AMBIENTMATERIALSOURCE, D3DMCS_COLOR1);
        /// FDevice.SetRenderState(D3DRS_EMISSIVEMATERIALSOURCE, D3DMCS_COLOR1);

      end;
    gfZBuffer: Result := D3DRS_ZENABLE;
    gfZWrite: Result := D3DRS_ZWRITEENABLE;
    gfNormalizeNormals: Result := D3DRS_NORMALIZENORMALS;
  end;

end;


function TD3DRenderer.Enable(F: TFeature): boolean;
var State: TD3DRenderStateType;
begin
  Result := false;
  if not Inited then Exit;

  State := GetState(F);
  Result := not Failed(FDevice.SetRenderState(State, 1));
  Graphic.FeatureSet[F] := true;
end;

function TD3DRenderer.Disable(F: TFeature): boolean;
var State: TD3DRenderStateType;
begin
  Result := false;
  if not Inited then Exit;

  State := GetState(F);
  Result := not Failed(FDevice.SetRenderState(State, 0));
  Graphic.FeatureSet[F] := false;
end;


function TD3DRenderer.SetOrtho(Width, Height: LongInt): boolean;
begin
  Result := false;
  if not Inited then Exit;

  CreateOrthoMatrix(Mat, Width, Height);
  FDevice.SetTransform(D3DTS_PROJECTION, Mat);

  Result := true;
end;

function TD3DRenderer.SetPerspective(Fov, NearP, BackP: single; Aspect: single):
boolean;
begin
  Result := false;
  if not Inited then Exit;
  if Aspect < 0.01 then
    Aspect := AspectRatio;
  SetProjectionMatrix(Mat, DegToRad(Fov), 1 / Aspect, NearP, BackP);
  FDevice.SetTransform(D3DTS_PROJECTION, Mat);
  Result := true;
end;

function TD3DRenderer.GetProjMatrix: TMatrix;
begin
  if not Inited then Exit;
  FDevice.GetTransform(D3DTS_PROJECTION, Mat);
  Result := TMatrix(Mat);
end;

function TD3DRenderer.ConfigView(X, Y, Width, Height: LongInt): boolean;
var View: TD3DView;
begin
  Result := false;
  if not Inited then Exit;
  View := D3DMisc.ViewPort(X, Y, Width, Height, 0, 1.0);
  Result := not Failed(FDevice.SetViewport(View));

end;

function TD3DRenderer.SetDWord(Mode: TSetDWord; Value: DWord): boolean;
var AVal: DWord;
begin
  Result := false;
  if not Inited then Exit;
  AVal := 0;
  try
    case Mode of
      sdTexFilter:
        begin
          case Value of
            TEXF_NEAREST: AVal := D3DTEXF_POINT;
            TEXF_BILINEAR,
              TEXF_TRILINEAR: AVal := D3DTEXF_LINEAR;
          end;
          FDevice.SetTextureStageState(0, D3DTSS_MINFILTER, AVal);
          FDevice.SetTextureStageState(0, D3DTSS_MAGFILTER, AVal);

          case Value of
            TEXF_NEAREST,
              TEXF_BILINEAR: AVal := D3DTEXF_POINT;
            TEXF_TRILINEAR: AVal := D3DTEXF_LINEAR;
          end;

          if TextureUseMipMaps then
            FDevice.SetTextureStageState(0, D3DTSS_MIPFILTER, AVAl);

        end;
      sdAmbient: FDevice.SetRenderState(D3DRS_AMBIENT, Value);
      sdFillMode: FDevice.SetRenderState(D3DRS_FILLMODE, Value + 1);
      sdShadeMode: FDevice.SetRenderState(D3DRS_SHADEMODE, Value + 1);
      sdCullMode: FDevice.SetRenderState(D3DRS_CULLMODE, Value + 1);
      sdFogMode:
        begin
          case Value of
            FOG_LINEAR: AVal := D3DFOG_LINEAR;
            FOG_EXP: AVal := D3DFOG_EXP;
            FOG_EXP2: AVal := D3DFOG_EXP2;
          end;
          FDevice.SetRenderState(D3DRS_FOGTABLEMODE, AVal);
        end;
      sdFogColor: FDevice.SetRenderState(D3DRS_FOGCOLOR, Value);
      sdZFunc: FDevice.SetRenderState(D3DRS_ZFUNC, Value + 1);
      sdAlphaFunc: FDevice.SetRenderState(D3DRS_ALPHAFUNC, Value + 1);
      sdAlphaRef: FDevice.SetRenderState(D3DRS_ALPHAREF, Value);
      sdStencilFunc: FDevice.SetRenderState(D3DRS_STENCILFUNC, Value + 1);
      sdStencilRef: FDevice.SetRenderState(D3DRS_STENCILREF, Value);
      sdStencilMask: FDevice.SetRenderState(D3DRS_STENCILMASK, Value);
      sdSrcBlend: FDevice.SetRenderState(D3DRS_SRCBLEND, Value + 1);
      sdDstBlend: FDevice.SetRenderState(D3DRS_DESTBLEND, Value + 1);
      sdTexAdrX: FDevice.SetTextureStageState(0, D3DTSS_ADDRESSU, Value + 1);
      sdTexAdrY: FDevice.SetTextureStageState(0, D3DTSS_ADDRESSV, Value + 1);

    end;
  except
    raise ED3DError.Create(sD3DErr +
      ' - TD3DRenderer.SetDWord - Device = nil or Invalid Value');
  end;
  Graphic.DWordSet[Mode] := Value;
  Result := true;

end;


function TD3DRenderer.SetFloat(Mode: TSetFloat; Value: Single): boolean;
begin
  Result := false;
  if not Inited then Exit;
  case Mode of
    sfFogStart: FDevice.SetRenderState(D3DRS_FOGSTART, FloatToDWord(Value));
    sfFogEnd: FDevice.SetRenderState(D3DRS_FOGEND, FloatToDWord(Value));
    sfFogDensity: FDevice.SetRenderState(D3DRS_FOGDENSITY, FloatToDWord(Value));
  end;
  Graphic.FloatSet[Mode] := Value;
end;



procedure TD3DRenderer.CreateDevTex(Tex: TTexture);
var DXT: ITexture;
  Fmt: DWord;
  Temp: TImage;
  TexRec: PD3DTex;
  R: TD3DLocked_Rect;
  Desc: TD3DSurface_Desc;
  Surf: ISurf;
  I, Levels: LongInt;
begin
  if not Inited then Exit;
  DeleteDevTex(Tex);

  Fmt := D3DFMT_A8R8G8B8;
  case Tex.DevTexBits of
    16:
      if Tex.Image.HasAlpha then
        Fmt := D3DFMT_A4R4G4B4
      else
        Fmt := D3DFMT_A1R5G5B5;
    32:
      Fmt := D3DFMT_A8R8G8B8
  end;

  if Tex.UseMipMaps then
    Levels := 0 //TextureMipLevels
  else
    Levels := 1;

  FDevice.CreateTexture(Tex.Width, Tex.Height,
    Levels, 0, Fmt, D3DResPool, DXT);

  Temp := Tex.Image.Clone;
  Mirror(Temp, false, true);

  Temp.BitCount := 32;
  Temp.BilinearResize := false;

  if Tex.ColorKeyEnable then
    SetColorKey32(Temp.Width, Temp.Height, Tex.ColorKey, Temp.Bits);

  Levels := DXT.GetLevelCount;

  for I := 0 to Levels - 1 do
  begin
    DXT.GetSurfaceLevel(I, Surf);
    Surf.LockRect(R, nil, 0);
    Surf.GetDesc(Desc);
    Temp.Resize(Desc.Width, Desc.Height);

    case Tex.DevTexBits of
      32: Move(Temp.Bits^, R.pBits^, Temp.Size);
      16:
        if Tex.Image.HasAlpha then
          ConvertARGB8ToARGB4(Desc.Width, Desc.Height, Temp.Bits, R.pBits)
        else
          ConvertARGB8ToA1RGB5(Desc.Width, Desc.Height, Temp.Bits, R.pBits);
    end;


    Surf.UnlockRect;
    Surf := nil;
  end;


  GetMem(TexRec, SizeOf(TD3DTex));
  ZeroMemory(TexRec, SizeOf(TD3DTex));
  TexRec^.ITex := DXT;

  Tex.DevTex := TexRec;
  Temp.Free;

  DXT := nil;

end;

procedure TD3DRenderer.DeleteDevTex(Tex: TTexture);
begin
  if Tex.DevTex <> nil then
  begin
    TD3DTex(Tex.DevTex^).ITex := nil;
    FreeMem(Tex.DevTex);
    Tex.DevTex := nil;
  end;


end;

function TD3DRenderer.SetTexture(Tex: TTexture): boolean;
var TR: TD3DTex;
begin
  Result := false;
  if not Inited then Exit;

  if not ((Tex = nil) and (FLastTex = nil)) then
  begin

    if (Tex = nil) and (FLastTex <> nil) then
    begin
      FDevice.SetTexture(0, nil);
      FLastTex := nil;
    end
    else
    begin
      if Tex.DevTex = nil then
      begin
        FDevice.SetTexture(0, nil);
        FLastTex := nil;
      end
      else
      begin
        TR := TD3DTex(Tex.DevTex^);
        if TR.ITex <> FLastTex then
        begin
          FDevice.SetTexture(0, TR.ITex);
          SetDWord(sdTexFilter, Tex.Filter);
          FLastTex := TR.ITex;
        end;
      end;
    end;
  end;

  Result := true;
end;

function TD3DRenderer.FreeDrawBuffer(Buff: TDrawBuffer): boolean;
begin
  Result := false;
  if Buff.RenderData = nil then Exit;
  FreeMem(TD3DBuffer2(Buff.RenderData^).Vertices);
  FreeMem(Buff.RenderData);
  Buff.RenderData := nil;
  Result := true;
end;


function TD3DRenderer.GetCanDraw: boolean;
var Hr: HResult;
begin
  Result := false;
  if not Inited then Exit;
  Hr := FDevice.TestCooperativeLevel;
  if (hr = D3DERR_DEVICELOST) then Exit;
  if (hr = D3DERR_DEVICENOTRESET) then Reset;
  Result := true;
end;

function TD3DRenderer.ViewIdentity: boolean;
begin
  Result := false;
  if not Inited then Exit;
  SetIdentityMatrix(Mat);
  Result := not Failed(FDevice.SetTransform(D3DTS_VIEW, Mat));
  FWorldMat:=Mat;
end;

function TD3DRenderer.GetViewMatrix: TMatrix;
begin
  if not Inited then Exit;
  Result:=TMatrix(FWorldMAt);
end;


function TD3DRenderer.ViewLookAt(Pos, At, Up: TPoint3S): boolean;
begin
  Result := false;
  if not Inited then Exit;
  SetIdentityMatrix(Mat);
  SetViewMatrix(Mat, TD3DVector(Pos), TD3DVector(At), TD3DVector(Up));
  Result := not Failed(FDevice.SetTransform(D3DTS_VIEW, Mat));
  FWorldMat:=Mat;
end;

function TD3DRenderer.WorldIdentity: boolean;
begin
  Result := false;
  if not Inited then Exit;
  Camera.UpdateView;
  Result := true;
end;

function TD3DRenderer.WorldTranslate(X, Y, Z: single): boolean;
begin
  Result := false;
  if not Inited then Exit;
  SetTranslateMatrix(Mat, X, Y, Z);
  Result := not Failed(FDevice.MultiplyTransform(D3DTS_VIEW, Mat));
  FWorldMat:=TD3DMatrix(MatrixMultiply(TMatrix(FWorldMat),TMatrix(Mat)));
end;

function TD3DRenderer.WorldScale(X, Y, Z: single): boolean;
begin
  Result := false;
  if not Inited then Exit;
  SetScaleMatrix(Mat, X, Y, Z);
  Result := not Failed(FDevice.MultiplyTransform(D3DTS_VIEW, Mat));
  FWorldMat:=TD3DMatrix(MatrixMultiply(TMatrix(FWorldMat),TMatrix(Mat)));
end;

function TD3DRenderer.WorldRotate(X, Y, Z: single): boolean;
begin
  Result := false;
  if not Inited then Exit;
  SetRotateZMatrix(Mat, DegToRad(Z));
  FDevice.MultiplyTransform(D3DTS_VIEW, Mat);
  FWorldMat:=TD3DMatrix(MatrixMultiply(TMatrix(FWorldMat),TMatrix(Mat)));
  SetRotateXMatrix(Mat, DegToRad(X));
  FDevice.MultiplyTransform(D3DTS_VIEW, Mat);
  FWorldMat:=TD3DMatrix(MatrixMultiply(TMatrix(FWorldMat),TMatrix(Mat)));
  SetRotateYMatrix(Mat, DegToRad(Y));
  FDevice.MultiplyTransform(D3DTS_VIEW, Mat);
  FWorldMat:=TD3DMatrix(MatrixMultiply(TMatrix(FWorldMat),TMatrix(Mat)));
  Result := true;
end;

function TD3DRenderer.WorldMatrix(Mat: TMatrix): boolean;
begin
  Result := false;
  if not Inited then Exit;
  FDevice.SetTransform(D3DTS_VIEW, TD3DMatrix(Mat));
  FWorldMat:=TD3DMatrix(Mat);
  Result := true;

end;

function TD3DRenderer.GetWorldMatrix: TMatrix;
begin
  if not Inited then Exit;
  Result:=TMatrix(FWorldMAt);
end;

function TD3DRenderer.WorldPush: boolean;
begin
// Result := false;
//  if not Inited then Exit;

  FStack[FStackPos]:=FWorldMat;
  Inc(FStackPos);
  Result := true;
end;

function TD3DRenderer.WorldPop: boolean;
begin
  Dec(FStackPos);
  FDevice.SetTransform(D3DTS_VIEW, FStack[FStackPos]);
  FWorldMat:=FStack[FStackPos];
  Result := true;
end;

function TD3DRenderer.WorldMultiply(Mat: TMatrix): boolean;
begin
  //Result := false;
  //if not Inited then Exit;
  FDevice.MultiplyTransform(D3DTS_VIEW, TD3DMatrix(Mat));
  FWorldMat:=TD3DMatrix(MatrixMultiply(TMatrix(FWorldMat),Mat));
  Result :=true;
end;

function TD3DRenderer.CreateLight(Light: TDynamicLight): boolean;
var D3DLight: TD3DLight8;
begin
  Result := false;
  if not Inited then Exit;
  ZeroMemory(@D3DLight, SizeOf(D3DLight));
  case Light.LightType of
    ltPoint: D3DLight._Type := D3DLIGHT_POINT;
    ltDirectional: D3DLight._Type := D3DLIGHT_DIRECTIONAL;
    ltSpot: D3DLight._Type := D3DLIGHT_SPOT;
  end;
  D3DLight.Diffuse := LCToD3DLC(Light.DiffuseColor);
  D3DLight.Specular := LCToD3DLC(Light.SpecularColor);
  D3DLight.Ambient := LCToD3DLC(Light.AmbientColor);
  D3DLight.Position := Point3SToD3DV(Light.Position);
  D3DLight.Direction := Point3SToD3DV(Light.AngleVector);
  D3DLight.Range := Light.Range;
  D3DLight.Falloff := Light.SpotFallOff;
  D3DLight.Attenuation0 := Light.ConstantAttenuation;
  D3DLight.Attenuation1 := Light.LinearAttenuation;
  D3DLight.Attenuation2 := Light.QuadraticAttenuation;
  D3DLight.Theta := DegToRad(Light.SpotInnerConeAngle);
  D3DLight.Phi := DegToRad(Light.SpotOuterConeAngle);

  Result := not Failed(FDevice.SetLight(Light.Index, D3DLight));



end;

function TD3DRenderer.EnableLight(Idx: LongInt): boolean;
begin
  Result := false;
  if not Inited then Exit;
  Result := not Failed(FDevice.LightEnable(Idx, true));
end;

function TD3DRenderer.DisableLight(Idx: LongInt): boolean;
begin
  Result := false;
  if not Inited then Exit;
  Result := not Failed(FDevice.LightEnable(Idx, false));
end;



//////////////////////////// DRAWING

procedure TD3DRenderer.BeginScene;
begin
  if FDevice = nil then Exit;
  if Failed(FDevice.BeginScene) then
    raise ED3DError.Create(sD3DErr + ' : BeginScene failed');
  FBegin := true;

end;

procedure TD3DRenderer.EndScene;
begin
  if FDevice = nil then Exit;
  if Failed(FDevice.EndScene) then
    raise ED3DError.Create(sD3DErr + ' : EndScene failed');
  FBegin := false;

end;

function TD3DRenderer.Present: boolean;
begin
  Result := false;
  if not Inited then Exit;
  if not CanDraw then Exit;
  if FBegin then EndScene;
  if not Failed(FDevice.Present(nil, nil, 0, nil)) then
    Result := true
  else
    raise ED3DError.Create(sD3DErr + ' : Error during presentation');

end;



function TD3DRenderer.ClearTarget(Color: TARGB): boolean;
begin
  Result := false;
  if not Inited then Exit;
  if not CanDraw then Exit;
  if not Failed(FDevice.Clear(0, nil, D3DCLEAR_TARGET, Color, 1, 0)) then
    Result := true

end;

function TD3DRenderer.ClearZBuffer(Value: single): boolean;
begin
  Result := false;
  if not Inited then Exit;
  if not CanDraw then Exit;
  if not Failed(FDevice.Clear(0, nil, D3DCLEAR_ZBUFFER, 0, Value, 0)) then
    Result := true

end;



function TD3DRenderer.DrawBuffer(Buff: TDrawBuffer): boolean;
var FVF: DWord;
  I, VCount: DWord;
  Prim: DWord;
  Stride, Mode: DWord;

  Pos: TPoint3S;
  Normal: TPoint3S;
  Col1: TARGB;
  Col2: TARGB;
  Coord1: TTexCoord;
  Coord2: TTexCoord;
  RD2: TD3DBuffer2;
  PRD2: PD3DBuffer2;

  ArrPNC1T1: array of TVPNC1T1;
  ArrPC1T1: array of TVPC1T1;
  ArrPC1: array of TVPC1;
  ArrPC2T1: array of TVPC2T1;
  Arr: Pointer;

begin
  Result := false;
  if not Inited then Exit;
  if Buff = nil then Exit;
  if not CanDraw then Exit;
  if not FBegin then
    BeginScene;



  Prim := Buff.Primitives;

  if ufTexture in Buff.Used then
    SetTexture(Buff.Texture);

  if Buff.RenderData = nil then
  begin

    Stride := 0;
    Mode := 0;

    VCount := Buff.Points.Count;
    if VCount <= 0 then Exit;

    if (Prim > VCount) or (Prim < 1) then
      Prim := VCount;

    case Buff.Mode of
      dmPoints: Mode := D3DPT_POINTLIST;
      dmLineList: Mode := D3DPT_LINELIST;
      dmLineStrip: Mode := D3DPT_LINESTRIP;
      dmTriangleList: Mode := D3DPT_TRIANGLELIST;
      dmTriangleStrip: Mode := D3DPT_TRIANGLESTRIP;
      dmTriangleFan: Mode := D3DPT_TRIANGLEFAN;
    end;

    FVF := D3DFVF_XYZ;
    if ufColors1 in Buff.Used then FVF := FVF or D3DFVF_DIFFUSE;
    if ufColors2 in Buff.Used then FVF := FVF or D3DFVF_SPECULAR;
    if ufCoords1 in Buff.Used then FVF := FVF or D3DFVF_TEX1;
    if ufCoords2 in Buff.Used then FVF := FVF or D3DFVF_TEX2;
    if ufNormals in Buff.Used then FVF := FVF or D3DFVF_NORMAL;
    if ufAll in Buff.Used then FVF := FVF_PNC2T2;

    ArrPNC1T1 := nil;
    ArrPC1T1 := nil;
    Arr := nil;
    case FVF of
      FVF_PNC1T1:
        begin
          SetLength(ArrPNC1T1, VCount);
          Arr := ArrPNC1T1;
          Stride := SizeOf(TVPNC1T1);
        end;
      FVF_PC1T1:
        begin
          SetLength(ArrPC1T1, VCount);
          Arr := ArrPC1T1;
          Stride := SizeOf(TVPC1T1);
        end;
      FVF_PC1:
        begin
          SetLength(ArrPC1, VCount);
          Arr := ArrPC1;
          Stride := SizeOf(TVPC1);
        end;
      FVF_PC2T1:
        begin
          SetLength(ArrPC2T1, VCount);
          Arr := ArrPC2T1;
          Stride := SizeOf(TVPC2T1);
        end;
    end;

    Normal := Point3S(0, 0, 1);
    Col1 := $FFFFFFFF;
    Col2 := $FFFFFFFF;
    Coord1 := Coord(0, 0);
    Coord2 := Coord(0, 0);

    for I := 0 to VCount - 1 do
    begin
      Pos := TPoint3S(Buff.Points[I]^);
      if Buff.Normals[I] <> nil then
        Normal := TPoint3S(Buff.Normals[I]^);

      if Buff.Colors1[I] <> nil then
        Col1 := TARGB(Buff.Colors1[I]^);

      if Buff.Colors2[I] <> nil then
        Col2 := TARGB(Buff.Colors2[I]^);

      if Buff.Coords1[I] <> nil then
        Coord1 := TTexCoord(Buff.Coords1[I]^);

      if Buff.Coords2[I] <> nil then
        Coord2 := TTexCoord(Buff.Coords2[I]^);


      case FVF of
        FVF_PNC1T1:
          begin
            ArrPNC1T1[I].Pos := Pos;
            ArrPNC1T1[I].Normal := Normal;
            ArrPNC1T1[I].Color1 := Col1;
            ArrPNC1T1[I].Coord1 := Coord1;
          end;
        FVF_PC1T1:
          begin
            ArrPC1T1[I].Pos := Pos;
            ArrPC1T1[I].Color1 := Col1;
            ArrPC1T1[I].Coord1 := Coord1;
          end;
        FVF_PC1:
          begin
            ArrPC1[I].Pos := Pos;
            ArrPC1[I].Color1 := Col1;
          end;
        FVF_PC2T1:
          begin
            ArrPC2T1[I].Pos := Pos;
            ArrPC2T1[I].Color1 := Col1;
            ArrPC2T1[I].Color2 := Col2;
            ArrPC2T1[I].Coord1 := Coord1;
          end;


      end;

    end;


    FDevice.SetVertexShader(FVF);
    if Arr <> nil then
      FDevice.DrawPrimitiveUP(Mode, Prim, Arr, Stride);



    if ufDeviceBuff in Buff.Used then
    begin

      GetMem(PRD2, SizeOf(RD2));
      ZeroMemory(PRD2, SizeOf(RD2));
      GetMem(PRD2^.Vertices, Stride * VCount);

      PRD2^.Mode := Mode;
      PRD2^.Stride := Stride;
      PRD2^.FVF := FVF;
      System.Move(Arr^, PRD2^.Vertices^, Stride * VCount);
      Buff.RenderData := PRD2;


    end;

  end
  else
  begin

    RD2 := TD3DBuffer2(Buff.RenderData^);
    FDevice.SetVertexShader(RD2.FVF);
    if RD2.Vertices <> nil then
      FDevice.DrawPrimitiveUP(RD2.Mode, Prim, RD2.Vertices, RD2.Stride);



  end;

  Inc(PrimsDrawn, Buff.Primitives);

  ArrPNC1T1 := nil;
  ArrPC1T1 := nil;
  ArrPC1 := nil;

  Result := true;
end;

procedure TD3DRenderer.DrawTriStrip(Points:array of TPoint3S;Coords:array of TTexCoord;
      Col:TARGB=$FFFFFFFF;Tex:TTexture=nil;Norm:PPoint3S=nil);
var
 Verts: array[0..3] of TVPNC1T1;
 I:LongInt;

begin
 if High(Points)<>3 then
  raise ED3DError.Create('Error in TD3DRenderer.DrawTriStrip  -  parameter "Ponits" must have 4 values.');

 //if not CanDraw then Exit;
  if not FBegin then
    BeginScene;

 for I:=0 to 3 do
  begin
   Verts[I].Pos:=Points[I];
   Verts[I].Color1:=Col;
   if Norm<>nil then
   Verts[I].Normal:=Norm^
   else
   Verts[I].Normal:=Point3S(0,0,1);
   if I<High(Coords)+1 then
   Verts[I].Coord1:=Coords[I]
   else
   Verts[I].Coord1:=Coord(0,0);
 end;

 SetTexture(Tex);
 FDevice.SetVertexShader(FVF_PNC1T1);
 FDevice.DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, @Verts, SizeOf(TVPNC1T1));

 Inc(PrimsDrawn, 2);

end;


procedure TD3DRenderer.DrawFace(Points:array of TPoint3S;Coords:array of TTexCoord;
      Cols:array of TARGB;Tex:TTexture=nil;Norm:PPoint3S=nil);
var
 Verts: array[0..3] of TVPNC1T1;
 I:LongInt;

begin
 if High(Points)<>2 then
  raise ED3DError.Create('Error in TD3DRenderer.DrawFace  -  parameter "Ponits" must have 2 values.');

 //if not CanDraw then Exit;
  if not FBegin then
    BeginScene;

 for I:=0 to 2 do
  begin
   Verts[I].Pos:=Points[I];
   if I<High(Cols)+1 then
    Verts[I].Color1:=Cols[I]
   else
     Verts[I].Color1:=$FFFFFFFF;
  if Norm<>nil then
   Verts[I].Normal:=Norm^
   else
   Verts[I].Normal:=Point3S(0,0,1);
   if I<High(Coords)+1 then
   Verts[I].Coord1:=Coords[I]
   else
   Verts[I].Coord1:=Coord(0,0);
 end;

 SetTexture(Tex);
 FDevice.SetVertexShader(FVF_PNC1T1);
 FDevice.DrawPrimitiveUP(D3DPT_TRIANGLELIST, 1, @Verts, SizeOf(TVPNC1T1));

 Inc(PrimsDrawn, 1);

end;



initialization

  Renderers.RegisterRenderer(CreateInfo(RenName, RenDesc, TD3DRenderer));

  DXGUtils.SetIdentityMatrix(Mat);

end.

