unit RenOGL;

interface

uses Windows, Classes, SysUtils, Renderers, OGL, GfxWin, WinUtils, WindowCls,
  DeviceModes, GfxTypes, Colors, Textures, Types3D, OGLMisc, BaseMgrs,
  DrawBuffers, Utils3D, Images, ImgUtils, TexUtils, GMessages, GfxMessages,
  DynLights, Lights;


type

  EOGLInit = class(Exception)
  end;
  EOGLError = class(Exception)
  end;


  TOGLRenderer = class(TRenderer)
  private
    FGLRC: HGLRC;
    FDC: HDC;
    FWin: TGfxWin;
    FOldCW: Word;
    FBitCount: byte;
    FDeskBits: byte;

    FLastTex: DWord;

    procedure SetPreInit;

  protected

    function GetDeviceDesc: string; override;
    function GetDeviceDriver: string; override;
    function GetTexMem: Dword; override;

  public
    constructor Create; override;
    destructor Destroy; override;

    procedure Init; override;
    procedure Done; override;
    function Reset: boolean; override;

    procedure EnumFullScrModes(List: TBaseMgr); override;


  // sets functions
    function SetDWord(Mode: TSetDWord; Value: DWord): boolean; override;
    function SetFloat(Mode: TSetFloat; Value: Single): boolean; override;

   // projection/viewports
    function SetOrtho(Width, Height: LongInt): boolean; override;
    function SetPerspective(Fov, NearP, BackP: single; Aspect: single = 0):
    boolean; override;
    function ConfigView(X, Y, Width, Height: LongInt): boolean; override;
    function GetProjMatrix: TMatrix; override;

     /////////// features
    function Enable(F: TFeature): boolean; override;
    function Disable(F: TFeature): boolean; override;

      ////////// drawing
    function Present: boolean; override;
    function ClearTarget(Color: TARGB): boolean; override;
    function ClearZBuffer(Value: single): boolean; override;
    function DrawBuffer(Buff: TDrawBuffer): boolean; override;
    function FreeDrawBuffer(Buff: TDrawBuffer): boolean; override;
    procedure DrawTriStrip(Points:array of TPoint3S;Coords:array of TTexCoord;
      Col:TARGB=$FFFFFFFF;Tex:TTexture=nil;Norm:PPoint3S=nil);override;
    procedure DrawFace(Points:array of TPoint3S;Coords:array of TTexCoord;
      Cols:array of TARGB;Tex:TTexture=nil;Norm:PPoint3S=nil);override;


    // tex
    procedure CreateDevTex(Tex: TTexture); override;
    procedure DeleteDevTex(Tex: TTexture); override;
    function SetTexture(Tex: TTexture): boolean; override;

    // view
    function ViewIdentity: boolean; override;
    function ViewLookAt(Pos, At, Up: TPoint3S): boolean; override;
    function GetViewMatrix: TMatrix; override;

    // world
    function WorldIdentity: boolean; override;
    function WorldTranslate(X, Y, Z: single): boolean; override;
    function WorldScale(X, Y, Z: single): boolean; override;
    function WorldRotate(X, Y, Z: single): boolean; override;
    function WorldMatrix(Mat: TMatrix): boolean; override;
    function GetWorldMatrix: TMatrix; override;
    function WorldPush: boolean; override;
    function WorldPop: boolean; override;
    function WorldMultiply(Mat: TMatrix): boolean; override;

  // dynamic lights
    function CreateLight(Light: TDynamicLight): boolean; override;
    function EnableLight(Idx: LongInt): boolean; override;
    function DisableLight(Idx: LongInt): boolean; override;


  published

  end;


const
  RenName = 'OGLRenderer';
  RenDesc = 'OpenGL 1.3 Renderer';

  sOGLErr = 'OGL Error';

implementation

uses GGraphic;

constructor TOGLRenderer.Create;
begin
  inherited Create;
  Name := RenName;
  Description := RenDesc;

end;

destructor TOGLRenderer.Destroy;
begin
  Done;
  inherited Destroy;
end;


function TOGLRenderer.GetDeviceDesc: string;
begin
  Result := glGetString(GL_RENDERER);
end;

function TOGLRenderer.GetDeviceDriver: string;
begin
  Result := glGetString(GL_VENDOR);

end;

function TOGLRenderer.GetTexMem: Dword;
begin
  Result := 0;
end;


procedure TOGLRenderer.Init;
var D, S: LongInt;
begin

  try

    FDeskBits := GetDeviceCaps(FDC, BITSPIXEL);
    FWin := TGfxWin(GetWin(Params.Handle));
    FDC := GetDC(Params.Handle);
    FOldCW := Default8087CW;
    Set8087CW($133F); { Disable all fpu exceptions }

    SetPreInit;
    FGLRC := CreateRenderingContext(FDC, [opDoubleBuffered], FBitCount,
      Params.DepthBits, Params.StencilBits, 0, 0,
      0);
    ActivateRenderingContext(FDC, FGLRC);

    glGetIntegerv(GL_DEPTH_BITS, @D);
    glGetIntegerv(GL_STENCIL_BITS, @S);

    UsedParams.DepthBits := D;
    UsedParams.StencilBits := S;
    UsedParams.UseDepth := UsedParams.DepthBits > 0;
    UsedParams.BackBuffers := 1;
    UsedParams.UseHardWare := true;
    UsedParams.UseVSync := true;
    UsedParams.UseHTL := not true;
    UsedParams.MultiSamples := 0;

  except
    raise EOGLInit.Create('Error when creating Rendering Context');
  end;
  if (FGLRC = 0) then
    raise EOGLInit.Create('Error when initializing OpenGL');

  glPixelTransferi(GL_MAP_COLOR, GL_FALSE);
  glPixelTransferi(GL_RED_SCALE, 1);
  glPixelTransferi(GL_RED_BIAS, 0);
  glPixelTransferi(GL_GREEN_SCALE, 1);
  glPixelTransferi(GL_GREEN_BIAS, 0);
  glPixelTransferi(GL_BLUE_SCALE, 1);
  glPixelTransferi(GL_BLUE_BIAS, 0);
  glPixelTransferi(GL_ALPHA_SCALE, 1);
  glPixelTransferi(GL_ALPHA_BIAS, 0);

  glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
  glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
  glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
  glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);


  glEnable(GL_TEXTURE_2D);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glClearColor(0.2, 0.2, 0.2, 1);
  glClear(GL_COLOR_BUFFER_BIT);

  glHint(GL_FOG_HINT, GL_NICEST);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
  glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);

  ReadImplementationProperties;
  ReadExtensions;




  Present;

  inherited Init;
end;

procedure TOGLRenderer.Done;
begin
  inherited Done;

  //DeactivateRenderingContext;
  //DestroyRenderingContext(FGLRC);

  Set8087CW(FOldCW);

end;



procedure TOGLRenderer.SetPreInit;
var Mode: LongInt;
begin
  UsedParams.Handle := Params.Handle;
  Mode := -1;
  if FWin = nil then
  begin
    FWin := TGfxWin.Create();
    FWin.SetHandle(Params.Handle);
  end;

  if Params.FullScreen then
  begin
    FWin.CanSetMode := false;
    FWin.SetFullScreen(Params.Width, Params.Height, Params.ColorBits);

    if Params.RefreshRate = 0 then
      Mode := FindModeHighestRefresh(Params.Width, Params.Height,
        Params.ColorBits);
    if Params.RefreshRate > 0 then
    begin
      Mode := FindModeEx(Params.Width, Params.Height, Params.ColorBits,
        Params.RefreshRate);
      if Mode = -1 then
        Mode := FindMode(Params.Width, Params.Height, Params.ColorBits);
    end;


// on my monitor 800x600 @ 75 Hz does HuhlyHuhly
// remove it
    if (params.Width = 800) and (DeviceModeList[Mode].Refresh = 75) then
      Mode := FindModeEx(800, 600, Params.ColorBits, 72);

    SetMode(Mode);

    FBitCount := Params.ColorBits;
    UsedParams.FullScreen := true;
    UsedParams.ColorBits := FBitCount;
    UsedParams.Width := Params.Width;
    UsedParams.Height := Params.Height;


  end
  else
  begin
    DeviceModes.ResetMode;
    FWin.SetWindowed;
    FWin.CanSetMode := true;

    FBitCount := FDeskBits;
    UsedParams.ColorBits := FDeskBits;
    UsedParams.FullScreen := false;
    UsedParams.Width := FWin.ClientWidth;
    UsedParams.Height := FWin.ClientHeight;


  end;


  UsedParams.RefreshRate := GetDeviceCaps(FDC, VREFRESH);


  if UsedParams.FullScreen then
    Messages.SendMessage(MSG_GFX_TOFULLSCREEN, Self)
  else
    Messages.SendMessage(MSG_GFX_TOWINDOW, Self)

end;


function TOGLREnderer.Reset: boolean;
begin
  Result := false;
  if not Inited then Exit;
  SetPreInit;
  FLastTex := 0;
  inherited Reset;

end;


procedure TOGLRenderer.EnumFullScrModes(List: TBaseMgr);
var I: LongInt;
  MI: TModeInfo;
begin
  GetDeviceModes;

  for I := 0 to MaxDeviceModes - 1 do
  begin
    MI := DeviceModeList[I];
    if MI.Bits >= 16 then
      List.Add(TFullScreenMode.Create(MI.Width, MI.Height, MI.Bits,
        MI.Refresh));
  end;

end;



function GetState(F: TFeature): TGLenum;
begin
  Result := 0;
  case F of
    gfLighting: Result := GL_LIGHTING;
    gfAlphaBlend: Result := GL_BLEND;
    gfStencil: Result := GL_STENCIL_TEST;
    gfFog: Result := GL_FOG;
    gfDither: Result := GL_DITHER;
    gfAlphaTest: Result := GL_ALPHA_TEST;
// gfClipPlane: Result:=GL_CLIP_PLANE1;
    gfVertexColors: Result := GL_COLOR_MATERIAL;
    gfZBuffer: Result := GL_DEPTH_TEST;
    gfNormalizeNormals: Result := GL_NORMALIZE;
    gfZWrite: Result := 0;

  end;

end;


function TOGLRenderer.Enable(F: TFeature): boolean;
var State: TGLenum;
begin
  Result := false;
  if not Inited then Exit;

  State := GetState(F);
  glEnable(State);
  Result := boolean(glIsEnabled(State));
  Graphic.FeatureSet[F] := Result;
end;

function TOGLRenderer.Disable(F: TFeature): boolean;
var State: TGLenum;
begin
  Result := false;
  if not Inited then Exit;

  State := GetState(F);
  glDisable(State);
  Result := not boolean(glIsEnabled(State));
  Graphic.FeatureSet[F] := false;
end;


function TOGLRenderer.SetOrtho(Width, Height: LongInt): boolean;
begin
  Result := false;
  if not Inited then Exit;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  glOrtho(0, Width, Height, 0, -1, 1);

  Result := true;
end;

function TOGLRenderer.SetPerspective(Fov, NearP, BackP: single; Aspect: single):
boolean;
begin
  Result := false;
  if not Inited then Exit;
  if Aspect < 0.01 then
    Aspect := AspectRatio;
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(Fov, Aspect, NearP, BackP);

  Result := true;
end;

function TOGLRenderer.GetProjMatrix: TMatrix;
begin
  if not Inited then Exit;
  glGetFloatv(GL_PROJECTION_MATRIX, @Result.f);
  Result._43 := Result._43 / 2;
  Result._33 := -Result._33;
  Result._34 := -Result._34;
end;


function TOGLRenderer.ConfigView(X, Y, Width, Height: LongInt): boolean;
begin
  Result := false;
  if not Inited then Exit;
  glViewPort(X, Y, Width, Height);

  Result := true;

end;

function TOGLRenderer.SetDWord(Mode: TSetDWord; Value: DWord): boolean;
var Col: TGLColor;
  AVal: TGLInt;
  AEnum: TGLEnum;
  AEnum2: TGLEnum;
begin
  Result := false;
  if not Inited then Exit;
  Aval := 0;
  AEnum := 0;
  AEnum2 := 0;
  case Mode of
    sdTexFilter:
      begin
        case Value of
          TEXF_NEAREST:
            if TextureUseMipMaps then
              AVal := GL_NEAREST_MIPMAP_NEAREST
            else
              AVal := GL_LINEAR;
          TEXF_BILINEAR:
            if TextureUseMipMaps then
              AVal := GL_LINEAR_MIPMAP_NEAREST
            else
              AVal := GL_LINEAR;
          TEXF_TRILINEAR:
            if TextureUseMipMaps then
              AVal := GL_LINEAR_MIPMAP_LINEAR
            else
              AVal := GL_LINEAR;
        end;
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, AVal);
        case Value of
          TEXF_NEAREST: AVal := GL_NEAREST;
          TEXF_BILINEAR,
            TEXF_TRILINEAR: AVal := GL_LINEAR;
        end;
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, AVAl);
      end;
    sdAmbient:
      begin
        Col := GLColFromARGB(Value);
        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, @Col);
      end;
    sdFogMode:
      begin
        case Value of
          FOG_LINEAR: AVal := GL_LINEAR;
          FOG_EXP: AVal := GL_EXP;
          FOG_EXP2: AVal := GL_EXP2;
        end;
        glFogi(GL_FOG_MODE, AVal);
      end;
    sdFogColor:
      begin
        Col := GLColFromARGB(Value);
        glFogfv(GL_FOG_COLOR, @Col);
      end;
    sdShadeMode:
      begin
        case Value of
          SHADE_FLAT: AVal := GL_FLAT;
          SHADE_GOURAUD: AVal := GL_SMOOTH;
        end;
        glShadeModel(AVal);
      end;
    sdFillMode:
      begin
        case Value of
          FILL_POINT: AVal := GL_POINT;
          FILL_WIREFRAME: AVal := GL_LINE;
          FILL_SOLID: AVal := GL_FILL;
        end;
        glPolygonMode(GL_FRONT_AND_BACK, AVal);
      end;
    sdCullMode:
      begin
        case Value of
          CULL_CCW: AVal := GL_CCW;
          CULL_CW: AVal := GL_CW;
        end;
        glEnable(GL_CULL_FACE);
        glFrontFace(AVal);

        if Value = CULL_NONE then
          glDisable(GL_CULL_FACE);

      end;
    sdZFunc, sdAlphaFunc, sdStencilFunc:
      begin
        case Value of
          CMP_NEVER: AEnum := GL_NEVER;
          CMP_LESS: AEnum := GL_LESS;
          CMP_EQUAL: AEnum := GL_EQUAL;
          CMP_LESSEQUAL: AEnum := GL_LEQUAL;
          CMP_GREATER: AEnum := GL_GREATER;
          CMP_NOTEQUAL: AEnum := GL_NOTEQUAL;
          CMP_GREATEREQUAL: AEnum := GL_GEQUAL;
          CMP_ALWAYS: AEnum := GL_ALWAYS;
        end;
        case Mode of
          sdZFunc: glDepthFunc(AEnum);
          sdAlphaFunc:
            glAlphaFunc(AEnum,
              Clamp255(Graphic.DWordSet[sdAlphaRef]));
          sdStencilFunc:
            glStencilFunc(AEnum, Graphic.DWordSet[sdStencilRef],
              Graphic.DWordSet[sdStencilMask]);

        end;

      end;
    sdAlphaRef:
      begin
        Graphic.DWordSet[sdAlphaRef] := Value;
        SetDWord(sdAlphaFunc, Graphic.DWordSet[sdAlphaFunc]);
      end;
    sdStencilRef:
      begin
        Graphic.DWordSet[sdStencilRef] := Value;
        SetDWord(sdStencilFunc, Graphic.DWordSet[sdStencilFunc]);
      end;
    sdStencilMask:
      begin
        Graphic.DWordSet[sdStencilMask] := Value;
        SetDWord(sdStencilFunc, Graphic.DWordSet[sdStencilFunc]);
      end;
    sdSrcBlend, sdDstBlend:
      begin
        Graphic.DWordSet[Mode] := Value;
        case Graphic.DWordSet[sdSrcBlend] of
          BLEND_ZERO: AEnum := GL_ZERO;
          BLEND_ONE: AEnum := GL_ONE;
          BLEND_SRCCOLOR: AEnum := GL_SRC_COLOR;
          BLEND_INVSRCCOLOR: AEnum := GL_ONE_MINUS_SRC_COLOR;
          BLEND_SRCALPHA: AEnum := GL_SRC_ALPHA;
          BLEND_INVSRCALPHA: AEnum := GL_ONE_MINUS_SRC_ALPHA;
          BLEND_DSTALPHA: AEnum := GL_DST_ALPHA;
          BLEND_INVDSTALPHA: AEnum := GL_ONE_MINUS_DST_ALPHA;
          BLEND_DSTCOLOR: AEnum := GL_DST_COLOR;
          BLEND_INVDSTCOLOR: AEnum := GL_ONE_MINUS_DST_COLOR;
          BLEND_SRCALPHASAT: AEnum := GL_SRC_ALPHA_SATURATE;
        end;
        case Graphic.DWordSet[sdDstBlend] of
          BLEND_ZERO: AEnum2 := GL_ZERO;
          BLEND_ONE: AEnum2 := GL_ONE;
          BLEND_SRCCOLOR: AEnum2 := GL_SRC_COLOR;
          BLEND_INVSRCCOLOR: AEnum2 := GL_ONE_MINUS_SRC_COLOR;
          BLEND_SRCALPHA: AEnum2 := GL_SRC_ALPHA;
          BLEND_INVSRCALPHA: AEnum2 := GL_ONE_MINUS_SRC_ALPHA;
          BLEND_DSTALPHA: AEnum2 := GL_DST_ALPHA;
          BLEND_INVDSTALPHA: AEnum2 := GL_ONE_MINUS_DST_ALPHA;
          BLEND_DSTCOLOR: AEnum2 := GL_DST_COLOR;
          BLEND_INVDSTCOLOR: AEnum2 := GL_ONE_MINUS_DST_COLOR;
          BLEND_SRCALPHASAT: AEnum2 := GL_SRC_ALPHA_SATURATE;
        end;
        glBlendFunc(AEnum, AEnum2);
      end;
    sdTexAdrX, sdTexAdrY:
      begin
        case Value of
          TEXADR_WRAP: AVal := GL_REPEAT;
          TEXADR_MIRROR: AVal := GL_REPEAT;
          TEXADR_CLAMP: AVAl := GL_CLAMP_TO_EDGE;
        end;

        if Mode = sdTexAdrX then
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, AVal)
        else
          glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, AVal);

      end;


  end;

  Graphic.DWordSet[Mode] := Value;
  Result := false;

end;


function TOGLRenderer.SetFloat(Mode: TSetFloat; Value: Single): boolean;
begin
  Result := false;
  case Mode of
    sfFogStart: glFogf(GL_FOG_START, Value);
    sfFogEnd: glFogf(GL_FOG_END, Value);
    sfFogDensity: glFogf(GL_FOG_DENSITY, Value);

  end;

  Graphic.FloatSet[Mode] := Value;
end;


procedure TOGLRenderer.CreateDevTex(Tex: TTexture);
var TexRec: POGLTex;
  Comps: TGLInt;
  Fmt: TGLEnum;
  Temp: TImage;
begin
  if not Inited then Exit;
  DeleteDevTex(Tex);

  GetMem(TexRec, SizeOf(TOGLTex));

  glGenTextures(1, @TexRec^.TexID);
  glBindTexture(GL_TEXTURE_2D, TexRec^.TexID);

  Temp := Tex.Image.Clone;
  Temp.BitCount := 32;
  Temp.BilinearResize := false;

  Mirror(Temp, false, true);
  if Tex.ColorKeyEnable then
    SetColorKey32(Temp.Width, Temp.Height, Tex.ColorKey, Temp.Bits);

  SwapRGB32(Temp.Width, Temp.Height, Temp.Bits);

  case Tex.DevTexBits of
    32:
      begin
        Comps := GL_RGBA8;
        Fmt := GL_RGBA;


       // glTexImage2D suddenly stopped working 22.10.2002

       // if not Tex.UseMipMaps then
       //   glTexImage2D(GL_TEXTURE_2D,0,Comps,Temp.Width,Temp.Height,0,Fmt,
       //    GL_UNSIGNED_BYTE,Temp.Bits);

       // else
          gluBuild2DMipmaps(GL_TEXTURE_2D,
            Comps,
            Tex.Width,
            Tex.Height,
            Fmt,
            GL_UNSIGNED_BYTE,
            Temp.Bits);
      end;

    16:
      begin
        if Tex.Image.HasAlpha then
        begin
          Comps := GL_RGBA4;
          Fmt := GL_RGBA;
        end
        else
        begin
          Comps := GL_RGB5_A1;
          Fmt := GL_RGBA;
        end;


       // if not Tex.UseMipMaps then
       //   glTexImage2D(GL_TEXTURE_2D, 0, Comps, Temp.Width, Temp.Height, 0,
       //     Fmt,GL_UNSIGNED_BYTE,Temp.Bits)
       // else
          gluBuild2DMipmaps(GL_TEXTURE_2D,
            Comps,
            Tex.Width,
            Tex.Height,
            Fmt,
            GL_UNSIGNED_BYTE,
            Temp.Bits);
      end;
  end;
 
  Tex.DevTex := TexRec;

  Temp.Free;

  glBindTexture(Gl_TEXTURE_2D, 0);
  FLastTex := 0;

end;

procedure TOGLRenderer.DeleteDevTex(Tex: TTexture);
begin
  if Tex.DevTex = nil then Exit;
  if TOGLTex(Tex.DevTex^).TexID <> 0 then
  begin
    glDeleteTextures(1, @TOGLTex(Tex.DevTex^).TexID);
    FreeMem(Tex.DevTex);
    Tex.DevTex := nil;
  end;

end;

function TOGLRenderer.SetTexture(Tex: TTexture): boolean;
var TR: TOGLTex;
begin
  Result := false;
  if not Inited then Exit;

  if Tex = nil then
  begin
    glBindTexture(GL_TEXTURE_2D, 0);
    FLastTex := 0;
  end
  else
  begin
    if Tex.DevTex = nil then
    begin
      glBindTexture(GL_TEXTURE_2D, 0);
      FLastTex := 0;
    end
    else
    begin
      TR := TOGLTex(Tex.DevTex^);
      if TR.TEXID <> FLastTex then
      begin
        glBindTexture(GL_TEXTURE_2D, TR.TexID);
        SetDWord(sdTexFilter, Tex.Filter);
        SetDWord(sdTexAdrX, Graphic.DwordSet[sdTexAdrX]);
        SetDWord(sdTexAdrY, Graphic.DwordSet[sdTexAdrY]);


        FLastTex := TR.TexID;
      end;
    end;
  end;



end;

function TOGLRenderer.FreeDrawBuffer(Buff: TDrawBuffer): boolean;
begin
  if Buff.RenderData <> nil then
  begin
    if glIsList(TGLBuffer(Buff.RenderData^).DisplayList) = 1 then
      glDeleteLists(TGLBuffer(Buff.RenderData^).DisplayList, 1);
    FreeMem(Buff.RenderData);
  end;
  Buff.RenderData := nil;
  Result := true;
end;

function TOGLRenderer.ViewIdentity: boolean;
begin
  Result := false;
  if not Inited then Exit;
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
  Result := true;
end;

function TOGLRenderer.ViewLookAt(Pos, At, Up: TPoint3S): boolean;
begin
//  Result := false;
//  if not Inited then Exit;

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;

  gluLookAt(Pos.X, Pos.Y, -Pos.z, At.X, At.Y, -At.Z, Up.X, Up.Y, -Up.Z);

  Result := true;
end;

function TOGLRenderer.GetViewMatrix: TMatrix;
begin
//  if not Inited then Exit;
  glGetFloatv(GL_MODELVIEW_MATRIX, @Result.f);
  Result._13 := -Result._13;
  Result._23 := -Result._23;
  Result._32 := -Result._32;
  Result._43 := -Result._43;
end;


function TOGLRenderer.WorldIdentity: boolean;
begin
//  Result := false;
//  if not Inited then Exit;
  Camera.UpdateView;
//  Result := true;
end;

function TOGlRenderer.WorldTranslate(X, Y, Z: single): boolean;
begin
  Result := false;
  if not Inited then Exit;
  glMatrixMode(GL_MODELVIEW);
  glTranslatef(X, Y, -Z);
  Result := true;
end;

function TOGlRenderer.WorldScale(X, Y, Z: single): boolean;
begin
  Result := false;
  if not Inited then Exit;
  glMatrixMode(GL_MODELVIEW);
  glScalef(X, Y, Z);
  Result := true;
end;

function TOGlRenderer.WorldRotate(X, Y, Z: single): boolean;
begin
  Result := false;
  if not Inited then Exit;
  glMatrixMode(GL_MODELVIEW);
  glRotatef(Z, 0, 0, 1);
  glRotatef(X, -1, 0, 0);
  glRotatef(Y, 0, -1, 0);

  Result := true;
end;

function TOGLRenderer.WorldMatrix(Mat: TMatrix): boolean;
begin
  Result := false;
  if not Inited then Exit;
  glMatrixMode(GL_MODELVIEW);
  Mat._13 := -Mat._13;
  Mat._23 := -Mat._23;
  Mat._32 := -Mat._32;
  Mat._43 := -Mat._43;
  glLoadMatrixf(@Mat.f);
  Result := true;

end;

function TOGLRenderer.GetWorldMatrix: TMatrix;
begin
  if not Inited then Exit;
  glGetFloatv(GL_MODELVIEW_MATRIX, @Result.f);
  Result._13 := -Result._13;
  Result._23 := -Result._23;
  Result._32 := -Result._32;
  Result._43 := -Result._43;
end;

function TOGLRenderer.WorldPush: boolean;
begin
//  Result := false;
//  if not Inited then Exit;
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix;
//  Result := true;
end;

function TOGLRenderer.WorldPop: boolean;
begin
//  Result := false;
//  if not Inited then Exit;
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix;
//  Result := true;
end;

function TOGLRenderer.WorldMultiply(Mat: TMatrix): boolean;
begin
//  Result := false;
//  if not Inited then Exit;
  glMatrixMode(GL_MODELVIEW);
  Mat._13 := -Mat._13;
  Mat._23 := -Mat._23;
  Mat._32 := -Mat._32;
  Mat._43 := -Mat._43;
  glMultMatrixf(@Mat.f);
//  Result := true;
end;

function TOGLRenderer.CreateLight(Light: TDynamicLight): boolean;
var GLIdx: TGLEnum;
  LC: TGLColor;
  P: TPoint3S;
  P4: array[0..3] of single;

begin
  Result := false;
  if not Inited then Exit;


  GlIdx := GL_LIGHT0 + Light.Index;
  LC := LCToGLC(Light.DiffuseColor);
  glLightfv(GLIdx, GL_DIFFUSE, @LC);
  LC := LCToGLC(Light.AmbientColor);
  glLightfv(GLIdx, GL_AMBIENT, @LC);
  LC := LCToGLC(Light.SpecularColor);
  glLightfv(GLIdx, GL_SPECULAR, @LC);
  P := Light.Position;
  P4[0] := P.X;
  P4[1] := P.Y;
  P4[2] := -P.Z;
  P4[3] := 1;
  glLightfv(GLIdx, GL_POSITION, @P4);
  P := Light.AngleVector;
  P.Z := -P.Z;
  glLightfv(GLIdx, GL_SPOT_DIRECTION, @P);

  if Light.LightType = ltSpot then
    glLightf(GLIdx, GL_SPOT_CUTOFF, Light.SpotOuterConeAngle)
  else
    glLightf(GLIdx, GL_SPOT_CUTOFF, 180);


  glLightf(GLIdx, GL_SPOT_EXPONENT, CoTan(DegToRad(Abs(Light.SpotOuterConeAngle
    - Light.SpotInnerConeAngle))));



  glLightf(GLIdx, GL_CONSTANT_ATTENUATION, Light.ConstantAttenuation);
  glLightf(GLIdx, GL_LINEAR_ATTENUATION, Light.LinearAttenuation);
  glLightf(GLIdx, GL_QUADRATIC_ATTENUATION, Light.QuadraticAttenuation);


  Result := true;
end;

function TOGLRenderer.EnableLight(Idx: LongInt): boolean;
begin
  Result := false;
  if not Inited then Exit;
  glEnable(GL_LIGHT0 + Idx);
  Result := boolean(glIsEnabled(GL_LIGHT0 + Idx));

end;

function TOGLRenderer.DisableLight(Idx: LongInt): boolean;
begin
  Result := false;
  if not Inited then Exit;
  glDisable(GL_LIGHT0 + Idx);
  Result := not boolean(glIsEnabled(GL_LIGHT0 + Idx));
end;

///////////////////////////  DRAWING

function TOGLRenderer.Present: boolean;
begin
  Result := false;
  if not Inited then Exit;

  if SwapBuffers(FDC) then
    Result := true
  else
    raise EOGLError.Create(sOGLErr + ' : SwapBuffers failed');
end;



function TOGLRenderer.ClearTarget(Color: TARGB): boolean;
begin
  Result := false;
  if not Inited then Exit;
  glClearColor(GetRS(Color), GetGS(Color), GetBS(Color), GetAS(Color));
  glClear(GL_COLOR_BUFFER_BIT);
  Result := true;

end;

function TOGLRenderer.ClearZBuffer(Value: single): boolean;
begin
  Result := false;
  if not Inited then Exit;
  glClearDepth(Value);
  glClear(GL_DEPTH_BUFFER_BIT);
  Result := true;

end;


function TOGLRenderer.DrawBuffer(Buff: TDrawBuffer): boolean;
var
  Mode: TGLEnum;
  I, VCount: LongInt;

  Pos: TPoint3S;
  Normal: TPoint3S;
  Col1: TARGB;
  Col2: TARGB;
  Coord1: TTexCoord;
  Coord2: TTexCoord;
  RD: TGLBuffer;
  PRD: PGLBuffer;

  Col: TGLColor;

begin
  Result := false;
  if not Inited then Exit;
  if Buff = nil then Exit;




  if ufTexture in Buff.Used then
    SetTexture(Buff.Texture);

  if Buff.RenderData = nil then
  begin

    VCount := Buff.Points.Count;
    if VCount <= 0 then Exit;

    Mode := GL_POINTS;
    case Buff.Mode of
      dmPoints: Mode := GL_POINTS;
      dmLineList: Mode := GL_LINES;
      dmLineStrip: Mode := GL_LINE_STRIP;
      dmTriangleList: Mode := GL_TRIANGLES;
      dmTriangleStrip: Mode := GL_TRIANGLE_STRIP;
      dmTriangleFan: Mode := GL_TRIANGLE_FAN;
    end;

    if ufDeviceBuff in Buff.Used then
    begin
      GetMem(PRD, SizeOf(RD));
      Buff.RenderData := PRD;

      PRD^.DisplayList := glGenLists(1);
      glNewList(PRD^.DisplayList, GL_COMPILE);
    end;


    glBegin(Mode);

    Normal := Point3S(0, 0, 1);
    Col1 := $FFFFFFFF;
    Col2 := $FFFFFFFF;
    Coord1 := Coord(0, 0);
    Coord2 := Coord(0, 0);


    for I := 0 to VCount - 1 do
    begin
      Pos := TPoint3S(Buff.Points[I]^);
      Pos.Z := -Pos.Z; // important

      if Buff.Normals[I] <> nil then
        Normal := TPoint3S(Buff.Normals[I]^);
      Normal.Z := -Normal.Z;

      if Buff.Colors1[I] <> nil then
        Col1 := TARGB(Buff.Colors1[I]^);

      if Buff.Colors2[I] <> nil then
        Col2 := TARGB(Buff.Colors2[I]^);

      if Buff.Coords1[I] <> nil then
        Coord1 := TTexCoord(Buff.Coords1[I]^);

      if Buff.Coords2[I] <> nil then
        Coord2 := TTexCoord(Buff.Coords2[I]^);


      if ufColors1 in Buff.Used then
      begin
        Col := GLColFromARGB(Col1);
        glColor4fv(@Col);
      end;

      if ufNormals in Buff.Used then
        glNormal3fv(@Normal);

      if ufCoords1 in Buff.Used then
        glTexCoord2fv(@Coord1);

      glVertex3fv(@Pos);

    end;
    glEnd;

    if ufDeviceBuff in Buff.Used then
      glEndList;

  end
  else
  begin
    RD := TGLBuffer(Buff.RenderData^);
   // if glIsList(RD.DisplayList) = 1 then
    glCallList(RD.DisplayList);

  end;

  Inc(PrimsDrawn, Buff.Primitives);


  Result := true;


end;

procedure TOGLRenderer.DrawTriStrip(Points:array of TPoint3S;Coords:array of TTexCoord;
      Col:TARGB=$FFFFFFFF;Tex:TTexture=nil;Norm:PPoint3S=nil);
var ANorm:TPoint3S;
    I:LongInt;
begin
 if High(Points)<>3 then
  raise EOGLError.Create('Error in TOGLRenderer.DrawTriStrip  -  parameter "Ponits" must have 4 values.');

 SetTexture(Tex);

 if Norm=nil then
 ANorm:=Point3S(0,0,1)
 else
 ANorm:=Norm^;

 glBegin(GL_TRIANGLE_STRIP);
 for I:=Low(Points) to High(Points) do
  begin
   glColor4ub(Col shr 16,Col shr 8,Col,Col shr 24);
   glNormal3f(ANorm.X,ANorm.Y,-ANorm.Z);
   glTexCoord2fv(@Coords[I]);
   glVertex3f(Points[I].X,Points[I].Y,-Points[I].Z);
 end;
  glEnd;

  Inc(PrimsDrawn, 2);
end;

procedure TOGLRenderer.DrawFace(Points:array of TPoint3S;Coords:array of TTexCoord;
      Cols:array of TARGB;Tex:TTexture=nil;Norm:PPoint3S=nil);
var ANorm:TPoint3S;
    I:LongInt;
begin
 if High(Points)<>2 then
  raise EOGLError.Create('Error in TOGLRenderer.DrawTriStrip  -  parameter "Ponits" must have 4 values.');

 SetTexture(Tex);

 if Norm=nil then
 ANorm:=Point3S(0,0,1)
 else
 ANorm:=Norm^;

 glBegin(GL_TRIANGLES);
 for I:=Low(Points) to High(Points) do
  begin
   glColor4ub(Cols[I] shr 16,Cols[I] shr 8,Cols[I],Cols[I] shr 24);
   glNormal3f(ANorm.X,ANorm.Y,-ANorm.Z);
   glTexCoord2fv(@Coords[I]);
   glVertex3f(Points[I].X,Points[I].Y,-Points[I].Z);
 end;
  glEnd;

  Inc(PrimsDrawn, 1);
end;

initialization

  Renderers.RegisterRenderer(CreateInfo(RenName, RenDesc, TOGLRenderer));

end.

