unit Renderers;

interface

uses
  Windows, Classes, SysUtils, BaseObjs, Engines,  GfxTypes,
  Colors, Textures, Types3D, DrawBuffers,TextureMgrs,
  WinUtils,GMessages,GfxMessages,DynLights,BaseMgrs;

type

  TRenderer = class;


  TRenderer = class(TBaseObj)
  private
    FOwner: TEngine;
    FDescription: string;
    FParams: TRenderInitParams;
    FUsedParams: TRenderInitParams;
    FActTex:TTexture;

  protected
    function GetDeviceDesc: string; virtual; abstract;
    function GetDeviceDriver: string; virtual; abstract;
    function GetTexMem: Dword; virtual; abstract;

  public
    PrimsDrawn:LongInt;

    constructor Create; override;
    destructor Destroy; override;

    procedure Init; override;
    procedure Done; override;
    function Reset:boolean; virtual;
    function CheckParams: boolean; virtual;
    procedure EnumFullScrModes(List:TBaseMgr);virtual;abstract;

    // set function
    function SetDWord(Mode: TSetDWord; Value: DWord): boolean; virtual;
      abstract;
    function SetFloat(Mode: TSetFloat; Value: Single): boolean; virtual;
      abstract;


    // viewports/projection
    function SetOrtho(Width, Height: LongInt): boolean; virtual; abstract;
    function SetPerspective(Fov,NearP,BackP: single;Aspect:single=0): boolean; virtual; abstract;
    function ConfigView(X,Y,Width,Height:LongInt):boolean;virtual;abstract;
    function GetProjMatrix:TMatrix;virtual;abstract;

    /// features - setrenderstate(d3drs_blablaenable,1) or glEnable()
    function Enable(F: TFeature): boolean; virtual; abstract;
    function Disable(F: TFeature): boolean; virtual; abstract;


    /// drawing functions
    function Present: boolean; virtual; abstract;
    function ClearTarget(Color: TARGB): boolean; virtual; abstract;
    function ClearZBuffer(Value: single): boolean; virtual; abstract;
    function DrawBuffer(Buff: TDrawBuffer): boolean; virtual; abstract;
    function FreeDrawBuffer(Buff:TDrawBuffer):boolean;virtual;abstract;
      // used for font's characters quads
    procedure DrawTriStrip(Points:array of TPoint3S;Coords:array of TTexCoord;
      Col:TARGB=$FFFFFFFF;Tex:TTexture=nil;Norm:PPoint3S=nil);virtual;abstract;
      // used for portal engine
    procedure DrawFace(Points:array of TPoint3S;Coords:array of TTexCoord;
      Cols:array of TARGB;Tex:TTexture=nil;Norm:PPoint3S=nil);virtual;abstract;


    // tex
    procedure CreateDevTex(Tex:TTexture);virtual;abstract;
    procedure DeleteDevTex(Tex:TTexture);virtual;abstract;
    function SetTexture(Tex:TTexture):boolean;virtual;abstract;

    // view
    function ViewIdentity:boolean;virtual;abstract;
    function ViewLookAt(Pos,At,Up:TPoint3S):boolean;virtual;abstract;
    function GetViewMatrix:TMatrix;virtual;abstract;

    // world
    function WorldIdentity:boolean;virtual;abstract;
    function WorldTranslate(X,Y,Z:single):boolean;virtual;abstract;
    function WorldScale(X,Y,Z:single):boolean;virtual;abstract;
    function WorldRotate(X,Y,Z:single):boolean;virtual;abstract;
    function WorldMatrix(Mat:TMatrix):boolean;virtual;abstract;
    function GetWorldMatrix:TMatrix;virtual;abstract;
    function WorldPush:boolean;virtual;abstract;
    function WorldPop:boolean;virtual;abstract;
    function WorldMultiply(Mat:TMatrix):boolean;virtual;abstract;

    // dynamic lights
    function CreateLight(Light:TDynamicLight):boolean;virtual;abstract;
    function EnableLight(Idx:LongInt):boolean;virtual;abstract;
    function DisableLight(Idx:LongInt):boolean;virtual;abstract;

    property Owner: TEngine read FOwner write FOwner;
    property Params: TRenderInitParams read FParams write FParams;
    property UsedParams: TRenderInitParams read FUsedParams write FUsedParams;
    property DeviceDesc: string read GetDeviceDesc;
    property DeviceDriver: string read GetDeviceDriver;
    property TextureMem: Dword read GetTexMem;

    property ActiveTexture:TTexture read FActTex;

  published


    property Description: string read FDescription write FDescription;
  end;


  TRendererClass = class of TRenderer;


  TRendererInfo = class
    Name: string;
    Desc: string;
    Cls: TRendererClass
  end;


procedure RegisterRenderer(Info: TRendererInfo);
function CreateInfo(Name, Desc: string; Cls: TRendererClass): TRendererInfo;

var RegRenderers: TBaseMgr;

const

  sBadParams = 'Bad Renderer Initialization Parameters';

implementation


function IsSame(RI1, RI2: TRendererInfo): boolean;
begin

  Result := Iff(CompStr(RI1.Name, RI2.Name), True, False) and
    Iff(CompStr(RI1.Desc, RI2.Desc), True, False) and
    Iff(RI1.Cls = RI2.Cls, True, False);


end;

procedure RegisterRenderer(Info: TRendererInfo);
var I: LongInt;
  Already: boolean;
begin
  if RegRenderers = nil then
    RegRenderers := TBaseMgr.Create;

  Already := false;
  for I := 0 to RegRenderers.Count - 1 do
    if IsSame(TRendererInfo(RegRenderers[I]), Info) then Already := true;

  if not Already then
    RegRenderers.Add(TBaseObj(Info));

end;

function CreateInfo(Name, Desc: string; Cls: TRendererClass): TRendererInfo;
var Inf: TRendererInfo;
begin
  Inf := TRendererInfo.Create;
  Inf.Name := Name;
  Inf.Desc := Desc;
  Inf.Cls := Cls;
  Result := Inf;
end;


constructor TRenderer.Create;
begin
  inherited Create;
  FUsedParams := TRenderInitParams.Create;
  FParams := TRenderInitParams.Create;


end;

destructor TRenderer.Destroy;
begin

  FUsedParams.Free;
  FParams.Free;
  inherited Destroy;
end;


function TRenderer.CheckParams: boolean;
var OK: boolean;

  procedure DoError;
  begin
    raise EBadParams.Create(sBadParams);
    OK := false;
  end;

begin
  OK := true;
  if not IsWindow(FParams.Handle) then DoError;
  if not IntInArray(FParams.Width, [320, 400, 512, 640, 800, 1024, 1280, 1600])
    then DoError;
  if not IntInArray(FParams.Height, [200, 240, 300, 384, 400, 480, 600, 768,
    1024, 1200]) then DoError;
  if not (FParams.ColorBits in [0,16, 32]) then DoError;
  if not (FParams.DepthBits in [0, 16, 24, 32]) then DoError;
  if not (FParams.StencilBits in [0, 8]) then DoError;
  if not (FParams.BackBuffers in [1, 2]) then DoError;

  Result := OK;
end;

procedure TRenderer.Init;
begin
  inherited Init;
  Messages.SendMessage(MSG_GFX_INIT,Self);
end;

function TRenderer.Reset:boolean;
begin
 Result:=false;
 Messages.SendMessage(MSG_GFX_RESET,Self);
end;

procedure TRenderer.Done;
begin
 
  inherited Done;


end;


initialization
 RegRenderers:=TBaseMgr.Create;

finalization

 RegRenderers.Free;

end.

