unit GeoExt;

interface

uses Windows, SysUtils, Classes, WinUtils, BaseObjs, Colors, Types3D,
  Utils3D, Objs3D, GGraphic, VisObjs, DrawBuffers, Textures, GfxTypes,
  GeoObjs, TexMapping;

type
  TAngleLim90 = -90..90;
  TAngleLim360 = 0..360;
  TCapType = (ctNone, ctCenter, ctFlat, ctAntiCenter);

  TSphereParts = (spTopCap, spMain, spBottomCap);
  TSphereBuffers = array[TSphereParts] of TDrawBuffer;
  TSphereColors = array[TSphereParts] of TARGB;
  TSphereTextures = array[TSphereParts] of TTexture;
  TSphereVisible = array[TSphereParts] of boolean;
  TSphereTexMappers = array[TSphereParts] of TTexMapper;

  TSphereBase = class(TGeoObj)
  private
    FTopCap: TCapType;
    FBottomCap: TCaptype;

    FBaseP: TPointDynArray;
  protected
    procedure SetTexture(Value: TTexture); override;
    procedure SetColor(Value: TARGB); override;
    procedure SetClockWise(Value: boolean); override;
    procedure SetTopCap(Value: TCapType); virtual;
    procedure SetBottomCap(Value: TCapType); virtual;

  public
    Colors: TSphereColors;
    Textures: TSphereTextures;
    Buffers: TSphereBuffers;
    VisibleParts: TSphereVisible;
    TexMappers: TSphereTexMappers;


    constructor Create; override;
    destructor Destroy; override;

    procedure Init; override;
    procedure Done; override;
    procedure UpdateTexCoords;override;

    procedure SetTexInvert(cX, cY: boolean); override;
    procedure SetTexClamp(cX, cY: boolean); override;
    procedure SetTexClampValues(MinX, MinY, MaxX, MaxY: single); override;
    procedure SetTexOffset(cX, cY: single); override;
    procedure SetTexTiling(cX, cY: single); override;
    procedure SetTexSwapXY(swap: boolean); override;

  published

    property TopCap: TCapType read FTopCap write SetTopCap;
    property BottomCap: TCapType read FBottomCap write SetBottomCap;



  end;



  TSphere = class(TSphereBase)
  private
    FRadius: single;
    FAntiLength: single;
    FSlices: LongInt; // meridians(poledniky)
    FStacks: LongInt; // parallelograms(rovnobezky)
    FTop: TAngleLim90; // do not combine
    FBottom: TAngleLim90; // top/bottom with start/stop
    FStart: TAngleLim360; // it doesnt work properly
    FStop: TAngleLim360;

  protected
    procedure SetRadius(Value: single); virtual;
    procedure SetAntiLength(Value: single); virtual;
    procedure SetSlices(Value: LongInt); virtual;
    procedure SetStacks(Value: LongInt); virtual;
    procedure SetTop(Value: TAngleLim90); virtual;
    procedure SetBottom(Value: TAngleLim90); virtual;
    procedure SetStart(Value: TAngleLim360); virtual;
    procedure SetStop(Value: TAngleLim360); virtual;
    function GetInFrustum:boolean;override;

  public

    constructor Create; overload; override;
    constructor Create(ARadius: single; ATex: TTexture = nil; ASegments: LongInt
      = 8); reintroduce; overload; virtual;

    procedure Draw; override;
    procedure UpdateBuffer; override;
    procedure UpdateBounds;override;

    procedure AddNoise(X, Y, Z: single; ReCreate: boolean = true); override;
    procedure AddNoisePercent(X, Y, Z: LongInt; ReCreate: boolean = true);
      override;


  published

    property Radius: single read FRadius write SetRadius;
    property AntiCenterLenght: single read FAntiLength write SetAntiLength;
    property Slices: LongInt read FSlices write SetSlices;
    property Stacks: LongInt read FStacks write SetStacks;
    property Top: TAngleLim90 read FTop write SetTop;
    property Bottom: TAngleLim90 read FBottom write SetBottom;
    property Start: TAngleLim360 read FStart write SetStart;
    property Stop: TAngleLim360 read FStop write SetStop;



  end;

  TCylinder = class(TSphereBase)
  private
    FSides: LongInt;
    FRadius: single;
    FHeight: single;
    FStart: TAngleLim360;
    FHasCenter: boolean;
    FAntiLength: single;
    FCapCenterUnitYCoord:boolean;

  protected
    procedure SetRadius(Value: single); virtual;
    procedure SetHeight(Value: single); virtual;
    procedure SetAntiLength(Value: single); virtual;
    procedure SetSides(Value: LongInt); virtual;
    procedure SetStart(Value: TAngleLim360); virtual;
    procedure SetHasCenter(Value: boolean); virtual;
    procedure SetCapUnit(Value:boolean);virtual;
    function GetInFrustum:boolean;override;

  public

    constructor Create; overload; override;
    constructor Create(ARadius: single; AHeight: single; ATex: TTexture = nil;
      ASides: LongInt
      = 16); reintroduce; overload; virtual;

    procedure UpdateBuffer; override;
    procedure UpdateBounds;override;
    procedure Draw; override;

    procedure Assign(Source: TBaseObj);override;

    procedure AddNoise(X, Y, Z: single; ReCreate: boolean = true); override;
    procedure AddNoisePercent(X, Y, Z: LongInt; ReCreate: boolean = true);
      override;


  published

    property Radius: single read FRadius write SetRadius;
    property Height: single read FHeight write SetHeight;
    property AntiCenterLength: single read FAntiLength write SetAntiLength;
    property Sides: LongInt read FSides write SetSides;
    property Start: TAngleLim360 read FStart write SetStart;
    property HasCenter: boolean read FHasCenter write SetHasCenter;
    property CapCenterUnitYCoord:boolean read FCapCenterUnitYCoord write SetCapUnit;

  end;


const
  MaxSphereSegments = 150 * 150;


implementation

constructor TSphereBase.Create;
var Part: TSphereParts;
begin
  inherited Create;

  for Part := Low(TSphereParts) to High(TSphereParts) do
    Buffers[Part] := TDrawBuffer.Create;

  for Part := Low(TSphereParts) to High(TSphereParts) do
  begin
    TexMappers[Part] := TTexMapper.Create;
    TexMappers[Part].SetInvert(false, false);
    TexMappers[Part].SetOffset(0, 0);
    TexMappers[Part].SetClamp(false, false);
    TexMappers[Part].SetTiling(1, 1);
    TexMappers[Part].SetClampValues(-1, -1, 1, 1);
    TexMappers[Part].SwapXY := false;


  end;

  FillChar(VisibleParts, 3, 1);

  Color := $FFFFFFFF;

end;

destructor TSphereBase.Destroy;
var Part: TSphereParts;
begin
  Done;
  for Part := Low(TSphereParts) to High(TSphereParts) do
    Buffers[Part].Free;
  for Part := Low(TSphereParts) to High(TSphereParts) do
    TexMappers[Part].Free;
  inherited Destroy;
end;

procedure TSphereBase.Init;
begin
  inherited Init;
end;

procedure TSphereBase.Done;
var Part: TSphereParts;
begin
  inherited Done;
  for Part := Low(TSphereParts) to High(TSphereParts) do
    Buffers[Part].ClearAll;
end;

procedure TSphereBase.SetTexture(Value: TTexture);
begin
  inherited SetTexture(Value);
  Textures[spTopCap] := Texture;
  Textures[spMain] := Texture;
  Textures[spBottomCap] := Texture;
  UpdateBuffer;
end;

procedure TSphereBase.SetColor(Value: TARGB);
begin
  inherited SetColor(Value);
  Colors[spTopCap] := Color;
  Colors[spMain] := Color;
  Colors[spBottomCap] := Color;
  UpdateBuffer;
end;

procedure TSphereBase.SetClockWise(Value: boolean);
begin
  if ClockWise = Value then Exit;
  inherited SetClockWise(Value);
  UpdateBuffer;
end;


procedure TSphereBase.SetTexInvert(cX, cY: boolean);
var Part: TSphereParts;
begin
  for Part := Low(TSphereParts) to High(TSphereParts) do
    TexMappers[Part].SetInvert(cX, cY);
  UpdateTexCoords;
end;

procedure TSphereBase.SetTexClamp(cX, cY: boolean);
var Part: TSphereParts;
begin
  for Part := Low(TSphereParts) to High(TSphereParts) do
    TexMappers[Part].SetClamp(cX, cY);
  UpdateTexCoords;
end;

procedure TSphereBase.SetTexClampValues(MinX, MinY, MaxX, MaxY: single);
var Part: TSphereParts;
begin
  for Part := Low(TSphereParts) to High(TSphereParts) do
    TexMappers[Part].SetClampValues(MinX, MinY, MaxX, MaxY);
  UpdateTexCoords;
end;

procedure TSphereBase.SetTexOffset(cX, cY: single);
var Part: TSphereParts;
begin
  for Part := Low(TSphereParts) to High(TSphereParts) do
    TexMappers[Part].SetOffset(cX, cY);
  UpdateTexCoords;
end;

procedure TSphereBase.SetTexTiling(cX, cY: single);
var Part: TSphereParts;
begin
  for Part := Low(TSphereParts) to High(TSphereParts) do
    TexMappers[Part].SetTiling(cX, cY);
  UpdateTexCoords;
end;

procedure TSphereBase.SetTexSwapXY(swap: boolean);
var Part: TSphereParts;
begin
  for Part := Low(TSphereParts) to High(TSphereParts) do
    TexMappers[Part].SwapXY := swap;
  UpdateTexCoords;
end;

procedure TSphereBase.SetTopCap(Value: TCapType);
begin
  if FTopCap = Value then Exit;
  FTopCap := Value;
  UpdateBuffer;
end;

procedure TSphereBase.SetBottomCap(Value: TCapType);
begin
  if FBottomCap = Value then Exit;
  FBottomCap := Value;
  UpdateBuffer;
end;

procedure TSphereBase.UpdateTexCoords;
var Part: TSphereParts;
begin
  if not Inited then Exit;
  for Part := Low(TSphereParts) to High(TSphereParts) do
    TexMappers[Part].ProcessBuffer(Buffers[Part]);
  Self.Update;

  if Buffers[spTopCap].Coords1.Count>0 then
  TTexCoord(Buffers[spTopCap].Coords1[0]^).X:=0.5;
  if Buffers[spBottomCap].Coords1.Count>0 then
  TTexCoord(Buffers[spBottomCap].Coords1[0]^).X:=0.5;

  if UseDeviceBuffer then
    for Part := Low(TSphereParts) to High(TSphereParts) do
    begin
      Renderer.FreeDrawBuffer(Buffers[Part]);
      Graphic.DrawBuffer(Buffers[Part]);
    end;
end;


 //////////////////// TSPHERE /////////////////

constructor TSphere.Create;
begin
  inherited Create;

  FTopCap := ctFlat;
  FBottomCap := ctFlat;
  FRadius := 100;
  Slices := 8;
  Stacks := 8;
  FTop := 90;
  FBottom := -90;
  FStart := 0;
  FStop := 360;
  FAntiLength := 100;


end;

constructor TSphere.Create(ARadius: single; ATex: TTexture; ASegments: LongInt);
begin
  Create;

  Texture := ATex;
  FRadius := ARadius;
  Slices := ASegments;
  Stacks := ASegments;

end;


procedure TSphere.SetRadius(Value: single);
begin
  FRadius := Value;
  UpdateBuffer;
end;

procedure TSphere.SetAntiLength(Value: single);
begin
  FAntiLength := Value;
  if (FTopCap = ctAntiCenter) or (FBottomCap = ctAntiCenter) then
    UpdateBuffer;
end;

procedure TSphere.SetSlices(Value: LongInt);
begin
  if (FStacks * Value) > MaxSphereSegments then
    Value := MaxSphereSegments div FStacks;

  if FSlices = Value then Exit;
  FSlices := Value;
  UpdateBuffer;
end;

procedure TSphere.SetStacks(Value: LongInt);
begin
  if (FSlices * Value) > MaxSphereSegments then
    Value := MaxSphereSegments div FSlices;

  if FStacks = Value then Exit;
  FStacks := Value;
  UpdateBuffer;
end;

procedure TSphere.SetTop(Value: TAngleLim90);
begin
  if FTop = Value then Exit;
  FTop := Value;
  UpdateBuffer;
end;

procedure TSphere.SetBottom(Value: TAngleLim90);
begin
  if FBottom = Value then Exit;
  FBottom := Value;
  UpdateBuffer;
end;

procedure TSphere.SetStart(Value: TAngleLim360);
begin
  if FStart = Value then Exit;
  FStart := Value;
  UpdateBuffer;
end;

procedure TSphere.SetStop(Value: TAngleLim360);
begin
  if FStop = Value then Exit;
  FStop := Value;
  UpdateBuffer;
end;

function TSphere.GetInFrustum:boolean;
begin
Result:=Camera.IsSphereVisible(AddPoint(WorldPos,Bounds.Center),Bounds.Rad);
end;

procedure TSphere.AddNoise(X, Y, Z: single; ReCreate: boolean);
var I, J: LongInt;
  P: TPoint3S;
  pp, Pbc, Pc: LongInt;
  V1, V2: TPoint3S;

begin
  if not Inited then Exit;
 // !! Works only on main sphere body
  if Recreate then
    UpdateBuffer;
  Pc := Length(FBaseP);

  for I := 0 to Pc - 1 do
  begin
    P := FBaseP[I];
    P.X := P.X + RandSignF(Random * X / FRadius);
    P.Y := P.Y + RandSignF(Random * Y / FRadius);
    P.Z := P.Z + RandSignF(Random * Z / FRadius);
    FBaseP[I] := P;
  end;

  pbc := 0;
  pp := 0;

  for J := FStacks - 1 downto 0 do
  begin
    V1.Arr[2] := FBaseP[pp].Z;
    V2.Arr[2] := FBaseP[pp + FSlices + 1].Z;

    for i := 0 to FSlices do
    begin
      V1.Arr[0] := FBaseP[pp].x; ;
      V2.Arr[0] := FBaseP[pp + FSlices + 1].x; ;
      V1.Arr[1] := FBaseP[pp].y; ;
      V2.Arr[1] := FBaseP[pp + FSlices + 1].y; ;

      TPoint3S(Buffers[spMain].Points[pbc]^) := V2;
      TPoint3S(Buffers[spMain].Points[pbc + 1]^) := V1;

      Inc(pbc, 2);
      Inc(pp);

    end;

  end;

  for I := 0 to Buffers[spMain].Points.Count - 1 do
  begin
    P := TPoint3S(Buffers[spMain].Points[I]^);
    if (Abs(P.X) < 0.000001)
      and (Abs(P.Y) < 0.000001)
      and (Abs(P.z) < 0.000001) then
    begin
      P.Z := -1;

    end;
    TPoint3S(Buffers[spMain].Points[I]^) := P;
  end;

  if UseDeviceBuffer then
  begin
    Renderer.FreeDrawBuffer(Buffers[spMain]);
    Graphic.DrawBuffer(Buffers[spMain]);
  end;

end;


procedure TSphere.AddNoisePercent(X, Y, Z: LongInt; ReCreate: boolean);
begin
  AddNoise(FRadius * X / 100, FRadius * Y / 100, FRadius * Z / 100, Recreate);
end;


procedure TSphere.Draw;
var Part: TSphereParts;
begin
  if not Inited then Exit;
  inherited Draw;
  PreDraw;
  Renderer.WorldScale(FRadius, FRadius, FRadius);
  for Part := Low(TSphereParts) to High(TSphereParts) do
    if VisibleParts[Part] then
      Graphic.DrawBuffer(Buffers[Part]);
  AfterDraw;

end;

procedure TSphere.UpdateBounds;
var I:LongInt;
    PP,SS,P:TPoint3S;
    Mat:TMatrix;
begin
 PP:=WorldPos;
 Mat:=CreateRotationMatrix(Angle);
 SS:=GetAbsScale;

 Bounds.Max.Z:=+FRadius;
 Bounds.Min.Z:=-FRadius;
 Bounds.Max.X:=+FRadius;
 Bounds.Min.X:=-FRadius;
 Bounds.Max.Y:=+FRadius;
 Bounds.Min.Y:=-FRadius;

 Bounds.Center.X:=0;
 Bounds.Center.Y:=0;
 Bounds.Center.Z:=0;

 Bounds.Rad:=FRadius;

 Bounds.Max:=MatrixTransform(Bounds.Max,Mat);
 Bounds.Max:=MulPoint(Bounds.Max,SS);
 Bounds.Max:=AddPoint(Bounds.Max,PP);

 Bounds.Min:=MatrixTransform(Bounds.Min,Mat);
 Bounds.Min:=MulPoint(Bounds.Min,SS);
 Bounds.Min:=AddPoint(Bounds.Min,PP);

end;

procedure TSphere.UpdateBuffer;
var
  V1, V2, N1: TPoint3S;
  AngTop, AngBottom, AngStart, AngStop, StepV, StepH: single;
  SinP, CosP, SinP2, CosP2, SinT, CosT, Phi, Phi2, Theta: single;
  uTexCoord, uTexFactor, vTexFactor, vTexCoord0, vTexCoord1: Single;
  II, I, J: Integer;
  Tris: DWord;
  DoReverse: Boolean;
  Used: TUsedFields;
begin
  if not Inited then Exit;
  DoReverse := not ClockWise;
  AngTop := DegToRad(-FTop);
  AngBottom := DegToRad(-FBottom);
  AngStart := DegToRad(FStart + 270);
  AngStop := DegToRad(FStop + 270);
  StepH := (AngStop - AngStart) / FSlices;
  StepV := (AngTop - AngBottom) / FStacks;
  Tris := 0;


  Used := [ufColors1, ufCoords1, ufTexture];

  if UseDeviceBuffer then
    Include(Used, ufDeviceBuff);

  if HasNormals then
    Include(Used, ufNormals);

  // top cap
  if (FTop < 90) and (FTopCap in [ctCenter, ctFlat, ctAntiCenter]) then
  begin
    Graphic.DrawBegin(dmTriangleFan, Buffers[spTopCap]);
    Graphic.DrawCount(FSlices);
    Graphic.DrawUse(Used);
    Graphic.DrawColor1(Colors[spTopCap]);

    SinCos(AngTop, SinP, CosP);
    Graphic.DrawCoord1(0.5, 0.5);
{      if DoReverse then
         Graphic.DrawNormal(0,-1,0)
     else Graphic.DrawNormal(0,1,0);}
    if FTopCap = ctCenter then
      Graphic.DrawVertex(0, 0, 0)
    else
      if FTopCap = ctFlat then
      begin
     {    if DoReverse then
            N1:=Point3S(0, -1, 0)
         else N1:=Point3S(0,1,0);}

        Graphic.DrawVertex(0, 0, -SinP);
      end
      else
      begin
    // norm
        Graphic.DrawVertex(0, 0, 1 + FAntiLength / FRadius);


      end;
    V1.Arr[2] := -SinP;
    Theta := AngStart;
    for I := 0 to FSlices do
    begin
      SinCos(Theta, SinT, CosT);
      V1.Arr[0] := CosP * SinT;
      if ClockWise then
      begin
        V1.Arr[1] := -CosP * CosT;
        Graphic.DrawCoord1(SinT / 2 + 0.5, 1 - (CosT / 2 + 0.5));
      end
      else
      begin
        V1.Arr[1] := CosP * CosT;
        Graphic.DrawCoord1(SinT / 2 + 0.5, CosT / 2 + 0.5);
      end;

    {     if FTopCap=ctCenter then begin
            N1:=VectorPerpendicular(YVector, V1);
            if DoReverse then Negate(N1);
         end; }

    //     Graphic.DrawNormal(N1);
      Graphic.DrawVertex(V1);
      Theta := Theta + StepH;
    end;
    Graphic.DrawTexture(Texture);
    Graphic.DrawEnd(false);
  end;



   // main body
  Phi := AngTop;
  Phi2 := Phi - StepV;
  uTexFactor := 1 / (FSlices);
  vTexFactor := 1 / (FStacks - 1);

  Graphic.DrawBegin(dmTriangleStrip, Buffers[spMain]);
  Graphic.DrawUse(Used);
  Graphic.DrawColor1(Colors[spMain]);

  FBaseP := nil;

  SetLength(FBaseP, FStacks * (FSlices + 1));
  II := 0;

  for J := FStacks - 1 downto 0 do
  begin
    Theta := AngStart;
    SinCos(Phi, SinP, CosP);
    SinCos(Phi2, SinP2, CosP2);
    V1.Arr[2] := -SinP;
    V2.Arr[2] := -SinP2;
    vTexCoord0 := 1 - (j + 1) * vTexFactor;
    vTexCoord1 := 1 - j * vTexFactor;


  //    Graphic.DrawColor1($FF00ff00);

    for i := 0 to FSlices do
    begin
      SinCos(Theta, SinT, CosT);
      V1.Arr[0] := CosP * SinT;
      V2.Arr[0] := CosP2 * SinT;
      V1.Arr[1] := CosP * CosT;
      V2.Arr[1] := CosP2 * CosT;
      uTexCoord := (i) * uTexFactor;

      FBaseP[II] := V1;

  {       if DoReverse then begin
            N1:=V1;
            NegateVector(N1);
            glNormal3fv(@N1);
         end else glNormal3fv(@V1);}
      if ClockWise then
      begin
        Graphic.DrawCoord1(uTexCoord, vTexCoord1);
        Graphic.DrawVertex(V2);
      end
      else
      begin
        Graphic.DrawCoord1(uTexCoord, vTexCoord0);
        Graphic.DrawVertex(V1);
      end;
{         if DoReverse then begin
            N1:=V2;
            NegateVector(N1);
            glNormal3fv(@N1);
         end else glNormal3fv(@V2);}

      if ClockWise then
      begin
        Graphic.DrawCoord1(uTexCoord, vTexCoord0);
        Graphic.DrawVertex(V1);
      end
      else
      begin
        Graphic.DrawCoord1(uTexCoord, vTexCoord1);
        Graphic.DrawVertex(V2);
      end;


      Inc(Tris, 2);
      Inc(II);


      Theta := Theta + StepH;
    end;
    Phi := Phi2;
    Phi2 := Phi2 - StepV;
  end;
  Graphic.DrawTexture(Texture);
  Graphic.DrawCount(Tris - 2);
  Graphic.DrawEnd(false);



  if (FBottom > -90) and (FBottomCap in [ctCenter, ctFlat, ctAntiCenter]) then
  begin
    Graphic.DrawBegin(dmTriangleFan, Buffers[spBottomCap]);
    Graphic.DrawCount(FSlices);
    Graphic.DrawUse(Used);
    Graphic.DrawColor1(Colors[spBottomCap]);


    SinCos(AngBottom, SinP, CosP);
    Graphic.DrawCoord1(0.5, 0.5);
{      if DoReverse then
         glNormal3f(0, 1, 0)
      else glNormal3f(0, -1, 0);}
    if FBottomCap = ctCenter then
      Graphic.DrawVertex(0, 0, 0)
    else
      if FBottomCap = ctFlat then
      begin
        Graphic.DrawVertex(0, 0, -SinP);
{         if DoReverse then
            MakeVector(N1, 0, -1, 0)
         else N1:=YVector;}
      end
      else
      begin
    // norm
        Graphic.DrawVertex(0, 0, -1 - FAntiLength / FRadius);

      end;

    V1.Arr[2] := -SinP;
    Theta := AngStop;
    for I := 0 to FSlices do
    begin
      SinCos(Theta, SinT, CosT);
      V1.Arr[0] := CosP * SinT;

      if ClockWise then
      begin
        V1.Arr[1] := -CosP * CosT;
        Graphic.DrawCoord1(SinT * 0.5 + 0.5, 1 - (CosT * 0.5 + 0.5));
      end
      else
      begin
        V1.Arr[1] := CosP * CosT;
        Graphic.DrawCoord1(SinT * 0.5 + 0.5, CosT * 0.5 + 0.5);
      end;
      {  if FTopCap = ctCenter then begin
            N1:=VectorPerpendicular(AffineVectorMake(0, -1, 0), V1);
            if DoReverse then NegateVector(N1);
         end;}

       //  glNormal3fv(@N1);
      Graphic.DrawVertex(V1);


      Theta := Theta - StepH;
    end;
    Graphic.DrawTexture(Texture);
    Graphic.DrawEnd(false);
  end;
  UpdateTexCoords;
  UpdateBounds;
end;


//////////////////// TCYLINDER

constructor TCylinder.Create;
begin
  inherited Create;

  FSides := 16;
  FRadius := 50;
  FHeight := 200;
  FStart := 0;
  FHasCenter := true;
  FTopCap := ctFlat;
  FBottomCap := ctFlat;
  FAntiLength := 100;
  Color := $FFFFFFFF;

end;

constructor TCylinder.Create(ARadius: single; AHeight: single; ATex: TTexture;
  ASides: LongInt);
begin
  Create;
  FRadius := ARadius;
  FHeight := AHeight;
  FSides := ASides;
  Texture := ATex;

end;


procedure TCylinder.SetRadius(Value: single);
begin
  if FRadius = Value then Exit;
  FRadius := Value;
  UpdateBuffer;
end;

procedure TCylinder.SetHeight(Value: single);
begin
  if FHeight = Value then Exit;
  FHeight := Value;
  UpdateBuffer;
end;


procedure TCylinder.SetAntiLength(Value: single);
begin
  FAntiLength := Value;
  if (FTopCap = ctAntiCenter) or (FBottomCap = ctAntiCenter) then
    UpdateBuffer;
end;

procedure TCylinder.SetSides(Value: LongInt);
begin
  if FSides = Value then Exit;
  FSides := Value;
  UpdateBuffer;
end;


procedure TCylinder.SetStart(Value: TAngleLim360);
begin
  if FStart = Value then Exit;
  FStart := Value;
  UpdateBuffer;
end;

procedure TCylinder.SetHasCenter(Value: boolean);
begin
  if FHasCenter = Value then Exit;
  FHasCenter := Value;
  if FStart > 180 then
    UpdateBuffer;
end;

procedure TCylinder.SetCapUnit(Value:boolean);
begin
 if FCapCenterUnitYCoord=Value then Exit;
 FCapCenterUnitYCoord:=Value;
 if (FTopCap<>ctNone)or(FBottomCap<>ctNone) then
  UpdateBuffer;
end;

procedure TCylinder.Assign(Source: TBaseObj);
var C: TCylinder;
    Side:TBoxSides;
    Part:TSphereParts;
begin
  inherited Assign(Source);
  if not (Source is TCylinder) then Exit;
  C := TCylinder(Source);
  FHeight:=C.Height;
  FRadius:=C.Radius;
  FBottomCap:=C.BottomCap;
  FTopCap:=C.TopCap;
  Colors:=C.Colors;
  VisibleParts:=C.VisibleParts;
  Textures:=C.Textures;

   for Part := Low(TSphereParts) to High(TSphereParts) do
    TexMappers[Part].Copy(C.TexMappers[Part]);

  FAntiLength:=C.AntiCenterLength;
  FCapCenterUnitYCoord:=C.CapCenterUnitYCoord;
  FHasCenter:=C.HasCenter;
  FSides:=C.Sides;
  FStart:=C.Start;

  Init;
end;


function TCylinder.GetInFrustum:boolean;
begin
// Result:=Camera.IsBoxVisible(Bounds);
 Result:=Camera.IsSphereVisible(AddPoint(WorldPos,Bounds.Center),Bounds.Rad);
end;

procedure TCylinder.AddNoise(X, Y, Z: single; ReCreate: boolean = true);
var I, Pc: LongInt;
  P, V1, V2: TPoint3S;
  Part: TSphereParts;
begin
  if not Inited then Exit;

  if Recreate then
    UpdateBuffer;
  Pc := Length(FBaseP);

  for I := 0 to Pc - 1 do
  begin
    P := FBaseP[I];
    P.X := P.X + RandSignF(Random * X);
    P.Y := P.Y + RandSignF(Random * Y);
    FBaseP[I] := P;
  end;

  for I := 0 to Pc - 1 do
  begin
    V1 := FBaseP[I];
    V1.Z := FHeight / 2 + RandSignF(Random * Z); ;
    V2 := FBaseP[I];
    V2.Z := -FHeight / 2 + RandSignF(Random * Z); ;

    if not ClockWise then
    begin
      TPoint3S(Buffers[spTopCap].Points[Pc - I]^) := V1;
      TPoint3S(Buffers[spBottomCap].Points[I + 1]^) := V2;
      SwapPoints(V1, V2);
    end
    else
    begin
      TPoint3S(Buffers[spTopCap].Points[I + 1]^) := V1;
      TPoint3S(Buffers[spBottomCap].Points[Pc - I]^) := V2;

    end;

    TPoint3S(Buffers[spMain].Points[I * 2]^) := V1;
    TPoint3S(Buffers[spMain].Points[I * 2 + 1]^) := V2;

  end;

  TPoint3S(Buffers[spMain].Points[Buffers[spMain].Points.Count - 2]^) :=
    TPoint3S(Buffers[spMain].Points[0]^);
  TPoint3S(Buffers[spMain].Points[Buffers[spMain].Points.Count - 1]^) :=
    TPoint3S(Buffers[spMain].Points[1]^);

  TPoint3S(Buffers[spTopCap].Points[Pc]^) :=
    TPoint3S(Buffers[spTopCap].Points[1]^);
  TPoint3S(Buffers[spBottomCap].Points[Pc]^) :=
    TPoint3S(Buffers[spBottomCap].Points[1]^);




  if UseDeviceBuffer then
    for Part := Low(TSphereParts) to High(TSphereParts) do
    begin
      Renderer.FreeDrawBuffer(Buffers[Part]);
      Graphic.DrawBuffer(Buffers[Part]);
    end;



end;


procedure TCylinder.AddNoisePercent(X, Y, Z: LongInt; ReCreate: boolean = true);
begin
  AddNoise(FRadius * X / 100, FRadius * Y / 100, FHeight * Z / 100, ReCreate);
end;


procedure TCylinder.Draw;
var Part: TSphereParts;
begin
  if not Inited then Exit;
  inherited Draw;
  PreDraw;
  for Part := Low(TSphereParts) to High(TSphereParts) do
    if VisibleParts[Part] then
      Graphic.DrawBuffer(Buffers[Part]);
  AfterDraw;
end;

procedure TCylinder.UpdateBounds;
var I:LongInt;
    PP,SS,P:TPoint3S;
    Mat:TMatrix;
begin
 PP:=WorldPos;
 Mat:=CreateRotationMatrix(Angle);
 SS:=GetAbsScale;

 Bounds:=BoundInfo(Point3S(1e20,1e20,1e20),Point3S(-1e20,-1e20,-1e20));

 Bounds.Max.Z:=+FHeight/2;
 Bounds.Min.Z:=-FHeight/2;
 Bounds.Max.X:=+FRadius;
 Bounds.Min.X:=-FRadius;
 Bounds.Max.Y:=+FRadius;
 Bounds.Min.Y:=-FRadius;

 Bounds.Center.X:=0;
 Bounds.Center.Y:=0;
 Bounds.Center.Z:=0;

 Bounds.Rad:=PointDistance(Bounds.Min,Bounds.Max)/2;

 if FTopCap=ctAntiCenter then
 begin
 Bounds.Max.Z:=FHeight/2+AntiCenterLength;
 Bounds.Min.Z:=-FHeight/2-AntiCenterLength;
 end;

 Bounds.Max:=MatrixTransform(Bounds.Max,Mat);
 Bounds.Max:=MulPoint(Bounds.Max,SS);
 Bounds.Max:=AddPoint(Bounds.Max,PP);

 Bounds.Min:=MatrixTransform(Bounds.Min,Mat);
 Bounds.Min:=MulPoint(Bounds.Min,SS);
 Bounds.Min:=AddPoint(Bounds.Min,PP);

end;


procedure TCylinder.UpdateBuffer;
var I: Integer;
  SinA, CosA, Ang, Step: single;
  BaseC: single;
  P1, P2: TPoint3S;
  Sds, Tris: DWord;
  Used: TUsedFields;
  UCoords: array of Single;
begin
  if not Inited then Exit;
  UpdateBounds;

  Step := 360 / FSides;
  Ang := 0;
  Tris := 0;
  Sds := FSides+1;
  BaseC := 1.0;

  Used := [ufColors1, ufCoords1, ufTexture];

  if UseDeviceBuffer then
    Include(Used, ufDeviceBuff);

  if HasNormals then
    Include(Used, ufNormals);

  FBaseP := nil;

 // Main Body
  if Height > 0.1 then
  begin
    Graphic.DrawBegin(dmTriangleStrip, Buffers[spMain]);
    Graphic.DrawColor1(Colors[spMain]);
    Graphic.DrawTexture(Textures[spMain]);
    Graphic.DrawUse(Used);
    if (FStart > 180) and (FHasCenter) then
    begin
      Sds := FSides - 1;

      P1 := Point3S(0, 0, -FHeight / 2);
      P2 := Point3S(0, 0, FHeight / 2);

      if ClockWise then
      begin
        Graphic.DrawCoord1(1, 0);
        Graphic.DrawVertex(P2);
        Graphic.DrawCoord1(1, 1);
        Graphic.DrawVertex(P1);
      end
      else
      begin
        Graphic.DrawCoord1(1, 1);
        Graphic.DrawVertex(P1);
        Graphic.DrawCoord1(1, 0);
        Graphic.DrawVertex(P2);
      end;

      SetLength(FBaseP, Length(FBaseP) + 1);
      FBaseP[Length(FBaseP) - 1] := P1;

      SetLength(UCoords, Length(UCoords) + 1);
      UCoords[Length(UCoords) - 1] := 1;

      BaseC := 1 - 0.25;

      Inc(Tris, 2);
    end;
    for I := 0 to Sds-1 do
    begin

      SinCos(DegToRad(180 + Ang), SinA, CosA);
      P1 := Point3S(CosA * FRadius, SinA * FRadius, -FHeight / 2);
      P2 := Point3S(CosA * FRadius, SinA * FRadius, FHeight / 2);


      if not ((Ang > (360 - FStart)) and (Ang < 360)) then
      begin
        if ClockWise then
        begin
          Graphic.DrawCoord1(BaseC - 1 / FSides * I, 0);
          Graphic.DrawVertex(P2);
          Graphic.DrawCoord1(BaseC - 1 / FSides * I, 1);
          Graphic.DrawVertex(P1);
        end
        else
        begin
          Graphic.DrawCoord1(BaseC - 1 / FSides * I, 1);
          Graphic.DrawVertex(P1);
          Graphic.DrawCoord1(BaseC - 1 / FSides * I, 0);
          Graphic.DrawVertex(P2);
        end;

        SetLength(FBaseP, Length(FBaseP) + 1);
        FBaseP[Length(FBaseP) - 1] := P1;

        SetLength(UCoords, Length(UCoords) + 1);
        UCoords[Length(UCoords) - 1] := BaseC - 1 / FSides * I;


        Inc(Tris, 2);
      end;

      Ang := Ang + Step;
    end;
    if (FStart > 180) and (FHasCenter) then
    begin
      P1 := Point3S(0, 0, -FHeight / 2);
      P2 := Point3S(0, 0, FHeight / 2);

      if ClockWise then
      begin
        Graphic.DrawCoord1(0, 0);
        Graphic.DrawVertex(P2);
        Graphic.DrawCoord1(0, 1);
        Graphic.DrawVertex(P1);
      end
      else
      begin
        Graphic.DrawCoord1(0, 1);
        Graphic.DrawVertex(P1);
        Graphic.DrawCoord1(0, 0);
        Graphic.DrawVertex(P2);
      end;

      SetLength(FBaseP, Length(FBaseP) + 1);
      FBaseP[Length(FBaseP) - 1] := P1;

      SetLength(UCoords, Length(UCoords) + 1);
      UCoords[Length(UCoords) - 1] := 0;

      Inc(Tris, 2);
    end;
    Graphic.DrawCount(Tris - 2);
    Graphic.DrawEnd(false);
  end;

 
 // Top Cap
  if FTopCap in [ctCenter..ctAntiCenter] then
  begin
    Graphic.DrawBegin(dmTriangleFan, Buffers[spTopCap]);
    Graphic.DrawColor1(Colors[spTopCap]);
    Graphic.DrawTexture(Textures[spTopCap]);
    Graphic.DrawUse(Used);

    if FTopCap = ctFlat then
      if FCapCenterUnitYCoord then
      Graphic.DrawCoord1(0.5,0)
      else
      Graphic.DrawCoord1(0.5, 1 - FRadius / FHeight)
    else
      if FCapCenterUnitYCoord then
      Graphic.DrawCoord1(0.5,0)
      else
      Graphic.DrawCoord1(0.5, 1 - FRadius / FHeight);

      Graphic.DrawCoord1(0.5, 1 - FAntiLength / FHeight);

    case FTopCap of
      ctFlat: Graphic.DrawVertex(0, 0, FHeight / 2);
      ctCenter: Graphic.DrawVertex(0, 0, 0);
      ctAntiCenter: Graphic.DrawVertex(0, 0, FHeight / 2 + FAntiLength);
    end;

    if not ClockWise then
      for I := Length(FBaseP) - 1 downto 0 do
      begin
        Graphic.DrawCoord1(UCoords[I], 1);
        Graphic.DrawVertex(FBaseP[I].X, FBaseP[I].Y, FHeight / 2);
      end
    else
      for I := 0 to Length(FBaseP) - 1 do
      begin
        Graphic.DrawCoord1(UCoords[I], 1);
        Graphic.DrawVertex(FBaseP[I].X, FBaseP[I].Y, FHeight / 2);
      end;

    Graphic.DrawCount(Length(FBaseP) - 1);
    Graphic.DrawEnd(false);


  end;

 // Bottom Cap
  if FBottomCap in [ctCenter..ctAntiCenter] then
  begin
    Graphic.DrawBegin(dmTriangleFan, Buffers[spBottomCap]);
    Graphic.DrawColor1(Colors[spBottomCap]);
    Graphic.DrawTexture(Textures[spBottomCap]);
    Graphic.DrawUse(Used);

    if FTopCap = ctFlat then
      Graphic.DrawCoord1(0.5, 1 - FRadius / FHeight)
    else
      Graphic.DrawCoord1(0.5, 1 - FAntiLength / FHeight);

    case FBottomCap of
      ctFlat: Graphic.DrawVertex(0, 0, -FHeight / 2);
      ctCenter: Graphic.DrawVertex(0, 0, 0);
      ctAntiCenter: Graphic.DrawVertex(0, 0, -FHeight / 2 - FAntiLength);
    end;

    if ClockWise then
      for I := Length(FBaseP) - 1 downto 0 do
      begin
        Graphic.DrawCoord1(UCoords[I], 1);
        Graphic.DrawVertex(FBaseP[I].X, FBaseP[I].Y, -FHeight / 2);
      end
    else
      for I := 0 to Length(FBaseP) - 1 do
      begin
        Graphic.DrawCoord1(UCoords[I], 1);
        Graphic.DrawVertex(FBaseP[I].X, FBaseP[I].Y, -FHeight / 2);
      end;
    Graphic.DrawCount(Length(FBaseP) - 1);
    Graphic.DrawEnd(false);


  end;

  UCoords := nil;
  UpdateTexCoords;
  UpdateBounds;
end;







end.

