unit GeoObjs;

interface

uses Windows, SysUtils, Classes, WinUtils, BaseObjs, Colors, Types3D,
  Utils3D, Objs3D, GGraphic, VisObjs, DrawBuffers, Textures, GfxTypes,
  TexMapping;

type


  TGeoObj = class(TVisibleObj)
  private
    FUseDeviceBuffer: boolean;
    FCulled: boolean;
    FTexture: TTexture;
    FColor: TARGB;
    FHasNormals: boolean;
    FClockWise: boolean;
    FOldCull: DWord;


  protected
    procedure SetTexture(Value: TTexture); virtual;
    procedure SetColor(Value: TARGB); virtual;
    procedure SetHasNormals(Value: boolean); virtual;
    procedure SetClockWise(Value: boolean); virtual;
    procedure SetUseDB(Value: boolean); virtual;

    procedure PreDraw; virtual;
    procedure AfterDraw; virtual;

  public
    constructor Create; override;
    destructor Destroy; override;

    procedure Init; override;

    procedure UpdateBuffer; virtual;
    procedure UpdateTexCoords;virtual;

    procedure Assign(Source: TBaseObj); override;

    // for global texture coords settings
    procedure SetTexInvert(cX,cY:boolean);virtual;
    procedure SetTexClamp(cX,cY:boolean);virtual;
    procedure SetTexClampValues(MinX,MinY,MaxX,MaxY:single);virtual;
    procedure SetTexOffset(cX,cY:single);virtual;
    procedure SetTexTiling(cX,cY:single);virtual;
    procedure SetTexSwapXY(swap:boolean);virtual;


    // adds value to each point in object;ReCreate - recreate original shape before noise application
    procedure AddNoise(X, Y, Z: single; ReCreate: boolean = true); virtual;
    procedure AddNoisePercent(X, Y, Z: LongInt; ReCreate: boolean = true);
      virtual;

  published
    property UseDeviceBuffer: boolean read FUseDeviceBuffer write SetUseDB;
    property Culled: boolean read FCulled write FCulled;
    property Texture: TTexture read FTexture write SetTexture;
    property Color: TARGB read FColor write SetColor;
    property HasNormals: boolean read FHasNormals write SetHasNormals;
    property ClockWise: boolean read FClockWise write SetClockWise default true;

  end;


  TQuadPoints = array[0..3] of TPoint3S;
  TQuadColors = array[0..3] of TARGB;
  TQuadCoords = array[0..3] of TTexCoord;


  TGeoQuad = class(TGeoObj)
  private
    FBuffer: TDrawBuffer;
    FWidth: Single; //  X Dimension
    FLength: Single; //  Y Dimension
    FHeight: Single; //  Z Position
    FTexMapper:TTexMapper;
  protected

    procedure SetWidth(Value: Single); virtual;
    procedure SetLength(Value: Single); virtual;
    procedure SetHeight(Value: Single); virtual;
    procedure SetTexture(Value: TTexture); override;
    procedure SetColor(Value: TARGB); override;
    procedure SetClockWise(Value: boolean); override;

    function GetInFrustum:boolean;override;

  public
    Points: TQuadPoints;
    Colors: TQuadColors;
    Coords: TQuadCoords;
    Normal: TPoint3S;

    constructor Create; overload; override;
    constructor Create(AWidth, ALength, AHeight: Single; ATex: TTexture = nil);
      reintroduce; overload; virtual;
    destructor Destroy; override;

    procedure Init; override;
    procedure Done; override;
    procedure Draw; override;
    procedure UpdateBuffer; override;
    procedure UpdateTexCoords;override;
    procedure UpdateBounds;override;

    procedure Assign(Source: TBaseObj); override;

    procedure SetTexInvert(cX,cY:boolean);override;
    procedure SetTexClamp(cX,cY:boolean);override;
    procedure SetTexClampValues(MinX,MinY,MaxX,MaxY:single);override;
    procedure SetTexOffset(cX,cY:single);override;
    procedure SetTexTiling(cX,cY:single);override;
    procedure SetTexSwapXY(swap:boolean);override;

    procedure AddNoise(X,Y,Z:single; ReCreate: boolean = true);override;
    procedure AddNoisePercent(X,Y,Z:LongInt; ReCreate: boolean = true);override;

  published

    property Width: Single read FWidth write SetWidth;
    property Height: Single read FHeight write SetHeight;
    property Length: Single read FLength write SetLength;
    property TexMapper:TTexMapper read FTexMapper;

  end;

  TBoxSides = (
    bsBottom, // points to negative Z
    bsFront, // points to positive Y
    bsBack, // points to negative Y
    bsLeft, // points to positive X
    bsRight, // points to negative X
    bsTop // points to positive Z
    );


  TBoxBasePoints = array[0..7] of TPoint3S;
  TBoxNormals = array[TBoxSides] of TPoint3S;
  TBoxTextures = array[TBoxSides] of TTexture;
  TBoxPoints = array[TBoxSides, 0..3] of PPoint3S; // pointers to TBoxBasePoints
  TBoxBuffers = array[TBoxSides] of TDrawBuffer;
  TBoxColors = array[TBoxSides] of TARGB;
  TBoxCoords = array[TBoxSides, 0..3] of TTexCoord;
  TBoxVisibleSides = array[TBoxSides] of boolean;
  TBoxTexMappers=array[TBoxSides]of TTexMapper;

  TGeoBox = class(TGeoObj)
  private
    FWidth: Single; //  X Dimension
    FLength: Single; //  Y Dimension
    FHeight: Single; //  Z Dimension

    procedure SetDefaultCoords;

  protected

    procedure SetWidth(Value: Single); virtual;
    procedure SetLength(Value: Single); virtual;
    procedure SetHeight(Value: Single); virtual;
    procedure SetTexture(Value: TTexture); override;
    procedure SetColor(Value: TARGB); override;
    procedure SetClockWise(Value: boolean); override;

    function GetInFrustum:boolean;override;

  public
    BasePoints: TBoxBasePoints;
    Points: TBoxPoints;
    Normals: TBoxNormals;
    Textures: TBoxTextures;
    Buffers: TBoxBuffers;
    Colors: TBoxColors;
    Coords: TBoxCoords;
    VisibleSides: TBoxVisibleSides;
    TexMappers:TBoxTexMappers;

    constructor Create; overload; override;
    constructor Create(AWidth, ALength, AHeight: Single; ATex: TTexture = nil);
      reintroduce; overload; virtual;
    destructor Destroy; override;

    procedure Init; override;
    procedure Done; override;
    procedure Draw; override;
    procedure UpdateBuffer; override;
    procedure UpdateTexCoords;override;
    procedure UpdateBounds;override;

    procedure Assign(Source: TBaseObj); override;

    procedure AddNoise(X,Y,Z:single; ReCreate: boolean = true);override;
    procedure AddNoisePercent(X,Y,Z:LongInt; ReCreate: boolean = true);override;

    procedure SetTexInvert(cX,cY:boolean);override;
    procedure SetTexClamp(cX,cY:boolean);override;
    procedure SetTexClampValues(MinX,MinY,MaxX,MaxY:single);override;
    procedure SetTexOffset(cX,cY:single);override;
    procedure SetTexTiling(cX,cY:single);override;
    procedure SetTexSwapXY(swap:boolean);override;


  published

    property Width: Single read FWidth write SetWidth;
    property Height: Single read FHeight write SetHeight;
    property Length: Single read FLength write SetLength;

  end;

  TCustomGeo = class(TGeoObj)
  private
    FBuffers: TDrawBufferMgr;
  protected

  public
    constructor Create; override;
    destructor Destroy; override;

    procedure Draw; override;

    procedure Assign(Source: TBaseObj); override;

    property Buffers: TDrawBufferMgr read FBuffers;

  published

  end;


implementation

///////////////////////  TGEOOBJ

constructor TGeoObj.Create;
begin
  inherited Create;
  FUseDeviceBuffer := true;
  FCulled := true;
  FClockWise := true;
  FHasNormals := false;
end;

destructor TGeoObj.Destroy;
begin

  inherited Destroy;
end;

procedure TGeoObj.Init;
begin
  inherited Init;
  UpdateBuffer;
end;

procedure TGeoObj.SetTexture(Value: TTexture);
begin
  FTexture := Value;
end;

procedure TGeoObj.SetColor(Value: TARGB);
begin
  FColor := Value;
end;

procedure TGeoObj.SetHasNormals(Value: boolean);
begin
  FHasNormals := Value;
  UpdateBuffer;
end;

procedure TGeoObj.SetUseDB(Value: boolean);
begin
  FUseDeviceBuffer := Value;
  UpdateBuffer;
end;

procedure TGeoObj.SetClockWise(Value: boolean);
begin
  FClockWise := Value;
end;

procedure TGeoObj.Assign(Source: TBaseObj);
var GO: TGeoObj;
begin
  inherited Assign(Source);
  if not (Source is TGeoObj) then Exit;
  GO := TGeoObj(Source);
  FUseDeviceBuffer:=GO.UseDeviceBuffer;
  FColor:=GO.Color;
  FCulled:=GO.Culled;
  FHasNormals:=GO.HasNormals;
  FTexture:=GO.Texture;
  FClockWise:=GO.ClockWise;

end;


procedure TGeoObj.UpdateBuffer;
begin
end;

procedure TGeoObj.UpdateTexCoords;
begin
end;

procedure TGeoObj.PreDraw;
begin
  FOldCull := 0;
  if not Culled then
  begin
    FOldCull := Graphic.DWordSet[sdCullMode];
    Renderer.SetDWord(sdCullMode, CULL_NONE);
  end;
end;


procedure TGeoObj.AfterDraw;
begin
  if not Culled then
    Renderer.SetDWord(sdCullMode, FOldCull);
  if Parent=nil then
  // Renderer.WorldIdentity;
end;

procedure TGeoObj.AddNoise(X, Y, Z: single; ReCreate: boolean);
begin
end;

procedure TGeoObj.AddNoisePercent(X, Y, Z: LongInt; ReCreate: boolean );
begin
end;

procedure TGeoObj.SetTexInvert(cX,cY:boolean);
begin
end;

procedure TGeoObj.SetTexClamp(cX,cY:boolean);
begin
end;

procedure TGeoObj.SetTexClampValues(MinX,MinY,MaxX,MaxY:single);
begin
end;

procedure TGeoObj.SetTexOffset(cX,cY:single);
begin
end;

procedure TGeoObj.SetTexTiling(cX,cY:single);
begin
end;

procedure TGeoObj.SetTexSwapXY(swap:boolean);
begin
end;



///////////////////////////  TGEOQUAD

constructor TGeoQuad.Create;
begin
  inherited Create;
  Angle := Point3S(0, 0, 0);
  Scale := Point3S(1, 1, 1);
  Position := Point3S(0, 0, 0);
  FBuffer := TDrawBuffer.Create;
  FTexMapper:=TTexMapper.Create(Self);
  FTexMapper.SetInvert(false,false);
  FTexMapper.SetOffset(0,0);
  FTexMapper.SetClamp(false,false);
  FTexMapper.SetTiling(1,1);
  FTexMapper.SetClampValues(-1,-1,1,1);
  FTexMapper.SwapXY:=false;


  FHasNormals := true;
  Width := 200;
  Length := 100;
  Height := 50;

  Coords[0] := Coord(0.01, 0.01);
  Coords[1] := Coord(0.99, 0.01);
  Coords[2] := Coord(0.01, 0.99);
  Coords[3] := Coord(0.99, 0.99);

  Normal := Point3S(0, 0, 1);

  Color := $FFFFFFFF;

end;


constructor TGeoQuad.Create(AWidth, ALength, AHeight: Single; ATex: TTexture);
begin
  Create;
  Width := AWidth;
  Length := ALength;
  Height := AHeight;
  Texture := ATex;

end;

destructor TGeoQuad.Destroy;
begin
  Done;
  FTexMapper.Free;
  FBuffer.Free;
  inherited Destroy;
end;

procedure TGeoQuad.Init;
begin
  inherited Init;
end;

procedure TGeoQuad.Done;
begin
  inherited Done;
  FBuffer.ClearAll;
end;

procedure TGeoQuad.SetWidth(Value: Single);
begin
  FWidth := Value;
  Points[0].X := -FWidth / 2;
  Points[1].X := FWidth / 2;
  Points[2].X := -FWidth / 2;
  Points[3].X := FWidth / 2;
  UpdateBuffer;
end;

procedure TGeoQuad.SetLength(Value: Single);
begin
  FLength := Value;
  Points[0].Y := -FLength / 2;
  Points[1].Y := -FLength / 2;
  Points[2].Y := FLength / 2;
  Points[3].Y := FLength / 2;
  UpdateBuffer;
end;

procedure TGeoQuad.SetHeight(Value: Single);
begin
  FHeight := Value;
  Points[0].Z := FHeight;
  Points[1].Z := FHeight;
  Points[2].Z := FHeight;
  Points[3].Z := FHeight;
  UpdateBuffer;
end;

procedure TGeoQuad.SetTexture(Value: TTexture);
begin
  inherited SetTexture(Value);
  UpdateBuffer;
end;

procedure TGeoQuad.SetColor(Value: TARGB);
begin
  inherited SetColor(Value);
  Colors[0] := FColor;
  Colors[1] := FColor;
  Colors[2] := FColor;
  Colors[3] := FColor;
  UpdateBuffer;
end;


procedure TGeoQuad.SetClockWise(Value: boolean);
var I: LongInt;
  TempPoints: TQuadPoints;
  TempColors: TQuadColors;
  TempCoords: TQuadCoords;
begin
  if FClockWise = Value then Exit;
  FClockWise := Value;

  Negate(Normal);

  TempPoints := Points;
  TempColors := Colors;
  TempCoords := Coords;

  Points[0] := TempPoints[2];
  Points[1] := TempPoints[3];
  Points[2] := TempPoints[0];
  Points[3] := TempPoints[1];

  Colors[0] := TempColors[2];
  Colors[1] := TempColors[3];
  Colors[2] := TempColors[0];
  Colors[3] := TempColors[1];

  Coords[0] := TempCoords[2];
  Coords[1] := TempCoords[3];
  Coords[2] := TempCoords[0];
  Coords[3] := TempCoords[1];


  UpdateBuffer;
end;

procedure TGeoQuad.SetTexInvert(cX,cY:boolean);
begin
 FTexMapper.SetInvert(cX,cY);
 UpdateTexCoords;
end;

procedure TGeoQuad.SetTexClamp(cX,cY:boolean);
begin
FTexMapper.SetClamp(cX,cY);
 UpdateTexCoords;
end;

procedure TGeoQuad.SetTexClampValues(MinX,MinY,MaxX,MaxY:single);
begin
FTexMapper.SetClampValues(MinX,MinY,MaxX,MaxY);
 UpdateTexCoords;
end;

procedure TGeoQuad.SetTexOffset(cX,cY:single);
begin
FTexMapper.SetOffset(cX,cY);
 UpdateTexCoords;
end;

procedure TGeoQuad.SetTexTiling(cX,cY:single);
begin
FTexMapper.SetTiling(cX,cY);
 UpdateTexCoords;
end;

procedure TGeoQuad.SetTexSwapXY(swap:boolean);
begin
 FTexMapper.SwapXY:=swap;
 UpdateTexCoords;
end;

function TGeoQuad.GetInFrustum:boolean;
begin

 //Result:=Camera.IsLineVisible(Bounds.Min,Bounds.Max);
 Result:=Camera.IsSphereVisible(AddPoint(WorldPos,Bounds.Center),Bounds.Rad);
end;

procedure TGeoQuad.Assign(Source: TBaseObj);
var GQ: TGeoQuad;
begin
  inherited Assign(Source);
  if not (Source is TGeoQuad) then Exit;
  GQ := TGeoQuad(Source);
  FHeight:=GQ.Height;
  FLength:=GQ.Length;
  FWidth:=GQ.Width;
  FTexMapper.Copy(GQ.TexMapper);
  Colors:=GQ.Colors;
  Coords:=GQ.Coords;
  Normal:=GQ.Normal;
  Points:=GQ.Points;

  Init;
end;

procedure TGeoQuad.AddNoise(X, Y, Z: single; ReCreate: boolean);
var  I:LongInt;
P:TPoint3S;
begin
  if not Inited then Exit;
 if Recreate then
  begin
   Done;
   Width:=FWidth;
   Length:=FLength;
   Height:=FHeight;
   Init;
  end;

   for I:=0 to 3 do
  begin
    P:=Points[I];
    P.X:=P.X+RandSignF(Random*X);
    P.Y:=P.Y+RandSignF(Random*Y);
    P.Z:=P.Z+RandSignF(Random*Z);
    Points[I]:=P;
  end;
 UpdateBuffer;

 if UseDeviceBuffer then
  begin
   Renderer.FreeDrawBuffer(FBuffer);
   Graphic.DrawBuffer(FBuffer);
  end;

end;

procedure TGeoQuad.AddNoisePercent(X, Y, Z: LongInt; ReCreate: boolean );
begin
 AddNoise(FWidth*X/100,FLength*Y/100,(FWidth+FLength)/2*Z/100,ReCReate);
end;


procedure TGeoQuad.Draw;
begin
  if not Inited then Exit;
  inherited Draw;
  PreDraw;
  Graphic.DrawBuffer(FBuffer);
  AfterDraw;
end;

procedure TGeoQuad.UpdateTexCoords;
begin
 FTexMapper.ProcessBuffer(FBuffer);

 if UseDeviceBuffer then
  begin
   Renderer.FreeDrawBuffer(FBuffer);
   Graphic.DrawBuffer(FBuffer);
  end;

end;

procedure TGeoQuad.UpdateBounds;
var I:LongInt;
    PP,SS,P:TPoint3S;
    Mat:TMatrix;

begin
 inherited;
 Bounds:=BoundInfo(Point3S(1e20,1e20,1e20),Point3S(-1e20,-1e20,-1e20));
 PP:=WorldPos;//GetAbsPosition;
 Mat:=CreateRotationMatrix(Angle);
 SS:=GetAbsScale;


 for I:=0 to 3 do
  begin
   P:=Points[I];
   P:=MatrixTransform(P,Mat);
   P:=MulPoint(P,SS);
   P:=AddPoint(P,PP);


   if P.X<Bounds.Min.X then
      Bounds.Min.X:=P.X;
   if P.Y<Bounds.Min.Y then
      Bounds.Min.Y:=P.Y;
   if P.Z<Bounds.Min.Z then
      Bounds.Min.Z:=P.Z;

   if P.X>Bounds.Max.X then
      Bounds.Max.X:=P.X;
   if P.Y>Bounds.Max.Y then
      Bounds.Max.Y:=P.Y;
   if P.Z>Bounds.Max.Z then
      Bounds.Max.Z:=P.Z;
  end;

 Bounds.Rad:=PointDistance(SubPoint(Bounds.Min,PP),SubPoint(Bounds.Max,PP))/2;

 Bounds.Center.X:=SubPoint(Bounds.Min,PP).X+Bounds.Rad;
 Bounds.Center.Y:=SubPoint(Bounds.Min,PP).Y+Bounds.Rad;
 Bounds.Center.Z:=SubPoint(Bounds.Min,PP).Z+Bounds.Rad;


end;

procedure TGeoQuad.UpdateBuffer;
var Used: TUsedFields;
begin
  if not Inited then Exit;
  UpdateBounds;
  with Graphic do
  begin
    DrawBegin(dmTriangleStrip, FBuffer);

    Used := [ufColors1, ufTexture];
    if FTexture <> nil then
      Include(Used, ufCoords1);

    if HasNormals then
      Include(Used, ufNormals);

    if FUseDeviceBuffer then
      Include(Used, ufDeviceBuff);

    DrawColors1(Colors);
    DrawTexture(FTexture);

    if ufNormals in Used then
      DrawNormal(Normal);

    if ufCoords1 in Used then
      DrawCoords1(Coords);

    DrawPoints(Points);

    DrawCount(2);
    DrawUse(Used);


    DrawEnd(false);
  end;
 UpdateTexCoords;

end;



///////////////////////////  TGEOBOX

constructor TGeoBox.Create;
var Side: TBoxSides;
begin
  inherited Create;

  for Side := Low(TBoxSides) to High(TBoxsides) do
    Buffers[Side] := TDrawBuffer.Create;

  for Side := Low(TBoxSides) to High(TBoxsides) do
  begin
    TexMappers[Side] := TTexMapper.Create(Self);
    TexMappers[Side].SetInvert(false,false);
  TexMappers[Side].SetOffset(0,0);
  TexMappers[Side].SetClamp(false,false);
  TexMappers[Side].SetTiling(1,1);
  TexMappers[Side].SetClampValues(-1,-1,1,1);
  TexMappers[Side].SwapXY:=false;

  end;

  FHasNormals:=true;

  Width := 100;
  Height := 100;
  Length := 100;
  Color := $FFFFFFFF;

  FillChar(VisibleSides, 6, 1);

  Normals[bsFront] := Point3S(0, 1, 0);
  Normals[bsBack] := Point3S(0, -1, 0);
  Normals[bsLeft] := Point3S(1, 0, 0);
  Normals[bsRight] := Point3S(-1, 0, 0);
  Normals[bsTop] := Point3S(0, 0, 1);
  Normals[bsBottom] := Point3S(0, 0, -1);

{  Normals[bsFront] := Point3S(0, -1, 0);
  Normals[bsBack] := Point3S(0, 1, 0);
  Normals[bsLeft] := Point3S(-1, 0, 0);
  Normals[bsRight] := Point3S(1, 0, 0);
  Normals[bsTop] := Point3S(0, 0, -1);
  Normals[bsBottom] := Point3S(0, 0, 1);
 }


  Points[bsFront, 0] := @BasePoints[7];
  Points[bsFront, 1] := @BasePoints[6];
  Points[bsFront, 2] := @BasePoints[3];
  Points[bsFront, 3] := @BasePoints[2];

  Points[bsLeft, 0] := @BasePoints[3];
  Points[bsLeft, 1] := @BasePoints[1];
  Points[bsLeft, 2] := @BasePoints[7];
  Points[bsLeft, 3] := @BasePoints[5];

  Points[bsTop, 0] := @BasePoints[0];
  Points[bsTop, 1] := @BasePoints[1];
  Points[bsTop, 2] := @BasePoints[2];
  Points[bsTop, 3] := @BasePoints[3];

  Points[bsBottom, 0] := @BasePoints[6];
  Points[bsBottom, 1] := @BasePoints[7];
  Points[bsBottom, 2] := @BasePoints[4];
  Points[bsBottom, 3] := @BasePoints[5];

  Points[bsBack, 0] := @BasePoints[1];
  Points[bsBack, 1] := @BasePoints[0];
  Points[bsBack, 2] := @BasePoints[5];
  Points[bsBack, 3] := @BasePoints[4];

  Points[bsRight, 0] := @BasePoints[0];
  Points[bsRight, 1] := @BasePoints[2];
  Points[bsRight, 2] := @BasePoints[4];
  Points[bsRight, 3] := @BasePoints[6];

  SetDefaultCoords;


end;

constructor TGeoBox.Create(AWidth, ALength, AHeight: Single; ATex: TTexture);
begin
  Create;

  Width := AWidth;
  Length := ALength;
  Height := AHeight;
  Texture := ATex;

end;

destructor TGeoBox.Destroy;
var Side: TBoxSides;
begin
  Done;
  for Side := Low(TBoxSides) to High(TBoxsides) do
    Buffers[Side].Free;
  inherited Destroy;
end;

procedure TGeoBox.Init;
begin
  inherited Init;
end;

procedure TGeoBox.Done;
var Side: TBoxSides;
begin
  inherited Done;
  for Side := Low(TBoxSides) to High(TBoxsides) do
    Graphic.DrawBuffer(Buffers[Side]);
end;

procedure TGeoBox.SetDefaultCoords;
begin
Coords[bsFront, 0] := Coord(1, 1);
  Coords[bsFront, 1] := Coord(0, 1);
  Coords[bsFront, 2] := Coord(1, 0);
  Coords[bsFront, 3] := Coord(0, 0);

  Coords[bsLeft, 0] := Coord(0, 0);
  Coords[bsLeft, 1] := Coord(1, 0);
  Coords[bsLeft, 2] := Coord(0, 1);
  Coords[bsLeft, 3] := Coord(1, 1);

  Coords[bsTop, 0] := Coord(0, 0);
  Coords[bsTop, 1] := Coord(1, 0);
  Coords[bsTop, 2] := Coord(0, 1);
  Coords[bsTop, 3] := Coord(1, 1);

  Coords[bsBottom, 0] := Coord(0, 1);
  Coords[bsBottom, 1] := Coord(1, 1);
  Coords[bsBottom, 2] := Coord(0, 0);
  Coords[bsBottom, 3] := Coord(1, 0);

  Coords[bsBack, 0] := Coord(0, 0);
  Coords[bsBack, 1] := Coord(1, 0);
  Coords[bsBack, 2] := Coord(0, 1);
  Coords[bsBack, 3] := Coord(1, 1);

  Coords[bsRight, 0] := Coord(0, 0);
  Coords[bsRight, 1] := Coord(1, 0);
  Coords[bsRight, 2] := Coord(0, 1);
  Coords[bsRight, 3] := Coord(1, 1);
end;

procedure TGeoBox.SetWidth(Value: Single);
begin
  FWidth := Value;
  BasePoints[0].X := -FWidth / 2;
  BasePoints[1].X := FWidth / 2;
  BasePoints[2].X := -FWidth / 2;
  BasePoints[3].X := FWidth / 2;
  BasePoints[4].X := -FWidth / 2;
  BasePoints[5].X := FWidth / 2;
  BasePoints[6].X := -FWidth / 2;
  BasePoints[7].X := FWidth / 2;

  UpdateBuffer;
end;

procedure TGeoBox.SetLength(Value: Single);
begin
  FLength := Value;
  BasePoints[0].Y := -FLength / 2;
  BasePoints[1].Y := -FLength / 2;
  BasePoints[2].Y := FLength / 2;
  BasePoints[3].Y := FLength / 2;
  BasePoints[4].Y := -FLength / 2;
  BasePoints[5].Y := -FLength / 2;
  BasePoints[6].Y := FLength / 2;
  BasePoints[7].Y := FLength / 2;
  UpdateBuffer;
end;

procedure TGeoBox.SetHeight(Value: Single);
begin
  FHeight := Value;
  BasePoints[0].Z := FHeight / 2;
  BasePoints[1].Z := FHeight / 2;
  BasePoints[2].Z := FHeight / 2;
  BasePoints[3].Z := FHeight / 2;
  BasePoints[4].Z := -FHeight / 2;
  BasePoints[5].Z := -FHeight / 2;
  BasePoints[6].Z := -FHeight / 2;
  BasePoints[7].Z := -FHeight / 2;
  UpdateBuffer;
end;

procedure TGeoBox.SetTexture(Value: TTexture);
begin
  inherited SetTexture(Value);
  Textures[bsFront] := FTexture;
  Textures[bsBack] := FTexture;
  Textures[bsLeft] := FTexture;
  Textures[bsRight] := FTexture;
  Textures[bsTop] := FTexture;
  Textures[bsBottom] := FTexture;
  UpdateBuffer;
end;

procedure TGeoBox.SetColor(Value: TARGB);
begin
  inherited SetColor(Value);
  Colors[bsFront] := FColor;
  Colors[bsBack] := FColor;
  Colors[bsLeft] := FColor;
  Colors[bsRight] := FColor;
  Colors[bsTop] := FColor;
  Colors[bsBottom] := FColor;
  UpdateBuffer;
end;


procedure TGeoBox.SetClockWise(Value: boolean);
var Side: TBoxSides;
  TempCoords: TBoxCoords;
  TempPoints: TBoxPoints;

begin
  if FClockWise = Value then Exit;
  FClockWise := Value;

  TempCoords := Coords;
  TempPoints := Points;

  for Side := Low(TBoxSides) to High(TBoxSides) do
  begin
    Negate(Normals[Side]);

    Coords[Side, 0] := TempCoords[Side, 2];
    Coords[Side, 1] := TempCoords[Side, 3];
    Coords[Side, 2] := TempCoords[Side, 0];
    Coords[Side, 3] := TempCoords[Side, 1];

    Points[Side, 0] := TempPoints[Side, 2];
    Points[Side, 1] := TempPoints[Side, 3];
    Points[Side, 2] := TempPoints[Side, 0];
    Points[Side, 3] := TempPoints[Side, 1];

  end;

  UpdateBuffer;
end;

procedure TGeoBox.SetTexInvert(cX,cY:boolean);
var Side:TBoxSides;
begin
 for Side := Low(TBoxSides) to High(TBoxsides) do
  TexMappers[Side].SetInvert(cX,cY);
 UpdateTexCoords;
end;

procedure TGeoBox.SetTexClamp(cX,cY:boolean);
var Side:TBoxSides;
begin
 for Side := Low(TBoxSides) to High(TBoxsides) do
  TexMappers[Side].SetClamp(cX,cY);
 UpdateTexCoords;
end;

procedure TGeoBox.SetTexClampValues(MinX,MinY,MaxX,MaxY:single);
var Side:TBoxSides;
begin
 for Side := Low(TBoxSides) to High(TBoxsides) do
  TexMappers[Side].SetClampValues(MinX,MinY,MaxX,MaxY);
 UpdateTexCoords;
end;

procedure TGeoBox.SetTexOffset(cX,cY:single);
var Side:TBoxSides;
begin
  for Side := Low(TBoxSides) to High(TBoxsides) do
  TexMappers[Side].SetOffset(cX,cY);
 UpdateTexCoords;
end;

procedure TGeoBox.SetTexTiling(cX,cY:single);
var Side:TBoxSides;
begin
 for Side := Low(TBoxSides) to High(TBoxsides) do
  TexMappers[Side].SetTiling(cX,cY);
 UpdateTexCoords;
end;

procedure TGeoBox.SetTexSwapXY(swap:boolean);
var Side:TBoxSides;
begin
 for Side := Low(TBoxSides) to High(TBoxsides) do
  TexMappers[Side].SwapXY:=swap;
 UpdateTexCoords;
end;

procedure TGeoBox.Assign(Source: TBaseObj);
var GB: TGeoBox;
    Side:TBoxSides;

begin
  inherited Assign(Source);
  if not (Source is TGeoBox) then Exit;
  GB := TGeoBox(Source);
  FHeight:=GB.Height;
  FLength:=GB.Length;
  FWidth:=GB.Width;

  for Side := Low(TBoxSides) to High(TBoxsides) do
  TexMappers[Side].Copy(GB.TexMappers[Side]);

  BasePoints:=GB.BasePoints;
  Colors:=GB.Colors;
  Coords:=GB.Coords;
  Normals:=GB.Normals;
  Textures:=GB.Textures;
  VisibleSides:=GB.VisibleSides;

  Init;
end;


function TGeoBox.GetInFrustum:boolean;
begin
// Result:=Camera.IsBoxVisible(Bounds);
 Result:=Camera.IsSphereVisible(AddPoint(WorldPos,Bounds.Center),Bounds.Rad);
end;


procedure TGeoBox.AddNoise(X, Y, Z: single; ReCreate: boolean);
var  I:LongInt;
P:TPoint3S;
Side:TBoxSides;
begin
  if not Inited then Exit;
 if Recreate then
  begin
   Done;
   Width:=FWidth;
   Length:=FLength;
   Height:=FHeight;
   Init;
  end;

  for I:=0 to 7 do
  begin
    P:=BasePoints[I];
    P.X:=P.X+RandSignF(Random*X);
    P.Y:=P.Y+RandSignF(Random*Y);
    P.Z:=P.Z+RandSignF(Random*Z);
    BasePoints[I]:=P;
  end;
 UpdateBuffer;

 if UseDeviceBuffer then
    for Side := Low(TBoxSides) to High(TBoxsides) do
  begin
   Renderer.FreeDrawBuffer(Buffers[Side]);
   Graphic.DrawBuffer(Buffers[Side]);
  end;

end;

procedure TGeoBox.AddNoisePercent(X, Y, Z: LongInt; ReCreate: boolean );
begin
 AddNoise(FWidth*X/100,FLength*Y/100,FHeight*Z/100,ReCreate);
end;

procedure TGeoBox.Draw;
var Side: TBoxSides;
begin
  if not Inited then Exit;
  inherited Draw;
  PreDraw;
  for Side := Low(TBoxSides) to High(TBoxsides) do
    if VisibleSides[Side] then
      Graphic.DrawBuffer(Buffers[Side]);
  AfterDraw;

end;

procedure TGeoBox.UpdateTexCoords;
var Side: TBoxSides;
begin
  for Side := Low(TBoxSides) to High(TBoxsides) do
  TexMappers[Side].ProcessBuffer(Buffers[Side]);

  if UseDeviceBuffer then
  for Side := Low(TBoxSides) to High(TBoxsides) do
  begin
   Renderer.FreeDrawBuffer(Buffers[Side]);
   Graphic.DrawBuffer(Buffers[Side]);
  end;
end;

procedure TGeoBox.UpdateBounds;
var I:LongInt;
    P,PP:TPoint3S;
    SS:TPoint3S;
    Mat:TMatrix;
begin
 inherited;
 PP:=WorldPos;//GetAbsPosition;
 Mat:=CreateRotationMatrix(Angle);
 SS:=GetAbsScale;

 Bounds:=BoundInfo(Point3S(1e20,1e20,1e20),Point3S(-1e20,-1e20,-1e20));
 for I:=0 to 7 do
  begin
   P:=BasePoints[I];
   P:=MatrixTransform(P,Mat);
   P:=MulPoint(P,SS);
   P:=AddPoint(P,PP);


   if P.X<Bounds.Min.X then
      Bounds.Min.X:=P.X;
   if P.Y<Bounds.Min.Y then
      Bounds.Min.Y:=P.Y;
   if P.Z<Bounds.Min.Z then
      Bounds.Min.Z:=P.Z;

   if P.X>Bounds.Max.X then
      Bounds.Max.X:=P.X;
   if P.Y>Bounds.Max.Y then
      Bounds.Max.Y:=P.Y;
   if P.Z>Bounds.Max.Z then
      Bounds.Max.Z:=P.Z;

  end;


 Bounds.Rad:=PointDistance(SubPoint(Bounds.Min,PP),SubPoint(Bounds.Max,PP))/2;

 Bounds.Center.X:=SubPoint(Bounds.Min,PP).X+Bounds.Rad;
 Bounds.Center.Y:=SubPoint(Bounds.Min,PP).Y+Bounds.Rad;
 Bounds.Center.Z:=SubPoint(Bounds.Min,PP).Z+Bounds.Rad;

end;

procedure TGeoBox.UpdateBuffer;
var Used: TUsedFields;
  Side: TBoxSides;
begin
  if not Inited then Exit;
  UpdateBounds;

  Used := [ufColors1, ufCoords1, ufTexture];


  if FUseDeviceBuffer then
    Include(Used, ufDeviceBuff);

  if HasNormals then
    Include(Used, ufNormals);

  with Graphic do
    for Side := Low(TBoxSides) to High(TBoxSides) do
    begin

      DrawBegin(dmTriangleStrip, Buffers[Side]);

      DrawColor1(Colors[Side]);

      if ufNormals in Used then
        DrawNormal(Normals[Side]);

      DrawCoords1(Coords[Side]);

      DrawPoints(Points[Side]);

      DrawTexture(Textures[Side]);

      DrawCount(2);
      DrawUse(Used);


      DrawEnd(false);
    end;
  UpdateTexCoords;
end;




///////////////////////////  TCUSTOMGEO

constructor TCustomGeo.Create;
begin
  inherited Create;
  FBuffers := TDrawBufferMgr.Create();

end;

destructor TCustomGeo.Destroy;
begin

  FBuffers.Free;
  inherited Destroy;
end;

procedure TCustomGeo.Draw;
begin
  inherited Draw;
  PreDraw;
  FBuffers.Draw;
  AfterDraw;

end;

procedure TCustomGeo.Assign(Source: TBaseObj);
var CG: TCustomGeo;
begin
  inherited Assign(Source);
  if not (Source is TCustomGeo) then Exit;
  CG := TCustomGeo(Source);

  // dangerous - change to clone drawbuffers
  Self.FBuffers:=CG.Buffers;
  Init;
end;


end.

