unit Objs3D;

interface

uses Windows, SysUtils, Classes, WinUtils, BaseObjs, Colors, Types3D,
  Utils3D,BaseMgrs;

type


  TBase3DObj = class(TBaseObj)
  private
    FPosition: TPoint3S;
    FDirection: TPoint3S;
    FAngleVector: TPoint3S;
    FAngle: TAngle3S;
    FScale:TPoint3S;

    FParent:TBase3DObj;
    FChildren:TBaseMgr;

    FRotMatrix: TMatrix;
    FTransMatrix:TMatrix;
    FScaleMatrix:TMatrix;

    FWorldMatrix:TMatrix;

    FMovable:boolean;
    FHideChildren:boolean;  // if obj is not drawn, children are not drawn too
    FInitChildren:boolean;  // if obj is inited/done, children are inited/done too

    FDist:single;
    FAChild:TBase3DObj;
    FAChild2:TBase3DObj;

    procedure SetWorldMatrix;

  protected

    function GetPosition: TPoint3S; virtual;
    function GetDirection: TPoint3S; virtual;
    function GetAngle: TAngle3S; virtual;
    function GetScale:TPoint3S;virtual;
    function GetAngleVector: TPoint3S; virtual;

    function GetChildren(Idx:LongInt):TBase3DObj;virtual;
    function GetChildCount:LongInt;virtual;

    procedure SetScale(Value:TPoint3S);virtual;
    procedure SetPosition(Value: TPoint3S); virtual;
    procedure SetAngle(Value: TPoint3S); virtual;
    procedure SetParent(Value:Tbase3DObj);virtual;

    function GetInFrustum: boolean; virtual;
    function GetHasAlpha:boolean;virtual;

  public

    constructor Create; override;
    destructor Destroy; override;

    procedure Init; override;
    procedure Done; override;

    procedure Draw; virtual;  // do not draw children,do not restore world matrix
    procedure DoDraw; virtual;// draw children,restore world matrix
    procedure DoDrawAndUpdate; virtual;// draw children,restore world matrix and call TryUpdate
    procedure Update;override;

    procedure MoveAbs(X,Y,Z:single);virtual;
    procedure MoveByAxes(X, Y, Z: single); virtual;
    procedure MoveByDir(Speed: single); virtual;
    procedure RotateByAxes(X, Y, Z: single); virtual;

    procedure Assign(Source: TBaseObj); override;
    function AngleTo(Obj: TBase3DObj): TAngle3S; overload; virtual;
    function AngleTo(P: TPoint3S): TAngle3S; overload; virtual;
    function DistanceTo(Obj: TBase3DObj): Single; overload;virtual;
    function DistanceTo(P:TPoint3S): Single; overload;virtual;
    procedure LookAt(Obj: TBase3DObj); overload; virtual;
    procedure LookAt(P: TPoint3S); overload; virtual;
    procedure LookAt(X,Y,Z:single);overload;virtual;
    procedure SideStep(Speed:single);virtual; // for strafing

    function GetAbsPosition:TPoint3S; // absolute position
    function GetAbsAngle:TAngle3S;  // absolute angle
    function GetAbsScale:TPoint3S; // absolute scale
    function WorldPos:TPoint3S; // real position in the world (with rotations)

    procedure SetUpWorld;virtual;
    procedure RestoreWorld;virtual;

    function AddChild(AChild:TBase3DObj):TBase3DObj;
    function ExtractChild(AChild:TBase3DObj):TBase3DObj;
    function HasChild(AChild:TBase3DObj):boolean;

    function Clone(AChildren:boolean=true):TBase3DObj;virtual;


    // for scene management - do not use directly
    procedure CalcDist;virtual;
    property  Dist:single read FDist write FDist;
    property AChild:TBase3DObj read FAChild write FAChild;
    //

    property Parent:TBase3DObj read FParent write SetParent;
    property Child[Idx:LongInt]:TBase3DObj read GetChildren;
    property ChildMgr:TBaseMgr read FChildren;
    property ChildCount:LongInt read GetChildCount;
    property RotationMatrix: TMatrix read FRotMatrix;
    property TranslationMatrix: TMatrix read FTransMatrix;
    property ScaleMatrix: TMatrix read FScaleMatrix;
    property WolrdMatrix: TMatrix read FWorldMatrix;
    property InFrustum: boolean read GetInFrustum;
    property HasAlpha:boolean read GetHasAlpha;

  published

    property Position: TPoint3S read GetPosition write SetPosition;
    property Direction: TPoint3S read GetDirection;
    property Angle: TPoint3S read GetAngle write SetAngle;
    property AngleVector: TPoint3S read GetAngleVector;
    property Scale:TPoint3S read GetScale write SetScale;

    property Movable:boolean read FMovable write FMovable;
    property InitChildren:boolean read FInitChildren write FInitChildren;
    property HideChildren:boolean read FHideChildren write FHideChildren;

  end;




implementation

uses GGraphic;

constructor TBase3DObj.Create;
begin
  inherited Create;
  UpdatePeriod:=0;
  Position := Point3S(0, 0, 0);
  Scale:=Point3S(1,1,1);
  Angle := Point3S(0, 0, 0);
  FChildren:=TBaseMgr.Create();

  FMovable:=true;
  FHideChildren:=false;
  FInitChildren:=true;

  FRotMatrix:=IdentityMatrix;
  FTransMatrix:=IdentityMatrix;
  FScaleMatrix:=IdentityMatrix;
  SetWorldMatrix;

end;

destructor TBase3DObj.Destroy;
begin
  FChildren.Free;
  inherited Destroy;
end;


procedure TBase3DObj.Init;
var I:LongInt;
begin
  if FInitChildren then
  for I:=0 to FChildren.Count-1 do
   Child[I].Init;


  inherited Init;
end;

procedure TBase3DObj.Done;
var I:LongInt;
begin
  if FInitChildren then
  for I:=0 to FChildren.Count-1 do
  Child[I].Done;

  inherited Done;
end;

procedure TBase3DObj.Update;
begin
// Draw;
end;

procedure TBase3DObj.SetPosition(Value: TPoint3S);
begin
  if not FMovable then Exit;
  FPosition := Value;
  FTransMatrix:=CreateTranslationMatrix(Position);
  SetWorldMatrix;
end;


procedure TBase3DObj.SetAngle(Value: TAngle3S);
begin
  if SamePoints(FAngle, Value) then Exit;

  if Value.Z>360 then Value.Z:=Value.Z-360;
  if Value.Z<0 then Value.Z:=360-Abs(Value.Z);

  FAngle := Value;
  FRotMatrix := CreateRotationMatrix(FAngle);
  SetWorldMatrix;
  FAngleVector := MatrixTransform(Point3S(0, 1, 0), FRotMatrix);

end;

procedure TBase3DObj.SetScale(Value:TPoint3S);
begin
 FScale:=Value;
 FScaleMatrix:=CreateScaleMatrix(FScale);
 SetWorldMatrix;
end;

procedure TBase3DObj.SetParent(Value:Tbase3DObj);
 begin
  FParent:=Value;
 end;

function TBase3DObj.GetPosition: TPoint3S;
begin
  Result := FPosition;
end;

function TBase3DObj.GetDirection: TPoint3S;
var AV: TPoint3S;
begin
   AV := AngleVector;
    FDirection := Point3S(AV.X + Position.X, AV.Y + Position.Y, AV.Z +
      Position.Z);
  Result := FDirection;
end;

function TBase3DObj.GetAngleVector: TPoint3S;
begin
  Result := FAngleVector;
end;


function TBase3DObj.GetAngle: TAngle3S;
begin
  Result := FAngle;
end;


function TBase3DObj.GetChildren(Idx:LongInt):TBase3DObj;
begin
if (Idx<0)or(Idx>FChildren.Count-1)then
 Result:=TBase3DObj(FChildren[0])
else
 Result:=TBase3DObj(FChildren[Idx]);
end;

function TBase3DObj.GetScale:TPoint3S;
begin
 Result:=FScale;
end;

procedure TBase3DObj.LookAt(Obj: TBase3DObj);
begin
  LookAt(Obj.WorldPos);
end;

procedure TBase3DObj.LookAt(P: TPoint3S);
begin
  Angle := PointToAngle(SubPoint(P,{NullPoint}WorldPos));
end;

procedure TBase3DObj.LookAt(X,Y,Z:single);
begin
 LookAt(Point3S(X,Y,Z));
end;

function TBase3DObj.AngleTo(Obj: TBase3DObj): TAngle3S;
begin
  Result := PointToAngle(SubPoint(Obj.WorldPos, WorldPos));
end;

function TBase3DObj.AngleTo(P: TPoint3S): TAngle3S;
begin
  Result := PointToAngle(SubPoint(P, WorldPos));
end;

function TBase3DObj.DistanceTo(Obj: TBase3DObj): TSingle;
begin
  Result:=PointDistance(WorldPos,Obj.WorldPos);
end;

function TBase3DObj.DistanceTo(P:TPoint3S): TSingle;
begin
  Result:=PointDistance(WorldPos,P);
end;

procedure TBase3DObj.Assign(Source: TBaseObj);
var BO: TBase3DObj;
begin
  inherited Assign(Source);
  if not (Source is TBase3DObj) then Exit;
  BO := TBase3DObj(Source);
  Position := BO.Position;
  Angle := BO.Angle;
  Scale:=BO.Scale;
  HideChildren:=Bo.HideChildren;
  InitChildren:=BO.InitChildren;
  Movable:=BO.Movable;
  Parent:=BO.Parent;

end;

procedure TBase3DObj.MoveAbs(X,Y,Z:single);
begin
 Position:=Point3S(X,Y,Z);
end;

procedure TBase3DObj.MoveByAxes(X, Y, Z: single);
begin
  Position := AddPoint(Position, Point3S(X, Y, Z));
end;

procedure TBase3DObj.MoveByDir(Speed: single);
var P: TPoint3S;
    D: TPoint3S;
begin
  D := Direction;

  P.X := D.X - Position.X;
  P.Y := D.Y - Position.Y;
  P.Z := D.Z - Position.Z;

  Position := Point3S(Position.X + P.X * Speed,
    Position.Y + P.Y * Speed, Position.Z + P.Z * Speed);

end;

procedure TBase3DObj.SideStep(Speed:single);
var P:TPoint3S;
begin
 P:=Position;

 P.X := P.X - Speed * Sin(DegToRad(Angle.Z-90));
 P.Y := P.Y + Speed * Cos(DegToRad(Angle.Z-90));

 Position:=P;

end;

procedure TBase3DObj.RotateByAxes(X, Y, Z: single);
begin
  Angle := AddPoint(Angle, Point3S(X, Y, Z));
end;

procedure TBase3DObj.SetWorldMatrix;
begin
 FWorldMatrix:=MatrixMultiply(FRotMatrix,FScaleMatrix);
 FWorldMatrix:=MatrixMultiply(FWorldMatrix,FTransMatrix);
end;

procedure TBase3DObj.SetUpWorld;
begin
 //if not Inited then Exit;
 Renderer.WorldPush;
 Renderer.WorldMultiply(FWorldMatrix);
end;

procedure TBase3DObj.RestoreWorld;
begin
 Renderer.WorldPop;
end;

procedure TBase3DObj.Draw;
begin
 //SetUpWorld;
end;

procedure TBase3DObj.DoDraw;
var I:LongInt;
begin
 SetUpWorld;
 Draw;

 for I:=0 to FChildren.Count-1 do
  Child[I].DoDraw;

 RestoreWorld;
end;

procedure TBase3DObj.DoDrawAndUpdate;
var I:LongInt;
begin
 TryUpdate;
 SetUpWorld;
 Draw;

 for I:=0 to FChildren.Count-1 do
  Child[I].DoDrawAndUpdate;

 RestoreWorld;
end;

function TBase3DObj.AddChild(AChild:TBase3DObj):TBase3DObj;
begin
 Result:=AChild;
 if AChild=nil then Exit;
 FChildren.Add(AChild);
 AChild.FParent:=Self;
end;

function TBase3DObj.ExtractChild(AChild:TBase3DObj):TBase3DObj;
begin
 Result:=AChild;
 if AChild=nil then Exit;
 FChildren.Extract(AChild);
 AChild.FParent:=nil;
end;

function TBase3DObj.HasChild(AChild:TBase3DObj):boolean;
begin
 Result:=false;
 if AChild=nil then Exit;
 Result:=FChildren.IndexOf(AChild)>-1;
end;


function TBase3DObj.GetChildCount:LongInt;
begin
 Result:=FChildren.Count;
end;

function TBase3DObj.GetAbsPosition:TPoint3S;
var Obj:TBase3DObj;
    P:TPoint3S;
begin
 Result:=NullPoint;
 Obj:=FParent;
 while Obj<>nil do
  begin
   Result:=AddPoint(Result,Obj.Position);
   Obj:=Obj.Parent;
  end;
 Result:=AddPoint(Result,Position);
end;

function TBase3DObj.WorldPos:TPoint3S;
var Obj:TBase3DObj;
    PP:TPoint3S;
begin
 PP:=Position;
 Obj:=FParent;
  while Obj<>nil do
  begin
   PP:=MatrixTransform(PP,Obj.FWorldMatrix);
   PP:=AddPoint(PP,Obj.Position);
   Obj:=Obj.Parent;
  end;
Result:=PP;
end;

function TBase3DObj.GetAbsAngle:TAngle3S;
var Obj:TBase3DObj;
begin
 Result:=NullPoint;
 Obj:=FParent;
 while Obj<>nil do
  begin
   Result:=AddPoint(Result,Obj.Angle);
   Obj:=Obj.Parent;
  end;

 Result:=AddPoint(Result,Angle);
end;

function TBase3DObj.GetAbsScale:TPoint3S;
var Obj:TBase3DObj;
begin
 Result:=NullPoint;
 Obj:=FParent;
 while Obj<>nil do
  begin
   Result:=AddPoint(Result,Obj.Scale);
   Obj:=Obj.Parent;
  end;

 Result:=AddPoint(Result,Scale);
end;

function TBase3DObj.GetInFrustum: boolean;
begin
 Result:=true;
end;

function TBase3DObj.GetHasAlpha: boolean;
begin
 Result:=false;
end;

procedure TBase3DObj.CalcDist;
begin
 FDist:=PointDistance(Camera.WorldPos,WorldPos);
end;

function TBase3DObj.Clone(AChildren:boolean):TBase3DObj;
var Obj,Ch:TBase3DObj;
    I:LongInt;
begin
 Obj:=TBase3DObj(Self.ClassType.Create);
 Obj.Create;
 Obj.Assign(Self);
 if Inited then
  Obj.Init;

 if AChildren then
 for I:=0 to ChildCount-1 do
  begin
   Ch:=Child[I].Clone(AChildren);
   Obj.AddChild(Ch);
  end;

 Result:=Obj;
end;


end.

