unit SkyBase;

interface

uses Windows, SysUtils, Classes, WinUtils, BaseObjs, Objs3D, Types3D, Utils3D,
  VisObjs, GeoObjs, Textures, TexMapping, DrawBuffers, Timing, Colors, GfxTypes;

type

  TSkyBase = class(TVisibleObj)
  private
    FTexture: TTexture;
    FColor: TARGB;
    FUseTexture: boolean;

    FWidth: single;
    FLength: single;
    FHeight: single;

    FViewHeight: single;
      // Height of Viewer in percents  0=bottom  ( z= -FHeight/2)
                        //         100=top  ( z= FHeight/2)
                       //         50=middle (z=0)


  protected

    procedure SetUseTex(Value: boolean); virtual;
    procedure SetTexture(Value: TTexture); virtual;
    procedure SetColor(Value: TARGB); virtual;
    procedure SetDim(Idx: LongInt; Value: single); virtual;

  public
    constructor Create; override;
    destructor Destroy; override;
  published

    property UseTexture: boolean read FUseTexture write SetUseTex;
    property Color: TARGB read FColor write SetColor;
    property Texture: TTexture read FTexture write SetTexture;

    property ViewerHeight: single read FViewHeight write FViewHeight;
    property Height: single index 3 read FHeight write SetDim;
    property Width: single index 1 read FWidth write SetDim;
    property Length: single index 2 read FLength write SetDim;


  end;

  TSkyBox = class;


  TSkyBoxSide = class(TBaseObj)
  private
    FBase: TSkyBase;
    FColor: TARGB;
  protected

    procedure SetColor(Value: TARGB); virtual;

  public
    Texture: TTexture;
    Points: TQuadPoints;
    Colors: TQuadColors;
    Coords: TQuadCoords;
    CoordXSpeed: single; // how much will be added to coords per second
    CoordYSpeed: single;
    TexMapper: TTexMapper;
    Buffer: TDrawBuffer;
    UseDeviceBuffer: boolean;
    UseTexture: boolean;

    constructor Create(ABase: TSkyBase); reintroduce; overload; virtual;

    procedure Update; override;
    procedure UpdateTexCoords;
    procedure Draw;
    procedure Init; override;
    procedure Done; override;
    procedure AnimateCoords;

    property Color: TARGB read FColor write SetColor;

  end;

  TSkyBoxSides = array[TBoxSides] of TSkyBoxSide;

  TSkyBox = class(TSkyBase)
  private


  protected
    procedure BuildPoints; virtual;

    procedure SetUseTex(Value: boolean); override;
    procedure SetTexture(Value: TTexture); override;
    procedure SetColor(Value: TARGB); override;
    procedure SetDim(Idx: LongInt; Value: single); override;

  public
    Sides: TSkyBoxSides;

    constructor Create; overload; override;
    constructor Create(AWidth, ALength, AHeight: single); reintroduce; overload;
      virtual;

    destructor Destroy; override;

    procedure Init; override;
    procedure Done; override;
    procedure Draw; override;
    procedure Update; override;

    procedure SetTextures(Top, Bottom, Front, Back, Left, Right: TTexture);
      overload;
    procedure SetTextures(Top, Bottom, Front, Back, Left, Right: string);
      overload;
    procedure SetTextures(AName: string); overload;
            // load from base name -  'night.jpg' will try to
            // load 'nighttop.jpg','nightbottom.jpg','...left.jpg',
            // '...right.jpg','...front.jpg','...back.jpg'



  published




  end;

  TSkyPlane = class(TSkyBox)
  private
    FTopSize: single;
    FTopIsPlane: boolean;
    FBottomIsPlane: boolean;
    FXSpeed: single;
    FYSpeed: single;
    FUseHorizon: boolean;
    FTopHColor: TARGB;
    FBottomHColor: TARGB;
    FTopHeight: single;
      // z position in percents of top plane ( use with horizons)
    FBottomHeight: single;
      // z position in percents of top plane ( use with horizons)

  protected
    procedure BuildPoints; override;

    procedure SetTopSize(Value: single); virtual;
    procedure SetSpeed(Idx: LongInt; Value: single); virtual;
    procedure SetPlane(Idx: LongInt; Value: boolean); virtual;
    procedure SetHColor(Idx: LongInt; Value: TARGB); virtual;
    procedure SetUseHorizon(Value: boolean); virtual;
    procedure SetHeights(Idx: LongInt; Value: single); virtual;

  public
    constructor Create; override;
    constructor Create(AWidth, ALength, AHeight, ATopSize: single); reintroduce;
      overload;
      virtual;
    destructor Destroy; override;

    procedure InvertXCoords;
    procedure InvertYCoords;
    procedure SwapXYCoords;

  published
    property TopSize: single read FTopSize write SetTopSize;
    property CoordXSpeed: single index 1 read FXSpeed write SetSpeed;
    property CoordYSpeed: single index 2 read FYSpeed write SetSpeed;
    property TopIsPlane: boolean index 1 read FTopIsPLane write SetPlane;
    property BottomIsPlane: boolean index 2 read FBottomIsPLane write SetPlane;
    property UseHorizon: boolean read FUseHorizon write SetUseHorizon;
    property TopHorizonColor: TARGB index 1 read FTopHColor write SetHColor;
    property BottomHorizonColor: TARGB index 2 read FBottomHColor write
    SetHColor;
    property TopHeight: single index 1 read FTopHeight write SetHeights;
    property BottomHeight: single index 2 read FBottomHeight write SetHeights;


  end;



type
  TBoxStrings = array[TBoxSides] of string;

const

  SkySize = 1000.0;
  SkyBoxSide: TBoxStrings = (
    'Bottom', 'Front', 'Back',
    'Left', 'Right', 'Top'
    );



implementation

uses GGraphic, GFileIO;

/////////////////////////// TSKYBASE

constructor TSkyBase.Create;
begin
  inherited Create;
  FUseTexture := true;
  FColor := $FFFFFFFF;

end;

destructor TSkyBase.Destroy;
begin

  inherited Create;
end;

procedure TSkyBase.SetUseTex(Value: boolean);
begin
  FUseTexture := Value;
end;

procedure TSkyBase.SetTexture(Value: TTexture);
begin
  FTexture := Value;
end;

procedure TSkyBase.SetColor(Value: TARGB);
begin
  FColor := Value;
end;

procedure TSkyBase.SetDim(Idx: LongInt; Value: single);
begin
  case Idx of
    1: FWidth := Value;
    2: FLength := Value;
    3: FHeight := Value;
  end;

end;


////////////////////////// TSKYBOXSIDE

constructor TSkyBoxSide.Create(ABase: TSkyBase);
begin
  inherited Create;
  FBase := ABase;
  FColor := $FFFFFFFF;
  UseTexture := true;

  TexMapper := TTexMapper.Create(Self);
  FillMem(@Colors, SizeOf(Colors), Integer(FBase.FColor));

  UpdateTexCoords;

  UseDeviceBuffer := false;
  Buffer := TDrawBuffer.Create;

end;

procedure TSkyBoxSide.Init;
begin
  inherited Init;
  UpdateTexCoords;

end;

procedure TSkyBoxSide.Done;
begin
  inherited Done;
  Buffer.ClearAll;
end;

procedure TSkyBoxSide.SetColor(Value: TARGB);
var I: LongInt;
begin
  FColor := Value;
  for I := 0 to 3 do
    Colors[I] := FColor;
end;

procedure TSkyBoxSide.Draw;
var Used: TUsedFields;
  AdrChanged: boolean;
begin
  if not Inited then Exit;
  AdrChanged := false;
  if (TexMapper.XTiling > 1) or (TexMapper.YTiling > 1) or
    (TexMapper.XTiling < -1) or (TexMapper.YTiling < -1) or
    (CoordXSpeed <> 0) or (CoordYSpeed <> 0) then
  begin
    Renderer.SetDWord(sdTexAdrX, TEXADR_WRAP);
    Renderer.SetDWord(sdTexAdrY, TEXADR_WRAP);
    AdrChanged := true;
  end;


  with Graphic do
  begin
    DrawBegin(dmTriangleStrip, Buffer);

    Used := [ufColors1, ufTexture];
    if Texture <> nil then
      Include(Used, ufCoords1);

    if UseDeviceBuffer then
      Include(Used, ufDeviceBuff);

    DrawColors1(Colors);
    if UseTexture then
      DrawTexture(Texture)
    else
      DrawTexture(nil);

    if ufCoords1 in Used then
    begin
      DrawCoords1(Coords);
    end;

    DrawPoints(Points);

    DrawCount(2);
    DrawUse(Used);

    DrawEnd(true);
  end;
  if AdrChanged then
  begin
    Renderer.SetDWord(sdTexAdrX, TEXADR_CLAMP);
    Renderer.SetDWord(sdTexAdrY, TEXADR_CLAMP);
  end;


end;

procedure TSkyBoxSide.Update;
begin
  inherited Update;
  AnimateCoords;
end;

procedure TSkyBoxSide.UpdateTexCoords;
begin

  Coords[0] := Coord(0, 0);
  Coords[1] := Coord(1, 0);
  Coords[2] := Coord(0, 1);
  Coords[3] := Coord(1, 1);

  if not Inited then Exit;

  TexMapper.ProcessArray(Coords);

end;

procedure TSkyBoxSide.AnimateCoords;
var I: LongInt;
begin
  for I := 0 to 3 do
  begin
    if CoordXSpeed <> 0 then
      Coords[I].X := Coords[I].X + CoordXSpeed /50;
    if CoordYSpeed <> 0 then
      Coords[I].Y := Coords[I].Y + CoordYSpeed / 50;
  end;

end;


/////////////////////////// TSKY BOX

constructor TSkyBox.Create;
var Side: TBoxSides;
begin
  inherited Create;
  for Side := Low(TBoxSides) to High(TBoxSides) do
  begin
    Sides[Side] := TSkyBoxSide.Create(Self);

  end;

  FWidth := SkySize;
  FHeight := SkySize;
  FLength := SkySize;

  BuildPoints;
  FViewHeight := 50;

end;

constructor TSkyBox.Create(AWidth, ALength, AHeight: single);
begin
  Create;
  FWidth := AWidth;
  FLength := ALength;
  FHeight := AHeight;
  BuildPoints;
end;

destructor TSkyBox.Destroy;
var Side: TBoxSides;
begin
  Done;

  for Side := Low(TBoxSides) to High(TBoxSides) do
    Sides[Side].Free;

  inherited Create;
end;


procedure TSkyBox.Init;
var Side: TBoxSides;
begin
  inherited Init;
  for Side := Low(TBoxSides) to High(TBoxSides) do
    Sides[Side].Init;
end;

procedure TSkyBox.Done;
var Side: TBoxSides;
begin
  inherited Done;
  for Side := Low(TBoxSides) to High(TBoxSides) do
    Sides[Side].Done;
end;

procedure TSkyBox.SetUseTex(Value: boolean);
var Side: TBoxSides;
begin
  inherited SetUseTex(Value);
  for Side := Low(TBoxSides) to High(TBoxSides) do
    Sides[Side].UseTexture := FUseTexture;
end;

procedure TSkyBox.SetTexture(Value: TTexture);
begin
  inherited SetTexture(Value);
  SetTextures(Value, Value, Value, Value, Value, Value);
end;

procedure TSkyBox.SetDim(Idx: LongInt; Value: single);
begin
  inherited SetDim(Idx, Value);
  BuildPoints;
end;

procedure TSkyBox.SetColor(Value: TARGB);
var Side: TBoxSides;
begin
  inherited SetColor(Value);
  for Side := Low(TBoxSides) to High(TBoxSides) do
    Sides[Side].Color := FColor;

end;

procedure TSkyBox.BuildPoints;
begin
  Sides[bsFront].Points[0] := Point3S(-FWidth / 2, FLength / 2, FHeight / 2);
  Sides[bsFront].Points[1] := Point3S(FWidth / 2, FLength / 2, FHeight / 2);
  Sides[bsFront].Points[2] := Point3S(-FWidth / 2, FLength / 2, -FHeight / 2);
  Sides[bsFront].Points[3] := Point3S(FWidth / 2, FLength / 2, -FHeight / 2);

  Sides[bsBack].Points[1] := Point3S(-FWidth / 2, -FLength / 2, FHeight / 2);
  Sides[bsBack].Points[0] := Point3S(FWidth / 2, -FLength / 2, FHeight / 2);
  Sides[bsBack].Points[3] := Point3S(-FWidth / 2, -FLength / 2, -FHeight / 2);
  Sides[bsBack].Points[2] := Point3S(FWidth / 2, -FLength / 2, -FHeight / 2);

  Sides[bsTop].Points[0] := Point3S(-FWidth / 2, -FLength / 2, FHeight / 2);
  Sides[bsTop].Points[1] := Point3S(FWidth / 2, -FLength / 2, FHeight / 2);
  Sides[bsTop].Points[2] := Point3S(-FWidth / 2, FLength / 2, FHeight / 2);
  Sides[bsTop].Points[3] := Point3S(FWidth / 2, FLength / 2, FHeight / 2);

  Sides[bsBottom].Points[2] := Point3S(-FWidth / 2, -FLength / 2, -FHeight / 2);
  Sides[bsBottom].Points[3] := Point3S(FWidth / 2, -FLength / 2, -FHeight / 2);
  Sides[bsBottom].Points[0] := Point3S(-FWidth / 2, FLength / 2, -FHeight / 2);
  Sides[bsBottom].Points[1] := Point3S(FWidth / 2, FLength / 2, -FHeight / 2);

  Sides[bsLeft].Points[0] := Point3S(FWidth / 2, FLength / 2, FHeight / 2);
  Sides[bsLeft].Points[1] := Point3S(FWidth / 2, -FLength / 2, FHeight / 2);
  Sides[bsLeft].Points[2] := Point3S(FWidth / 2, FLength / 2, -FHeight / 2);
  Sides[bsLeft].Points[3] := Point3S(FWidth / 2, -FLength / 2, -FHeight / 2);

  Sides[bsRight].Points[1] := Point3S(-FWidth / 2, FLength / 2, FHeight / 2);
  Sides[bsRight].Points[0] := Point3S(-FWidth / 2, -FLength / 2, FHeight / 2);
  Sides[bsRight].Points[3] := Point3S(-FWidth / 2, FLength / 2, -FHeight / 2);
  Sides[bsRight].Points[2] := Point3S(-FWidth / 2, -FLength / 2, -FHeight / 2);
end;


procedure TSkyBox.SetTextures(Top, Bottom, Front, Back, Left, Right: TTexture);
begin
  Sides[bsTop].Texture := Top;
  Sides[bsBottom].Texture := Bottom;
  Sides[bsLeft].Texture := Left;
  Sides[bsRight].Texture := Right;
  Sides[bsFront].Texture := Front;
  Sides[bsBack].Texture := Back;
end;

procedure TSkyBox.SetTextures(Top, Bottom, Front, Back, Left, Right: string);
begin
  Sides[bsTop].Texture := TextureMgr.GetTexture(Top);
  Sides[bsBottom].Texture := TextureMgr.GetTexture(Bottom);
  Sides[bsLeft].Texture := TextureMgr.GetTexture(Left);
  Sides[bsRight].Texture := TextureMgr.GetTexture(Right);
  Sides[bsFront].Texture := TextureMgr.GetTexture(Front);
  Sides[bsBack].Texture := TextureMgr.GetTexture(Back);
end;

procedure TSkyBox.SetTextures(AName: string);
var Base,
  Ext: string;
  Side: TBoxSides;
begin
  Ext := ExtractFileExt(AName);
  Base := Copy(AName, 1, System.Length(AName) - System.Length(Ext));

  for Side := Low(TBoxSides) to High(TBoxSides) do
    if FileIO.IsFile(Base + SkyBoxSide[Side] + Ext) then
      Sides[Side].Texture := TextureMgr.GetTexture(Base + SkyBoxSide[Side] + Ext)
    else
      Sides[Side].Texture := nil;


end;

procedure TSkyBox.Draw;
var
  View: single;
begin
  if not Inited then Exit;

  inherited Draw;

  Graphic.PushFeatures;
  Graphic.PushCamera;
  Graphic.PushDWords;
  Graphic.PushViewPort;
  ViewPort.NearPlane:=0.1;

  Graphic.Disable(gfFog);
  Graphic.Disable(gfLighting);
  Graphic.Disable(gfZBuffer);
  if FViewHeight < 50 then
    View := (100 - FViewHeight * 2) / 100 * (-FHeight / 2)
  else
    View := (FViewHeight - 50) * 2 / 100 * FHeight / 2;

  Camera.Position := Point3S(0, 0, View);
  Renderer.SetDWord(sdCullMode, CULL_NONE);
  Renderer.SetDWord(sdTexAdrX, TEXADR_CLAMP);
  Renderer.SetDWord(sdTexAdrY, TEXADR_CLAMP);

  if (Camera.Angle.X < 40) then
    Sides[bsBottom].Draw;
  if (Camera.Angle.X > -40) then
    Sides[bsTop].Draw;
  if (Camera.Angle.Z < 80) or (Camera.Angle.Z > 280) then
    Sides[bsFront].Draw;
  if (Camera.Angle.Z < 260) and (Camera.Angle.Z > 100) then
    Sides[bsBack].Draw;
  if (Camera.Angle.Z < 350) and (Camera.Angle.Z > 190) then
    Sides[bsLeft].Draw;
  if (Camera.Angle.Z < 170) and (Camera.Angle.Z > 10) then
    Sides[bsRight].Draw;

  Renderer.WorldIdentity;
  Graphic.PopFeatures([gfFog, gfLighting, gfZBuffer]);
  Graphic.PopDWords([sdCullMode, sdTexAdrX, sdTexAdrY]);
  Graphic.PopViewPort;
  Graphic.PopCamera;


end;

procedure TSkyBox.Update;
var Side: TBoxSides;
begin
  if not Inited then Exit;

  for Side := Low(TBoxSides) to High(TBoxSides) do
    Sides[Side].Update;

end;

//////////////////////   TSKY PLANE

constructor TSkyPlane.Create;
begin
  inherited Create;
  FTopSize := SkySize * 3;
  FTopIsPlane := true;
  FUseHorizon := false;
  FTopHColor := $00FFFFFF;
  FBottomHColor := $FFFFFFFF;
  FTopHeight := 100;
  FBottomHeight := 0;

end;

constructor TSkyPlane.Create(AWidth, ALength, AHeight, ATopSize: single);
begin
  Create;
  FTopSize := ATopSize;
end;


destructor TSkyPlane.Destroy;
begin

  inherited Destroy;
end;

procedure TSkyPlane.SetTopSize(Value: single);
begin
  FTopSize := Value;
  BuildPoints;
end;

procedure TSkyPlane.SetSpeed(Idx: LongInt; Value: single);
var Side: TBoxSides;
begin
  for Side := Low(TBoxSides) to High(TBoxSides) do
  begin
    Sides[Side].CoordXSpeed := 0;
    Sides[Side].CoordYSpeed := 0;
  end;

  case Idx of
    1: FXSpeed := Value;
    2: FYSpeed := Value;
  end;

  if FTopIsPlane then
  begin
    Sides[bsTop].CoordXSpeed := FXSpeed;
    Sides[bsTop].CoordYSpeed := FYSpeed;
  end;
  if FBottomIsPlane then
  begin
    Sides[bsBottom].CoordXSpeed := FXSpeed;
    Sides[bsBottom].CoordYSpeed := FYSpeed;
  end;


end;


procedure TSkyPlane.SetPlane(Idx: LongInt; Value: boolean);
begin
  case Idx of
    1: FTopIsPlane := Value;
    2: FBottomIsPlane := Value;
  end;
  SetSpeed(1, FXSpeed);
  SetSpeed(2, FYSpeed);
  BuildPoints;

end;

procedure TSkyPlane.SetHColor(Idx: LongInt; Value: TARGB);
begin
  case Idx of
    1: FTopHColor := Value;
    2: FBottomHColor := Value;
  end;
  SetUseHorizon(FUseHorizon);
end;

procedure TSkyPlane.SetUseHorizon(Value: boolean);
var Side: TBoxSides;
begin
  FUseHorizon := Value;
  if FUseHorizon then
    for Side := bsFront to bsRight do
    begin
      Sides[Side].UseTexture := false;
      Sides[side].Colors[0] := FTopHColor;
      Sides[side].Colors[1] := FTopHColor;
      Sides[side].Colors[2] := FBottomHColor;
      Sides[side].Colors[3] := FBottomHColor;
    end
  else
    for Side := bsFront to bsRight do
    begin
      Sides[Side].UseTexture := FUseTexture;
      FillMem(@Sides[Side].Colors, SizeOf(Sides[Side].Colors),
        Sides[Side].FColor);
    end;
end;

procedure TSkyPlane.SetHeights(Idx: LongInt; Value: single);
begin
  case Idx of
    1: FTopHeight := Value;
    2: FBottomHeight := Value;
  end;
  BuildPoints;
end;

procedure TSkyPlane.BuildPoints;
var I: LongInt;
begin
  inherited BuildPoints;

  for I := 0 to 3 do
  begin
    Sides[bsTop].Points[I].Z := (100 - FTopHeight * 2) / 100 * (-FHeight / 2);
    Sides[bsBottom].Points[I].Z := (FBottomHeight - 50) * 2 / 100 * FHeight / 2;
  end;

  if FTopIsPlane then
  begin
    Sides[bsTop].Points[0] := Point3S(-FTopSize / 2, -FTopSize / 2,
      Sides[bsTop].Points[0].Z);
    Sides[bsTop].Points[1] := Point3S(FTopSize / 2, -FTopSize / 2,
      Sides[bsTop].Points[1].Z);
    Sides[bsTop].Points[2] := Point3S(-FTopSize / 2, FTopSize / 2,
      Sides[bsTop].Points[2].Z);
    Sides[bsTop].Points[3] := Point3S(FTopSize / 2, FTopSize / 2,
      Sides[bsTop].Points[3].Z);
  end;

  if FBottomIsPlane then
  begin
    Sides[bsBottom].Points[0] := Point3S(-FTopSize / 2, -FTopSize / 2,
      Sides[bsBottom].Points[0].Z);
    Sides[bsBottom].Points[1] := Point3S(FTopSize / 2, -FTopSize / 2,
      Sides[bsBottom].Points[1].Z);
    Sides[bsBottom].Points[2] := Point3S(-FTopSize / 2, FTopSize / 2,
      Sides[bsBottom].Points[2].Z);
    Sides[bsBottom].Points[3] := Point3S(FTopSize / 2, FTopSize / 2,
      Sides[bsBottom].Points[3].Z);
  end;

end;


procedure TSkyPlane.InvertXCoords;
var Side: TBoxSides;
  Old: boolean;
begin
  if not Inited then Exit;
  if FTopIsPlane then
    Side := bsTop
  else
    if FBottomIsPlane then
      Side := bsBottom
    else
      Exit;

  Old := Sides[Side].TexMapper.XInvert;
  Sides[Side].TexMapper.XInvert := true;
  Sides[Side].Coords[0] := Coord(0, 0);
  Sides[Side].Coords[1] := Coord(1, 0);
  Sides[Side].Coords[2] := Coord(0, 1);
  Sides[Side].Coords[3] := Coord(1, 1);
  Sides[Side].TexMapper.ProcessArray(Sides[Side].Coords);
  Sides[Side].TexMapper.XInvert := Old;

end;

procedure TSkyPlane.InvertYCoords;
var Side: TBoxSides;
  Old: boolean;
begin
  if not Inited then Exit;
  if FTopIsPlane then
    Side := bsTop
  else
    if FBottomIsPlane then
      Side := bsBottom
    else
      Exit;

  Old := Sides[Side].TexMapper.YInvert;
  Sides[Side].TexMapper.YInvert := true;
  Sides[Side].Coords[0] := Coord(0, 0);
  Sides[Side].Coords[1] := Coord(1, 0);
  Sides[Side].Coords[2] := Coord(0, 1);
  Sides[Side].Coords[3] := Coord(1, 1);
  Sides[Side].TexMapper.ProcessArray(Sides[Side].Coords);
  Sides[Side].TexMapper.YInvert := Old;

end;

procedure TSkyPlane.SwapXYCoords;
var Side: TBoxSides;
  Old: boolean;
begin
  if not Inited then Exit;
  if FTopIsPlane then
    Side := bsTop
  else
    if FBottomIsPlane then
      Side := bsBottom
    else
      Exit;

  Old := Sides[Side].TexMapper.SwapXY;
  Sides[Side].TexMapper.SwapXY := true;
  Sides[Side].Coords[0] := Coord(0, 0);
  Sides[Side].Coords[1] := Coord(1, 0);
  Sides[Side].Coords[2] := Coord(0, 1);
  Sides[Side].Coords[3] := Coord(1, 1);
  Sides[Side].TexMapper.ProcessArray(Sides[Side].Coords);
  Sides[Side].TexMapper.SwapXY := Old;
end;


end.

