unit Textures;

interface

uses Windows,SysUtils,Classes,WinUtils,BaseObjs,Types3D,Graphics,
     Images,Math,GfxTypes,Colors,BaseMgrs;


type

   ETexErr=class(Exception)
   end;


   TTexture=class(TBaseObj)
     private
      FImage:TImage;

      FXFactor:single;
      FYFactor:single;

      FWidth:LongInt;
      FHeight:LongInt;


      protected
       function GetWidth:LongInt;
       function GetHeight:LongInt;
       function GetBits:LongInt;
       function GetScaledSize(Idx:LongInt):LongInt;

       procedure SetImage(Value:TImage);

     public
       DevTex:Pointer;

       DevTexBits:LongInt;
       UseMipMaps:boolean;
       Filter:LongInt;
       ClearAfterInit:boolean;
       InitAfterLoad:boolean;
       Detail:LongInt;
       ColorKey:TARGB;
       ColorKeyEnable:boolean;

       constructor Create(AName:string='');reintroduce;overload;virtual;
       constructor Create(Img:TImage;AName:string='');reintroduce;overload;virtual;

       destructor Destroy; override;

       procedure Init;override;
       procedure Done;override;

       procedure Resize(W,H:LongInt);

       procedure Load(AName:string);
       procedure CheckSize;

       property Image:TImage read FImage write SetImage;
       property Width:LongInt read GetWidth;
       property Height:LongInt read GetHeight;
       property BitCount:LongInt read GetBits;

       // for drawing pictures in orig sizes
       property ScaleXFactor:single read FXFactor;
       property ScaleYFactor:single read FYFactor;
       property ScaledWidth:LongInt index 1 read GetScaledSize;
       property ScaledHeight:LongInt index 2 read GetScaledSize;

     published

     end;


const

 sTexErr='Error in TTexture.';

var

 // Texture Presets

 TextureDetail:LongInt=1;     // powers of 2; higher = lesser img resolution
 TextureBitCount:LongInt=32 div 1;
  // 32 bit or 16 bit
  // 16 bit - if image has no alpha channel 5 bits are used for each color channel  1 bit for alpha
  //             if alpha is used there are 4 bits for each channel

 TextureUseMipMaps:boolean=true;
 TextureFilter:LongInt=TEXF_TRILINEAR;
 TextureClearAfterInit:boolean=true;
 TextureInitAfterLoad:boolean=true;
 TextureColorKey:TARGB=$FFFF00FF;
 TextureColorKeyEnable:boolean=true;


 TextureMinSize:LongInt=16;
 TextureMaxSize:LongInt=1024;



 function CreateTexture(FileName:string):TTexture;overload;
 function CreateTexture(FileName:string;Width,Height:LongInt):TTexture;overload;
 function CreateTexture(FileName:string;Bits:LongInt;Detail:LongInt;
   Filter:LongInt):TTexture;overload;
 function CreateTexture(FileName:string;Bits:LongInt;Detail:LongInt=1;
  Filter:LongInt=TEXF_BILINEAR;UseMipMaps:boolean=true;
  InitAfterLoad:boolean=true;ClearAfterInit:boolean=true
   ):TTexture;overload;

var

  TextureList:TBaseMgr;    // stores all created textures


implementation

uses
 GGraphic,TextureMgrs;

function CreateTexture(FileName:string):TTexture;
begin
 Result:=CreateTexture(FileName,TextureBitCount,TextureDetail,
  TextureFilter,TextureUseMipMaps,TextureInitAfterLoad,
  TextureClearAfterInit);
end;

function CreateTexture(FileName:string;Width,Height:LongInt):TTexture;
begin
 Result:=CreateTexture(FileName,TextureBitCount,TextureDetail,
  TextureFilter,TextureUseMipMaps,TextureInitAfterLoad,
  TextureClearAfterInit);
 if Result<>nil then
 Result.Resize(Width,Height);

end;


function CreateTexture(FileName:string;Bits:LongInt;
 Detail:LongInt;Filter:LongInt):TTexture;
begin
 Result:=CreateTexture(FileName,Bits,Detail,
  Filter,TextureUseMipMaps,TextureInitAfterLoad,
  TextureClearAfterInit);
end;


function CreateTexture(FileName:string;Bits:LongInt;Detail:LongInt;
  Filter:LongInt;UseMipMaps:boolean;InitAfterLoad:boolean;ClearAfterInit:boolean   ):TTexture;
var Tex:TTexture;
    Old:boolean;
 begin

 Old:=TextureInitAfterLoad;
 TextureInitAfterLoad:=false;
 Tex:=TTexture.Create(FileName);
 TextureInitAfterLoad:=Old;
 Tex.Name:=FileName;
 Tex.UseMipMaps:=UseMipMaps;
 Tex.Filter:=Filter;
 Tex.ClearAfterInit:=ClearAfterInit;
 Tex.InitAfterLoad:=InitAfterLoad;
 Tex.Detail:=Detail;
 Tex.DevTexBits:=Bits;

 Tex.Init;
 Result:=Tex;

end;





constructor TTexture.Create(AName:string);
begin
 inherited Create;
 ClearAfterInit:=TextureClearAfterInit;
 InitAfterLoad:=TextureInitAfterLoad;
 DevTexBits:=TextureBitCount;
 Detail:=TextureDetail;
 UseMipMaps:=TextureUseMipMaps;
 ColorKey:=TextureColorKey;
 ColorKeyEnable:=TextureColorKeyEnable;
 Filter:=TextureFilter;

 Name:=AName;
 FImage:=TImage.Create;
 if AName<>'' then
  Load(AName);

 //TextureList.Add(Self);

end;

constructor TTexture.Create(Img:TImage;AName:string='');
begin
 Create();
 Image:=Img;
 if AName<>'' then
 Name:=AName;
end;


destructor TTexture.Destroy;
begin
 Done;
 //TextureList.Extract(Self);
 FImage.Free;
 inherited Destroy;
end;

procedure TTexture.Init;
begin
 //if Inited then Exit;
 try
 if FImage=nil then
  begin
   FImage:=TImage.Create;
   if Name<>'' then
    FImage.Load(Name);

  end;
 Done;
 CheckSize;
 Renderer.CreateDevTex(Self);
 if ClearAfterInit then
  begin FImage.Free;FImage:=nil;end;

 inherited Init;

 except
  raise ETexErr.Create(sTexErr+'Init');
 end;

end;


procedure TTexture.Done;
begin
 if not Inited then Exit;
 try
 inherited Done;
 Renderer.DeleteDevTex(Self);
 except
  raise ETexErr.Create(sTexErr+'Done');
 end;

end;


function TTexture.GetWidth:LongInt;
begin
 if FImage<>nil then
 Result:=FImage.Width
 else
 Result:=FWidth;
end;

function TTexture.GetHeight:LongInt;
begin
 if FImage<>nil then
 Result:=FImage.Height
 else
 Result:=FHeight;
end;

function TTexture.GetScaledSize(Idx:LongInt):LongInt;
begin
 Result:=0;
 case Idx of
 1:Result:=Round(FWidth/FXFactor);
 2:Result:=Round(FHeight/FYFactor);
 end;

end;

function TTexture.GetBits:LongInt;
begin
 Result:=DevTexBits;
 if FImage<>nil then
 Result:=FImage.BitCount;
end;

procedure TTexture.Resize(W,H:LongInt);
begin
 if FImage=nil then
  begin
   FImage:=TImage.Create;
   if Name<>'' then
    FImage.Load(Name);
   FImage.Resize(W,H);
  end;

 if not Inited then Exit;
 Done;
 Init;


end;


procedure TTexture.CheckSize;
var oW,oH,W,H:LongInt;
    sX,sY:single;

begin
 if FImage=nil then
  raise ETexErr.Create(sTexErr+'CheckSize - FImage=nil');

 try
 oW:=FImage.Width;
 oH:=FImage.Height;

 sX:=Log2(oW);
 sY:=Log2(oH);

 if Trunc(sX)<>sX then
  sX:=Trunc(sX)+1;

 if Trunc(sY)<>sY then
  sY:=Trunc(sY)+1;

 W:=Round(Power(2,sX));
 H:=Round(Power(2,sY));

 W:=W div Detail;
 H:=H div Detail;

 W:=WinUtils.Max(TextureMinSize,WinUtils.Min(TextureMaxSize,W));
 H:=WinUtils.Max(TextureMinSize,WinUtils.Min(TextureMaxSize,H));

 FImage.BilinearResize:=false;
 FImage.Resize(W,H);


 FXFactor:=W*Detail/oW;
 FYFactor:=H*Detail/oH;

 FWidth:=W;
 FHeight:=H;

 except
  raise ETexErr.Create(sTexErr+'CheckSize');
 end;

 end;




procedure TTexture.Load(AName:string);
begin
 if FImage=nil then
  FImage:=TImage.Create;
 FImage.Load(AName);
 Name:=AName;
 //if InitAfterLoad then
 // Init;

end;


procedure TTexture.SetImage(Value:TImage);
begin
 FImage.Free;
 FImage:=nil;
 if Value<>nil then
 begin
 FImage:=Value.Clone;
 Name:=Value.Filename;
 end;
 if FImage<>nil then
 begin
 //Init;
 end;

end;



initialization

 TextureList:=TBaseMgr.Create();

finalization

 TextureList.Free;

end.
